Package: Rsolnp
Type: Package
Title: General Non-Linear Optimization
Version: 2.0.1
Date: 2025-06-30
Authors@R: c(person("Alexios", "Galanos", role = c("aut","cre","cph"), email = "alexios@4dscape.com", comment = c(ORCID = "0009-0000-9308-0457")), person("Stefan", "Theussl", role = c("ctb"), email = "Stefan.Theussl@R-project.org", comment = c(ORCID = "0000-0002-6523-4620")), person("Yinyu", "Ye", role = "aut", comment = "Original author of the solnp Matlab code, Stanford University"))
Maintainer: Alexios Galanos <alexios@4dscape.com>
Depends: R (>= 3.0.2)
LinkingTo: Rcpp (>= 0.10.6), RcppArmadillo
Imports: Rcpp, truncnorm, parallel, stats, numDeriv, future.apply
Description: General Non-linear Optimization Using Augmented Lagrange Multiplier Method.
LazyLoad: yes
License: GPL-2
Encoding: UTF-8
NeedsCompilation: yes
RoxygenNote: 7.3.2
Suggests: knitr, rmarkdown
VignetteBuilder: knitr
Packaged: 2025-06-30 03:33:57 UTC; alexios
Author: Alexios Galanos [aut, cre, cph] (ORCID:
    <https://orcid.org/0009-0000-9308-0457>),
  Stefan Theussl [ctb] (ORCID: <https://orcid.org/0000-0002-6523-4620>),
  Yinyu Ye [aut] (Original author of the solnp Matlab code, Stanford
    University)
Repository: CRAN
Date/Publication: 2025-06-30 04:20:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 21:26:18 UTC; unix
Archs: Rsolnp.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,solnp_kkt_summary)
export(csolnp)
export(csolnp_ms)
export(gosolnp)
export(kkt_diagnose)
export(solnp)
export(solnp_problem_suite)
export(solnp_problems_table)
export(solnp_standardize_problem)
export(startpars)
importFrom(Rcpp,evalCpp)
importFrom(future.apply,future_lapply)
importFrom(numDeriv,grad)
importFrom(numDeriv,hessian)
importFrom(numDeriv,jacobian)
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(parallel,parLapply)
importFrom(stats,optim)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(truncnorm,rtruncnorm)
importFrom(utils,tail)
useDynLib(Rsolnp, .registration=TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo=FALSE--------------------------------------------------------
library(Rsolnp)

## -----------------------------------------------------------------------------
args(csolnp)

## -----------------------------------------------------------------------------
args(solnp)

## -----------------------------------------------------------------------------
prob <- solnp_problem_suite(number = 10)
sol <- csolnp(pars = prob$start, fn = prob$fn, gr = prob$gr, eq = prob$eq_fn, eq_b = prob$eq_b, 
              eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, ineq_lower = prob$ineq_lower, 
              ineq_upper = prob$ineq_upper, ineq_jac = prob$ineq_jac, lower = prob$lower,
              upper = prob$upper)
print(prob$name)
print(c("convergence" = sol$convergence))
print(c("csolnp objective" = sol$objective, "best objective" = prob$best_fn))
print(sol$elapsed_time)

## -----------------------------------------------------------------------------
prob <- solnp_problem_suite(suite = "Other", number = 4)
sol <- csolnp(pars = prob$start, fn = prob$fn, gr = prob$gr, eq = prob$eq_fn, eq_b = prob$eq_b, 
              eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, ineq_lower = prob$ineq_lower, 
              ineq_upper = prob$ineq_upper, ineq_jac = prob$ineq_jac, lower = prob$lower,
              upper = prob$upper)
print(prob$name)
print(c("convergence" = sol$convergence))
print(c("csolnp objective" = sol$objective, "best objective" = prob$best_fn))
print(sol$elapsed_time)

## ----warning=FALSE------------------------------------------------------------
prob <- solnp_problem_suite(number = 55)
sol <- csolnp_ms(prob$fn, gr = prob$gr, eq_fn = prob$eq_fn, eq_b = prob$eq_b, 
              eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, ineq_lower = prob$ineq_lower, 
              ineq_upper = prob$ineq_upper, ineq_jac = prob$ineq_jac, lower = prob$lower,
              upper = prob$upper, n_candidates = 200, penalty = 1e5, 
              control = list(min_iter = 1000, max_iter = 100, tol = 1e-8),
              seed = 300)

print(paste0("Solution : ", round(sol$objective,3), " | Best Objective :", round(prob$best_fn,3)))
print(paste0("Equaility Violation : ", round(sol$kkt_diagnostics$eq_violation,3)))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo=FALSE--------------------------------------------------------
library(Rsolnp)

## -----------------------------------------------------------------------------
head(solnp_problems_table())

## -----------------------------------------------------------------------------
prob <- solnp_problem_suite(suite = "Hock-Schittkowski", number = 60)
sol <- csolnp(prob$start, fn = prob$fn, gr = prob$gr, eq_fn = prob$eq_fn, 
              eq_b = prob$eq_b, eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, 
              ineq_lower = prob$ineq_lower, ineq_upper = prob$ineq_upper, 
              ineq_jac = prob$ineq_jac, lower = prob$lower, upper = prob$upper, 
              control = list(trace = 0, min_iter = 1000, max_iter = 300, tol = 1e-8, rho = 1))
print(c("solnp" = sol$objective, "benchmark" = prob$best_fn))
print(rbind("solnp" = sol$pars, "benchmark" = prob$best_par))

#################################################################################
##
##   R package Rsolnp by Alexios Ghalanos and Stefan Theussl Copyright (C) 2009-2013
##   This file is part of the R package Rsolnp.
##
##   The R package Rsolnp is free software: you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation, either version 3 of the License, or
##   (at your option) any later version.
##
##   The R package Rsolnp is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.
##
#################################################################################
.onLoad <- function(lib, pkg) {
	if(interactive() || getOption("verbose")){
		packageStartupMessage(sprintf("Package %s (%s) loaded.  To cite, see citation(\"%s\")\n",pkg, utils::packageDescription(pkg)$Version, pkg))
	}
}

#' Random Initialization and Multiple Restarts of the solnp solver.
#'
#' When the objective function is non-smooth or has many local minima, it is hard
#' to judge the optimality of the solution, and this usually depends critically on
#' the starting parameters. This function enables the generation of a set of
#' randomly chosen parameters from which to initialize multiple restarts of the
#' solver (see note for details).
#'
#' @param pars The starting parameter vector. This is not required unless the fixed option is
#' also used.
#' @param fixed The numeric index which indicates those parameters which should stay fixed
#' instead of being randomly generated.
#' @param fun The main function which takes as first argument the parameter vector and returns
#' a single value.
#' @param eqfun (Optional) The equality constraint function returning the vector of evaluated
#' equality constraints.
#' @param eqB (Optional) The equality constraints.
#' @param ineqfun (Optional) The inequality constraint function returning the vector of evaluated
#' inequality constraints.
#' @param ineqLB (Optional) The lower bound of the inequality constraints.
#' @param ineqUB (Optional) The upper bound of the inequality constraints.
#' @param LB The lower bound on the parameters. This is not optional in this function.
#' @param UB The upper bound on the parameters. This is not optional in this function.
#' @param control (Optional) The control list of optimization parameters. The \code{eval.type}
#' option in this control list denotes whether to evaluate the function as is and
#' exclude inequality violations in the final ranking (default, value = 1),
#' else whether to evaluate a penalty barrier function comprised of the objective
#' and all constraints (value = 2). See \code{solnp} function documentation for
#' details of the remaining control options.
#' @param distr A numeric vector of length equal to the number of parameters, indicating the
#' choice of distribution to use for the random parameter  generation. Choices are
#' uniform (1), truncated normal (2), and normal (3).
#' @param distr.opt If any choice in \code{distr} was anything other than uniform (1), this is a
#' list equal to the length of the parameters with sub-components for the mean and
#' sd, which are required in the truncated normal and normal distributions.
#' @param n.restarts The number of solver restarts required.
#' @param n.sim The number of random parameters to generate for every restart of the solver.
#' Note that there will always be significant rejections if inequality bounds are
#' present. Also, this choice should also be motivated by the width of the upper
#' and lower bounds.
#' @param cluster If you want to make use of parallel functionality, initialize and pass a cluster
#' object from the parallel package (see details), and remember to terminate it!
#' @param rseed (Optional) A seed to initiate the random number generator, else system time will
#' be used.
#' @param ... (Optional) Additional parameters passed to the main, equality or inequality
#' functions
#' @details
#' Given a set of lower and upper bounds, the function generates, for those
#' parameters not set as fixed, random values from one of the 3 chosen
#' distributions. Depending on the \code{eval.type} option of the \code{control}
#' argument, the function is either directly evaluated for those points not
#' violating any inequality constraints, or indirectly via a penalty barrier
#' function jointly comprising the objective and constraints. The resulting values
#' are then sorted, and the best N (N = random.restart) parameter vectors
#' (corresponding to the best N objective function values) chosen in order to
#' initialize the solver. Since version 1.14, it is up to the user to prepare and
#' pass a cluster object from the parallel package for use with gosolnp, after
#' which the parLapply function is used. If your function makes use of additional
#' packages, or functions, then make sure to export them via the \code{clusterExport}
#' function of the parallel package. Additional arguments passed to the solver via the
#' \dots option are evaluated and exported by gosolnp to the cluster.
#' @return A list containing the following values:
#' \item{pars}{Optimal Parameters.}
#' \item{convergence }{Indicates whether the solver has converged (0) or not (1).}
#' \item{values}{Vector of function values during optimization with last one the
#' value at the optimal.}
#' \item{lagrange}{The vector of Lagrange multipliers.}
#' \item{hessian}{The Hessian at the optimal solution.}
#' \item{ineqx0}{The estimated optimal inequality vector of slack variables used
#' for transforming the inequality into an equality constraint.}
#' \item{nfuneval}{The number of function evaluations.}
#' \item{elapsed}{Time taken to compute solution.}
#' \item{start.pars}{The parameter vector used to start the solver}
#' @references
#' Y.Ye, \emph{Interior algorithms for linear, quadratic, and linearly constrained
#' non linear programming}, PhD Thesis, Department of EES Stanford University,
#' Stanford CA.\cr
#' Hu, X. and Shonkwiler, R. and Spruill, M.C. \emph{Random Restarts in Global
#' Optimization}, 1994, Georgia Institute of technology, Atlanta.
#' @author Alexios Galanos and Stefan Theussl\cr
#' Y.Ye (original matlab version of solnp)
#' @note
#' The choice of which distribution to use for randomly sampling the parameter
#' space should be driven by the user's knowledge of the problem and confidence or
#' lack thereof of the parameter distribution. The uniform distribution indicates
#' a lack of confidence in the location or dispersion of the parameter, while the
#' truncated normal indicates a more confident choice in both the location and
#' dispersion. On the other hand, the normal indicates perhaps a lack of knowledge
#' in the upper or lower bounds, but some confidence in the location and dispersion
#' of the parameter. In using choices (2) and (3) for \code{distr},
#' the \code{distr.opt} list must be supplied with \code{mean} and \code{sd} as
#' subcomponents for those parameters not using the uniform (the examples section
#' hopefully clarifies the usage).
#' @examples
#' \dontrun{
#' # [Example 1]
#' # Distributions of Electrons on a Sphere Problem:
#' # Given n electrons, find the equilibrium state distribution (of minimal Coulomb
#' # potential) of the electrons positioned on a conducting sphere. This model is
#' # from the COPS benchmarking suite. See http://www-unix.mcs.anl.gov/~more/cops/.
#' gofn = function(dat, n)
#' {
#'
#' 	x = dat[1:n]
#' 	y = dat[(n+1):(2*n)]
#' 	z = dat[(2*n+1):(3*n)]
#' 	ii = matrix(1:n, ncol = n, nrow = n, byrow = TRUE)
#' 	jj = matrix(1:n, ncol = n, nrow = n)
#' 	ij = which(ii<jj, arr.ind = TRUE)
#' 	i = ij[,1]
#' 	j = ij[,2]
#' 	#  Coulomb potential
#' 	potential = sum(1.0/sqrt((x[i]-x[j])^2 + (y[i]-y[j])^2 + (z[i]-z[j])^2))
#' 	potential
#' }
#'
#' goeqfn = function(dat, n)
#' {
#' 	x = dat[1:n]
#' 	y = dat[(n+1):(2*n)]
#' 	z = dat[(2*n+1):(3*n)]
#' 	apply(cbind(x^2, y^2, z^2), 1, "sum")
#' }
#'
#' n = 25
#' LB = rep(-1, 3*n)
#' UB = rep(1,  3*n)
#' eqB = rep(1, n)
#' ans = gosolnp(pars  = NULL, fixed = NULL, fun = gofn, eqfun = goeqfn, eqB = eqB,
#' LB = LB, UB = UB, control = list(outer.iter = 100, trace = 1),
#' distr = rep(1, length(LB)), distr.opt = list(), n.restarts = 2, n.sim = 20000,
#' rseed = 443, n = 25)
#' # should get a function value around 243.813
#'
#' # [Example 2]
#' # Parallel functionality for solving the Upper to Lower CVaR problem (not properly
#' # formulated...for illustration purposes only).
#'
#' mu =c(1.607464e-04, 1.686867e-04, 3.057877e-04, 1.149289e-04, 7.956294e-05)
#' sigma = c(0.02307198,0.02307127,0.01953382,0.02414608,0.02736053)
#' R = matrix(c(1, 0.408, 0.356, 0.347, 0.378,  0.408, 1, 0.385, 0.565, 0.578, 0.356,
#' 0.385, 1, 0.315, 0.332, 0.347, 0.565, 0.315, 1, 0.662, 0.378, 0.578,
#' 0.332, 0.662, 1), 5,5, byrow=TRUE)
#' # Generate Random deviates from the multivariate Student distribution
#' set.seed(1101)
#' v = sqrt(rchisq(10000, 5)/5)
#' S = chol(R)
#' S = matrix(rnorm(10000 * 5), 10000) %*% S
#' ret = S/v
#' RT = as.matrix(t(apply(ret, 1, FUN = function(x) x*sigma+mu)))
#' # setup the functions
#' .VaR = function(x, alpha = 0.05)
#' {
#' 	VaR = quantile(x, probs = alpha, type = 1)
#' 	VaR
#' }
#'
#' .CVaR = function(x, alpha = 0.05)
#' {
#' 	VaR = .VaR(x, alpha)
#' 	X = as.vector(x[, 1])
#' 	CVaR = VaR - 0.5 * mean(((VaR-X) + abs(VaR-X))) / alpha
#' 	CVaR
#' }
#' .fn1 = function(x,ret)
#' {
#' 	port=ret%*%x
#' 	obj=-.CVaR(-port)/.CVaR(port)
#' 	return(obj)
#' }
#'
#' # abs(sum) of weights ==1
#' .eqn1  = function(x,ret)
#' {
#' 	sum(abs(x))
#' }
#'
#' LB=rep(0,5)
#' UB=rep(1,5)
#' pars=rep(1/5,5)
#' ctrl = list(delta = 1e-10, tol = 1e-8, trace = 0)
#' cl = makePSOCKcluster(2)
#' # export the auxilliary functions which are used and cannot be seen by gosolnp
#' clusterExport(cl, c(".CVaR", ".VaR"))
#' ans = gosolnp(pars, fun = .fn1, eqfun = .eqn1, eqB = 1, LB = LB, UB = UB,
#' n.restarts = 2, n.sim=500, cluster = cl, ret = RT)
#' ans
#' # don't forget to stop the cluster!
#' stopCluster(cl)
#' }
#' @export
gosolnp = function(pars = NULL, fixed = NULL, fun, eqfun = NULL, eqB = NULL, ineqfun = NULL, ineqLB = NULL,
		ineqUB = NULL, LB = NULL, UB = NULL, control = list(), distr = rep(1, length(LB)), distr.opt = list(),
		n.restarts = 1, n.sim = 20000, cluster = NULL, rseed = NULL, ...)
{
    # allowed distributions:
    # 1: uniform (no confidence in the location of the parameter...somewhere in LB-UB space)
    # 2: truncnorm (high confidence in the location of the parameter)
    # 3: normal (Uncertainty in Lower and Upper bounds, but some idea about the dispersion about the location)
    # ...
	if( !is.null(pars) ) gosolnp_parnames = names(pars) else gosolnp_parnames = NULL
	if(is.null(control$trace)) trace = FALSE else trace = as.logical(control$trace)
	if(is.null(control$eval.type)) parmethod = 1 else parmethod = as.integer(min(abs(control$eval.type),2))
	if(parmethod == 0) parmethod = 1
	control$eval.type = NULL
	# use a seed to initialize random no. generation
	if(is.null(rseed)) rseed = as.numeric(Sys.time()) else rseed = as.integer(rseed)
	# function requires both upper and lower bounds
	if(is.null(LB))
		stop("\ngosolnp-->error: the function requires lower parameter bounds\n", call. = FALSE)
	if(is.null(UB))
		stop("\ngosolnp-->error: the function requires upper parameter bounds\n", call. = FALSE)
	# allow for fixed parameters (i.e. non randomly chosen), but require pars vector in that case
	if(!is.null(fixed) && is.null(pars))
		stop("\ngosolnp-->error: you need to provide a pars vector if using the fixed option\n", call. = FALSE)
	if(!is.null(pars)) n = length(pars) else n = length(LB)

	np = 1:n

	if(!is.null(fixed)){
		# make unique
		fixed = unique(fixed)
		# check for violations in indices
		if(any(is.na(match(fixed, np))))
			stop("\ngosolnp-->error: fixed indices out of bounds\n", call. = FALSE)
	}
	# check distribution options
	# truncated normal
	if(any(distr == 2)){
		d2 = which(distr == 2)
		for(i in 1:length(d2)) {
			if(is.null(distr.opt[[d2[i]]]$mean))
				stop(paste("\ngosolnp-->error: distr.opt[[,",d2[i],"]] missing mean\n", sep = ""), call. = FALSE)
			if(is.null(distr.opt[[d2[i]]]$sd))
				stop(paste("\ngosolnp-->error: distr.opt[[,",d2[i],"]] missing sd\n", sep = ""), call. = FALSE)
		}
	}
	#  normal
	if(any(distr == 3)){
		d3 = which(distr == 3)
		for(i in 1:length(d3)) {
			if(is.null(distr.opt[[d3[i]]]$mean))
				stop(paste("\ngosolnp-->error: distr.opt[[,",d3[i],"]] missing mean\n", sep = ""), call. = FALSE)
			if(is.null(distr.opt[[d3[i]]]$sd))
				stop(paste("\ngosolnp-->error: distr.opt[[,",d3[i],"]] missing sd\n", sep = ""), call. = FALSE)
		}
	}
	# setup cluster exports:
	if( !is.null(cluster) ){
		clusterExport(cluster, c("gosolnp_parnames", "fun", "eqfun",
						"eqB", "ineqfun", "ineqLB", "ineqUB", "LB", "UB"), envir = environment())
		if( !is.null(names(list(...))) ){
			# evaluate promises
			xl = names(list(...))
			for(i in 1:length(xl)){
				eval(parse(text=paste(xl[i],"=list(...)[[i]]",sep="")))
			}
			clusterExport(cluster, names(list(...)), envir = environment())
		}
		clusterEvalQ(cluster, require(Rsolnp))
	}
	# initiate random search
	gosolnp_rndpars = switch(parmethod,
			.randpars(pars = pars, fixed = fixed, fun = fun, eqfun = eqfun,
					eqB = eqB, ineqfun = ineqfun, ineqLB = ineqLB,
					ineqUB = ineqUB, LB = LB, UB = UB, distr = distr,
					distr.opt = distr.opt, n.restarts = n.restarts,
					n.sim = n.sim, trace = trace, rseed = rseed,
					gosolnp_parnames = gosolnp_parnames, cluster = cluster, ...),
			.randpars2(pars = pars, fixed = fixed, fun = fun, eqfun = eqfun,
					eqB = eqB, ineqfun = ineqfun, ineqLB = ineqLB,
					ineqUB = ineqUB, LB = LB, UB = UB, distr = distr,
					distr.opt = distr.opt, n.restarts = n.restarts,
					n.sim = n.sim, rseed = rseed, trace = trace,
					gosolnp_parnames = gosolnp_parnames, cluster = cluster, ...))

	gosolnp_rndpars = gosolnp_rndpars[,1:n, drop = FALSE]
	# initiate solver restarts
	if( trace ) cat("\ngosolnp-->Starting Solver\n")
	solution = vector(mode = "list", length = n.restarts)
	if( !is.null(cluster) )
	{
		clusterExport(cluster, c("gosolnp_rndpars"), envir = environment())
		solution = parLapply(cluster, as.list(1:n.restarts), fun = function(i) {
					xx = gosolnp_rndpars[i,]
					names(xx) = gosolnp_parnames
					ans = try(solnp(pars = xx, fun = fun, eqfun = eqfun,
									eqB = eqB, ineqfun = ineqfun,
									ineqLB = ineqLB, ineqUB = ineqUB,
									LB = LB, UB = UB,
									control = control, ...), silent = TRUE)
					if(inherits(ans, "try-error")){
						ans = list()
						ans$values = 1e10
						ans$convergence = 0
						ans$pars = rep(NA, length(xx))
					}
					return( ans )
				})
	} else {
		solution = lapply(as.list(1:n.restarts), FUN = function(i){
					xx = gosolnp_rndpars[i,]
					names(xx) = gosolnp_parnames
					ans = try(solnp(pars = xx, fun = fun, eqfun = eqfun,
									eqB = eqB, ineqfun = ineqfun,
									ineqLB = ineqLB, ineqUB = ineqUB,
									LB = LB, UB = UB,
									control = control, ...), silent = TRUE)
					if(inherits(ans, "try-error")){
						ans = list()
						ans$values = 1e10
						ans$convergence = 0
						ans$pars = rep(NA, length(xx))
					}
					return( ans )
				})
	}
	if(n.restarts>1){
		best = sapply(solution, FUN = function(x) if(x$convergence!=0) NA else x$values[length(x$values)])
		if(all(is.na(best)))
			stop("\ngosolnp-->Could not find a feasible starting point...exiting\n", call. = FALSE)
		nb = which(best == min(best, na.rm = TRUE))[1]
		solution = solution[[nb]]
		if( trace ) cat("\ngosolnp-->Done!\n")
		solution$start.pars = gosolnp_rndpars[nb,]
		names(solution$start.pars) = gosolnp_parnames
		solution$rseed = rseed
	} else{
		solution = solution[[1]]
		solution$start.pars = gosolnp_rndpars[1,]
		names(solution$start.pars) = gosolnp_parnames
		solution$rseed = rseed
	}
	return(solution)
}


#' Generates and returns a set of starting parameters by sampling the parameter
#' space based on the evaluation of the function and constraints.
#'
#' A simple penalty barrier function is formed which is then evaluated at randomly
#' sampled points based on the upper and lower parameter bounds
#' (when \code{eval.type} = 2), else the objective function directly for values not
#' violating any inequality constraints (when \code{eval.type} = 1). The sampled
#' points can be generated from the uniform, normal or truncated normal
#' distributions.
#'
#' @param pars The starting parameter vector. This is not required unless the fixed option is
#' also used.
#' @param fixed The numeric index which indicates those parameters which should stay fixed
#' instead of being randomly generated.
#' @param fun The main function which takes as first argument the parameter vector and returns
#' a single value.
#' @param eqfun (Optional) The equality constraint function returning the vector of evaluated
#' equality constraints.
#' @param eqB (Optional) The equality constraints.
#' @param ineqfun (Optional) The inequality constraint function returning the vector of evaluated
#' inequality constraints.
#' @param ineqLB (Optional) The lower bound of the inequality constraints.
#' @param ineqUB (Optional) The upper bound of the inequality constraints.
#' @param LB The lower bound on the parameters. This is not optional in this function.
#' @param UB The upper bound on the parameters. This is not optional in this function.
#' @param distr A numeric vector of length equal to the number of parameters, indicating the
#' choice of distribution to use for the random parameter generation. Choices are
#' uniform (1), truncated normal (2), and normal (3).
#' @param distr.opt If any choice in \code{distr} was anything other than uniform (1), this is a
#' list equal to the length of the parameters with sub-components for the mean and
#' sd, which are required in the truncated normal and normal distributions.
#' @param bestN The best N (less than or equal to n.sim) set of parameters to return.
#' @param n.sim The number of random parameter sets to generate.
#' @param cluster If you want to make use of parallel functionality, initialize and pass a cluster
#' object from the parallel package (see details), and remember to terminate it!
#' @param rseed (Optional) A seed to initiate the random number generator, else system time will
#' be used.
#' @param eval.type Either 1 (default) for the direction evaluation of the function (excluding
#' inequality constraint violations) or 2 for the penalty barrier method.
#' @param trace (logical) Whether to display the progress of the function evaluation.
#' @param ... (Optional) Additional parameters passed to the main, equality or inequality
#' functions
#' @details
#' Given a set of lower and upper bounds, the function generates, for those
#' parameters not set as fixed, random values from one of the 3 chosen
#' distributions. For simple functions with only inequality constraints, the direct
#' method (\code{eval.type} = 1) might work better. For more complex setups with
#' both equality and inequality constraints the penalty barrier method
#' (\code{eval.type} = 2)might be a better choice.
#' @return A matrix of dimension bestN x (no.parameters + 1). The last column is the
#' evaluated function value.
#' @author Alexios Galanos and Stefan Theussl\cr
#' @note
#' The choice of which distribution to use for randomly sampling the parameter
#' space should be driven by the user's knowledge of the problem and confidence or
#' lack thereof of the parameter distribution. The uniform distribution indicates a
#' lack of confidence in the location or dispersion of the parameter, while the
#' truncated normal indicates a more confident choice in both the location and
#' dispersion. On the other hand, the normal indicates perhaps a lack
#' of knowledge in the upper or lower bounds, but some confidence in the location
#' and dispersion of the parameter. In using choices (2) and (3) for \code{distr},
#' the \code{distr.opt} list must be supplied with \code{mean} and \code{sd} as
#' subcomponents for those parameters not using the uniform.
#' @examples
#' \dontrun{
#' library(Rsolnp)
#' library(parallel)
#' # Windows
#' cl = makePSOCKcluster(2)
#' # Linux:
#' # makeForkCluster(nnodes = getOption("mc.cores", 2L), ...)
#'
#' gofn = function(dat, n)
#' {
#'
#' 	x = dat[1:n]
#' 	y = dat[(n+1):(2*n)]
#' 	z = dat[(2*n+1):(3*n)]
#' 	ii = matrix(1:n, ncol = n, nrow = n, byrow = TRUE)
#' 	jj = matrix(1:n, ncol = n, nrow = n)
#' 	ij = which(ii<jj, arr.ind = TRUE)
#' 	i = ij[,1]
#' 	j = ij[,2]
#' 	#  Coulomb potential
#' 	potential = sum(1.0/sqrt((x[i]-x[j])^2 + (y[i]-y[j])^2 + (z[i]-z[j])^2))
#' 	potential
#' }
#'
#' goeqfn = function(dat, n)
#' {
#' 	x = dat[1:n]
#' 	y = dat[(n+1):(2*n)]
#' 	z = dat[(2*n+1):(3*n)]
#' 	apply(cbind(x^2, y^2, z^2), 1, "sum")
#' }
#' n = 25
#' LB  = rep(-1, 3*n)
#' UB  = rep( 1, 3*n)
#' eqB = rep( 1,   n)
#'
#' sp = startpars(pars = NULL, fixed = NULL, fun = gofn , eqfun = goeqfn,
#' eqB = eqB, ineqfun = NULL, ineqLB = NULL, ineqUB = NULL, LB = LB, UB = UB,
#' distr = rep(1, length(LB)), distr.opt = list(), n.sim = 2000,
#' cluster = cl, rseed = 100, bestN = 15, eval.type = 2, n = 25)
#' #stop cluster
#' stopCluster(cl)
#' # the last column is the value of the evaluated function (here it is the barrier
#' # function since eval.type = 2)
#' print(round(apply(sp, 2, "mean"), 3))
#' # remember to remove the last column
#' ans = solnp(pars=sp[1,-76],fun = gofn , eqfun = goeqfn , eqB = eqB, ineqfun = NULL,
#' ineqLB = NULL, ineqUB = NULL, LB = LB, UB = UB, n = 25)
#' # should get a value of around 243.8162
#' }
#' @export
startpars = function(pars = NULL, fixed = NULL, fun, eqfun = NULL, eqB = NULL,
		ineqfun = NULL, ineqLB = NULL, ineqUB = NULL, LB = NULL, UB = NULL,
		distr = rep(1, length(LB)), distr.opt = list(), n.sim = 20000, cluster = NULL,
		rseed = NULL, bestN = 15, eval.type = 1, trace = FALSE, ...)
{
	if( !is.null(pars) ) gosolnp_parnames = names(pars) else gosolnp_parnames = NULL
	if(is.null(eval.type)) parmethod = 1 else parmethod = as.integer(min(abs(eval.type),2))
	if(parmethod == 0) parmethod = 1
	eval.type = NULL
	#trace = FALSE
	# use a seed to initialize random no. generation
	if(is.null(rseed)) rseed = as.numeric(Sys.time()) else rseed = as.integer(rseed)
	# function requires both upper and lower bounds
	if(is.null(LB))
		stop("\nstartpars-->error: the function requires lower parameter bounds\n", call. = FALSE)
	if(is.null(UB))
		stop("\nstartpars-->error: the function requires upper parameter bounds\n", call. = FALSE)

	# allow for fixed parameters (i.e. non randomly chosen), but require pars vector in that case
	if(!is.null(fixed) && is.null(pars))
		stop("\nstartpars-->error: you need to provide a pars vector if using the fixed option\n", call. = FALSE)
	if(!is.null(pars)) n = length(pars) else n = length(LB)

	np = seq_len(n)

	if(!is.null(fixed)){
		# make unique
		fixed = unique(fixed)
		# check for violations in indices
		if(any(is.na(match(fixed, np))))
			stop("\nstartpars-->error: fixed indices out of bounds\n", call. = FALSE)
	}

	# check distribution options
	# truncated normal
	if(any(distr == 2)){
		d2 = which(distr == 2)
		for(i in 1:length(d2)) {
			if(is.null(distr.opt[[d2[i]]]$mean))
				stop(paste("\nstartpars-->error: distr.opt[[,",d2[i],"]] missing mean\n", sep = ""), call. = FALSE)
			if(is.null(distr.opt[[d2[i]]]$sd))
				stop(paste("\nstartpars-->error: distr.opt[[,",d2[i],"]] missing sd\n", sep = ""), call. = FALSE)
		}
	}
	#  normal
	if(any(distr == 3)){
		d3 = which(distr == 3)
		for(i in 1:length(d3)) {
			if(is.null(distr.opt[[d3[i]]]$mean))
				stop(paste("\nstartpars-->error: distr.opt[[,",d3[i],"]] missing mean\n", sep = ""), call. = FALSE)
			if(is.null(distr.opt[[d3[i]]]$sd))
				stop(paste("\nstartpars-->error: distr.opt[[,",d3[i],"]] missing sd\n", sep = ""), call. = FALSE)
		}
	}

	# setup cluster exports:
	if( !is.null(cluster) ){
		clusterExport(cluster, c("gosolnp_parnames", "fun", "eqfun",
						"eqB", "ineqfun", "ineqLB", "ineqUB", "LB", "UB"), envir = environment())
		if( !is.null(names(list(...))) ){
			# evaluate promises
			xl = names(list(...))
			for(i in 1:length(xl)){
			  eval(parse(text = paste(xl[i], "=list(...)", "[[" , i, "]]", sep = "")))
			}
			clusterExport(cluster, names(list(...)), envir = environment())
		}
		if( !is.null(names(list(...))) ) parallel::clusterExport(cluster, names(list(...)), envir = environment())
		clusterEvalQ(cluster, require(Rsolnp))
	}

	# initiate random search
	gosolnp_rndpars = switch(parmethod,
			.randpars(pars = pars, fixed = fixed, fun = fun, eqfun = eqfun,
					eqB = eqB, ineqfun = ineqfun, ineqLB = ineqLB, ineqUB = ineqUB,
					LB = LB, UB = UB, distr = distr, distr.opt = distr.opt,
					n.restarts = as.integer(bestN), n.sim = n.sim, trace = trace,
					rseed = rseed, gosolnp_parnames = gosolnp_parnames,
					cluster = cluster, ...),
			.randpars2(pars = pars, fixed = fixed, fun = fun, eqfun = eqfun,
					eqB = eqB, ineqfun = ineqfun, ineqLB = ineqLB, ineqUB = ineqUB,
					LB = LB, UB = UB, distr = distr, distr.opt = distr.opt,
					n.restarts = as.integer(bestN), n.sim = n.sim, trace = trace,
					rseed = rseed, gosolnp_parnames = gosolnp_parnames,
					cluster = cluster, ...))
	return(gosolnp_rndpars)
}


.randpars = function(pars, fixed, fun, eqfun, eqB,  ineqfun, ineqLB, ineqUB,
		LB, UB, distr, distr.opt, n.restarts, n.sim, trace = TRUE, rseed,
		gosolnp_parnames, cluster, ...)
{
	if( trace ) cat("\nCalculating Random Initialization Parameters...")
	N = length(LB)
	gosolnp_rndpars = matrix(NA, ncol = N, nrow = n.sim * n.restarts)
	if(!is.null(fixed)) for(i in 1:length(fixed)) gosolnp_rndpars[,fixed[i]] = pars[fixed[i]]
	nf = 1:N
	if(!is.null(fixed)) nf = nf[-c(fixed)]
	m = length(nf)
	set.seed(rseed)
	for(i in 1:m){
		j = nf[i]
		gosolnp_rndpars[,j] = switch(distr[j],
				.distr1(LB[j], UB[j], n.restarts*n.sim),
				.distr2(LB[j], UB[j], n.restarts*n.sim, mean = distr.opt[[j]]$mean, sd = distr.opt[[j]]$sd),
				.distr3(n.restarts*n.sim, mean = distr.opt[[j]]$mean, sd = distr.opt[[j]]$sd)
		)
	}

	if( trace ) cat("ok!\n")

	if(!is.null(ineqfun)){
		if( trace ) cat("\nExcluding Inequality Violations...\n")
		ineqv = matrix(NA, ncol = length(ineqLB), nrow = n.restarts*n.sim)
		# ineqv = t(apply(rndpars, 1, FUN = function(x) ineqfun(x)))
		if(length(ineqLB) == 1){
			ineqv = apply(gosolnp_rndpars, 1, FUN = function(x){
						names(x) = gosolnp_parnames
						ineqfun(x, ...)} )
			lbviol = sum(ineqv<ineqLB)
			ubviol = sum(ineqv>ineqUB)
			if( lbviol > 0 | ubviol > 0 ){
				vidx = c(which(ineqv<ineqLB), which(ineqv>ineqUB))
				vidx = unique(vidx)
				gosolnp_rndpars = gosolnp_rndpars[-c(vidx),,drop=FALSE]
				lvx = length(vidx)
			} else{
				vidx = 0
				lvx = 0
			}
		} else{
			ineqv = t(apply(gosolnp_rndpars, 1, FUN = function(x){
								names(x) = gosolnp_parnames
								ineqfun(x, ...)} ))

			# check lower and upper violations
			lbviol = apply(ineqv, 1, FUN = function(x) sum(any(x<ineqLB)))
			ubviol = apply(ineqv, 1, FUN = function(x) sum(any(x>ineqUB)))
			if( any(lbviol > 0) | any(ubviol > 0) ){
				vidx = c(which(lbviol>0), which(ubviol>0))
				vidx = unique(vidx)
				gosolnp_rndpars = gosolnp_rndpars[-c(vidx),,drop=FALSE]
				lvx = length(vidx)

			} else{
				vidx = 0
				lvx = 0
			}
		}
		if( trace ) cat(paste("\n...Excluded ", lvx, "/",n.restarts*n.sim, " Random Sequences\n", sep = ""))
	}
	# evaluate function value
	if( trace ) cat("\nEvaluating Objective Function with Random Sampled Parameters...")
	if( !is.null(cluster) ){
		nx = dim(gosolnp_rndpars)[1]
		clusterExport(cluster, c("gosolnp_rndpars", ".safefun"), envir = environment())
		evfun = parLapply(cluster, as.list(1:nx), fun = function(i){
					.safefun(gosolnp_rndpars[i, ], fun, gosolnp_parnames, ...)
				})
		evfun = as.numeric( unlist(evfun) )
	} else{
		evfun = apply(gosolnp_rndpars, 1, FUN = function(x) .safefun(x, fun, gosolnp_parnames, ...))
	}
	if( trace ) cat("ok!\n")
	if( trace ) cat("\nSorting and Choosing Best Candidates for starting Solver...")
	z = sort.int(evfun, index.return = T)
	ans = gosolnp_rndpars[z$ix[1:n.restarts],,drop = FALSE]
	prtable = cbind(ans, z$x[1:n.restarts])
	if( trace ) cat("ok!\n")
	colnames(prtable) = c(paste("par", 1:N, sep = ""), "objf")
	if( trace ){
		cat("\nStarting Parameters and Starting Objective Function:\n")
		if(n.restarts == 1) print(t(prtable), digits = 4) else print(prtable, digits = 4)
	}
	return(prtable)
}

# form a barrier function before passing the parameters
.randpars2 = function(pars, fixed, fun, eqfun, eqB,  ineqfun, ineqLB, ineqUB, LB,
		UB, distr, distr.opt, n.restarts, n.sim, rseed, trace = TRUE,
		gosolnp_parnames, cluster, ...)
{
	if( trace ) cat("\nCalculating Random Initialization Parameters...")
	N = length(LB)
	gosolnp_idx = "a"
	gosolnp_R = NULL
	if(!is.null(ineqfun) && is.null(eqfun) ){
		gosolnp_idx = "b"
		gosolnp_R = 100
	}
	if( is.null(ineqfun) && !is.null(eqfun) ){
		gosolnp_idx = "c"
		gosolnp_R = 100
	}
	if(!is.null(ineqfun) && !is.null(eqfun) ){
		gosolnp_idx = "d"
		gosolnp_R = c(100,100)
	}
	gosolnp_rndpars = matrix(NA, ncol = N, nrow = n.sim * n.restarts)
	if(!is.null(fixed)) for(i in 1:length(fixed)) gosolnp_rndpars[,fixed[i]] = pars[fixed[i]]
	nf = 1:N
	if(!is.null(fixed)) nf = nf[-c(fixed)]
	gosolnp_m = length(nf)
	set.seed(rseed)
	for(i in 1:gosolnp_m){
		j = nf[i]
		gosolnp_rndpars[,j] = switch(distr[j],
				.distr1(LB[j], UB[j], n.restarts*n.sim),
				.distr2(LB[j], UB[j], n.restarts*n.sim, mean = distr.opt[[j]]$mean, sd = distr.opt[[j]]$sd),
				.distr3(n.restarts*n.sim, mean = distr.opt[[j]]$mean, sd = distr.opt[[j]]$sd)
		)
	}
	if( trace ) cat("ok!\n")
	# Barrier Function
	pclfn = function(x){
		z=x
		z[x<=0] = 0
		z[x>0] = (0.9+z[x>0])^2
		z
	}
	.lagrfun = function(pars, m, idx, fun, eqfun = NULL, eqB = 0, ineqfun = NULL, ineqLB = NULL, ineqUB = NULL, ...)
	{
		fn = switch(idx,
				"a" = fun(pars[1:m], ...),
				"b" = fun(pars[1:m], ...) + pars[m+1]* sum( pclfn( c(ineqLB - ineqfun(pars[1:m], ...), ineqfun(pars[1:m], ...) - ineqUB) ) ),
				"c" = fun(pars[1:m], ...) + sum( (eqfun(pars[1:m], ...) - eqB )^2 / pars[m+1]),
				"d" = fun(pars[1:m], ...) + sum( (eqfun(pars[1:m], ...) - eqB )^2 / pars[m+1]) + pars[m+2]* sum( pclfn( c(ineqLB - ineqfun(pars[1:m], ...), ineqfun(pars[1:m], ...) - ineqUB) ) ) )
		return(fn)
	}

	# evaluate function value
	if( trace ) cat("\nEvaluating Objective Function with Random Sampled Parameters...")
	if( !is.null(cluster) ){
		nx = dim(gosolnp_rndpars)[1]
		clusterExport(cluster, c("gosolnp_rndpars", "gosolnp_m", "gosolnp_idx",
						"gosolnp_R"), envir = environment())
		clusterExport(cluster, c("pclfn", ".lagrfun"), envir = environment())
		evfun = parallel::parLapply(cluster, as.list(1:nx), fun = function(i){
					.lagrfun(c(gosolnp_rndpars[i,], gosolnp_R), gosolnp_m,
							gosolnp_idx, fun, eqfun, eqB, ineqfun, ineqLB,
							ineqUB, ...)
				})
		evfun = as.numeric( unlist(evfun) )
	} else{
		evfun = apply(gosolnp_rndpars, 1, FUN = function(x){
					.lagrfun(c(x,gosolnp_R), gosolnp_m, gosolnp_idx, fun, eqfun,
							eqB, ineqfun, ineqLB, ineqUB, ...)})
	}
	if( trace ) cat("ok!\n")
	if( trace ) cat("\nSorting and Choosing Best Candidates for starting Solver...")
	z = sort.int(evfun, index.return = T)
	#distmat = dist(evfun, method = "euclidean", diag = FALSE, upper = FALSE, p = 2)
	ans = gosolnp_rndpars[z$ix[1:n.restarts],,drop = FALSE]
	prtable = cbind(ans, z$x[1:n.restarts])
	colnames(prtable) = c(paste("par", 1:N, sep = ""), "objf")
	if( trace ){
		cat("\nStarting Parameters and Starting Objective Function:\n")
		if(n.restarts == 1) print(t(prtable), digits = 4) else print(prtable, digits = 4)
	}
	return(prtable)
}


.distr1 = function(LB, UB, n)
{
	runif(n, min = LB, max = UB)
}

.distr2 = function(LB, UB, n, mean, sd)
{
	rtruncnorm(n, a = as.double(LB), b = as.double(UB), mean = as.double(mean), sd = as.double(sd))
}

.distr3 = function(n, mean, sd)
{
	rnorm(n, mean = mean, sd = sd)
}

.safefun = function(pars, fun, gosolnp_parnames, ...){
	# gosolnp_parnames = get("gosolnp_parnames", envir = .env)
	names(pars) = gosolnp_parnames
	v  = fun(pars, ...)
	if(is.na(v) | !is.finite(v) | is.nan(v)) {
		warning(paste("\ngosolnp-->warning: ", v , " detected in function call...check your function\n", sep = ""), immediate. = FALSE)
		v = 1e24
	}
	v
}

benchmarkids <- function()
{
return(c("Powell", "Wright4", "Wright9", "Alkylation", "Entropy", "Box", "RosenSuzuki",
				"Electron", "Permutation"))
}


benchmark <- function( id = "Powell")
{
  if( !any(benchmarkids() == id[ 1L ]) )
    stop( "invalid benchmark id" )
 	ans = switch(id,
			Powell = .powell(),
			Wright4 = .wright4(),
			Wright9 = .wright9(),
			Alkylation = .alkylation(),
			Entropy = .entropy(),
			Box = .box(),
			RosenSuzuki = .rosensuzuki(),
			Electron = .electron(),
			Permutation = .permutation())
	return(ans)
}

.powell = function()
{
	.fn1 = function(x)
	{
		exp(x[1]*x[2]*x[3]*x[4]*x[5])
	}

	.eqn1 = function(x){
		z1=x[1]*x[1]+x[2]*x[2]+x[3]*x[3]+x[4]*x[4]+x[5]*x[5]
		z2=x[2]*x[3]-5*x[4]*x[5]
		z3=x[1]*x[1]*x[1]+x[2]*x[2]*x[2]
		return(c(z1,z2,z3))
	}

	.x0 = c(-2, 2, 2, -1, -1)
	ctrl=list(trace=0)
	ans = solnp(.x0, fun = .fn1, eqfun = .eqn1, eqB = c(10,0,-1), control=ctrl)
	minos = list()
	minos$fn = 0.05394985
	minos$pars = c(-1.717144, 1.595710, 1.827245, 0.763643, 0.763643)
	minos$nfun = 524
	minos$iter = 12
	minos$elapsed = 0.2184

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("Powell's exponential problem is a function of five variables with
three nonlinear equality constraints on the variables.")

	return(bt)
}

.wright4 = function()
{
	.fn1 = function(x)
	{
		(x[1]-1)^2+(x[1]-x[2])^2+(x[2]-x[3])^3+(x[3]-x[4])^4+(x[4]-x[5])^4
	}

	.eqn1 = function(x){
		z1=x[1]+x[2]*x[2]+x[3]*x[3]*x[3]
		z2=x[2]-x[3]*x[3]+x[4]
		z3=x[1]*x[5]
		return(c(z1,z2,z3))
	}

	.x0 = c(1, 1, 1, 1, 1)
	ctrl=list(trace=0)
	ans = solnp(.x0, fun = .fn1, eqfun = .eqn1, eqB = c(2+3*sqrt(2),-2+2*sqrt(2),2), control=ctrl)
	minos = list()
	minos$fn = 0.02931083
	minos$pars = c(1.116635, 1.220442, 1.537785, 1.972769, 1.791096)
	minos$nfun = 560
	minos$iter = 9
	minos$elapsed = 0.249

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("Wright's fourth problem is a function of five variables with
three non linear equality constraints on the variables. This popular
test problem has several local solutions and taken from Wright (1976).")
	return(bt)
}

.wright9 = function()
{
	.fn1 = function(x)
	{
		10*x[1]*x[4]-6*x[3]*x[2]*x[2]+x[2]*(x[1]*x[1]*x[1])+
				9*sin(x[5]-x[3])+x[5]^4*x[4]*x[4]*x[2]*x[2]*x[2]
	}

	.ineqn1 = function(x){
		z1=x[1]*x[1]+x[2]*x[2]+x[3]*x[3]+x[4]*x[4]+x[5]*x[5]
		z2=x[1]*x[1]*x[3]-x[4]*x[5]
		z3=x[2]*x[2]*x[4]+10*x[1]*x[5]
		return(c(z1,z2,z3))
	}
	ineqLB = c(-100, -2, 5)
	ineqUB = c(20, 100, 100)
	.x0 = c(1, 1, 1, 1, 1)
	ctrl=list(trace=0)
	ans = solnp(.x0, fun = .fn1, ineqfun = .ineqn1, ineqLB = ineqLB, ineqUB = ineqUB, control=ctrl)
	minos = list()
	minos$fn = -210.4078
	minos$pars = c(-0.08145219, 3.69237756, 2.48741102,  0.37713392, 0.17398257)
	minos$nfun = 794
	minos$iter = 11
	minos$elapsed = 0.281

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("Wright's ninth problem is a function of five variables with three
non linear inequality constraints on the variables. This popular test
problem has several local solutions and taken from Wright (1976).")
	return(bt)
}

.alkylation = function()
{
	.fn1 = function(x)
	{
		-0.63*x[4]*x[7]+50.4*x[1]+3.5*x[2]+x[3]+33.6*x[5]
	}

	.eqn1 = function(x){
		z1=98*x[3]-0.1*x[4]*x[6]*x[9]-x[3]*x[6]
		z2=1000*x[2]+100*x[5]-100*x[1]*x[8]
		z3=122*x[4]-100*x[1]-100*x[5]
		return(c(z1,z2,z3))
	}
	.ineqn1 = function(x){
		z1=(1.12*x[1]+0.13167*x[1]*x[8]-0.00667*x[1]*x[8]*x[8])/x[4]
		z2=(1.098*x[8]-0.038*x[8]*x[8]+0.325*x[6]+57.25)/x[7]
		z3=(-0.222*x[10]+35.82)/x[9]
		z4=(3*x[7]-133)/x[10]
		return(c(z1,z2,z3,z4))
	}
	ineqLB = c(0.99,0.99,0.9,0.99)
	ineqUB = c(100/99,100/99,10/9,100/99)
	eqB = c(0,0,0)
	LB = c(0,0,0,10,0,85,10,3,1,145)
	UB = c(20,16,120,50,20,93,95,12,4,162)
	.x0 = c(17.45,12,110,30,19.74,89.2,92.8,8,3.6,155)
	ctrl = list(rho = 0, trace=0)
	ans = solnp(.x0, fun = .fn1, eqfun = .eqn1, eqB = eqB, ineqfun = .ineqn1, ineqLB = ineqLB,
			ineqUB = ineqUB, LB = LB, UB = UB, control = ctrl)
	minos = list()
	minos$fn = -172.642
	minos$pars = c(16.996427, 16.000000, 57.685751, 30.324940, 20.000000, 90.565147, 95.000000, 10.590461, 1.561636, 153.535354)
	minos$nfun = 2587
	minos$iter = 13
	minos$elapsed = 0.811

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("The Alkylation problem models a simplified alkylation process. It
is a function of ten variables with four non linear inequality and
three non linear equality constraints as well as variable bounds.
The problem is taken from Locke and Westerberg (1980).")
	return(bt)
}

.entropy = function()
{
	.fn1 = function(x)
	{
		m = length(x)
		f = 0
		for(i in 1:m){
			f = f-log(x[i])
		}
		ans = f-log(.vnorm(x-1) + 0.1)
		ans
	}

	.eqn1 = function(x){
		sum(x)
	}
	eqB = 10
	LB = rep(0,10)
	UB = rep(1000,10)
	.x0 = runif(10, 0, 1000)
	ctrl=list(trace=0)
	ans = solnp(.x0, fun = .fn1, eqfun = .eqn1, eqB = eqB, LB = LB, UB = UB, control=ctrl)
	minos = list()
	minos$fn = 0.1854782
	minos$pars = c(2.2801555, 0.8577605, 0.8577605, 0.8577605, 0.8577605, 0.8577605,
			0.8577605, 0.8577605, 0.8577605, 0.8577605)
	minos$nfun = 886
	minos$iter = 4
	minos$elapsed = 0.296

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("The Entropy problem is non convex in n variables with one linear
equality constraint and variable positivity bounds.")
	return(bt)
}

.box = function()
{
	.fn1 = function(x)
	{
		-x[1]*x[2]*x[3]
	}

	.eqn1 = function(x){
		4*x[1]*x[2]+2*x[2]*x[3]+2*x[3]*x[1]
	}

	eqB = 100
	LB = rep(1, 3)
	UB = rep(10, 3)

	.x0 = c(1.1, 1.1, 9)
	ctrl=list(trace=0)
	ans = solnp(.x0, fun = .fn1, eqfun = .eqn1, eqB = eqB, LB = LB, UB = UB, control=ctrl)
	minos = list()
	minos$fn = -48.11252
	minos$pars = c(2.886751, 2.886751, 5.773503)
	minos$nfun = 394
	minos$iter = 9
	minos$elapsed = 0.156

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("The box problem is a function of three variables with one non
linear equality constraint and variable bounds.")
	return(bt)
}

.rosensuzuki = function()
{
	.fn1 = function(x)
	{
		x[1]*x[1]+x[2]*x[2]+2*x[3]*x[3]+x[4]*x[4]-5*x[1]-5*x[2]-21*x[3]+7*x[4]
	}

	.ineqn1 = function(x){
		z1=8-x[1]*x[1]-x[2]*x[2]-x[3]*x[3]-x[4]*x[4]-x[1]+x[2]-x[3]+x[4]
		z2=10-x[1]*x[1]-2*x[2]*x[2]-x[3]*x[3]-2*x[4]*x[4]+x[1]+x[4]
		z3=5-2*x[1]*x[1]-x[2]*x[2]-x[3]*x[3]-2*x[1]+x[2]+x[4]
		return(c(z1,z2,z3))
	}
	ineqLB = rep(0, 3)
	ineqUB = rep(1000, 3)
	.x0 = c(1, 1, 1, 1)
	ctrl=list(trace=0)
	ans = solnp(.x0, fun = .fn1, ineqfun = .ineqn1, ineqLB = ineqLB, ineqUB = ineqUB, control=ctrl)
	minos = list()
	minos$fn = -44
	minos$pars = c(2.502771e-07, 9.999997e-01, 2.000000e+00, -1.000000e+00)
	minos$nfun = 527
	minos$iter = 12
	minos$elapsed = 0.203

	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			minos =  rbind(round(minos$fn, 5L),
					round(minos$iter, 0L),
					round(0, 0L),
					round(minos$nfun, 0L),
					round(minos$elapsed, 3L),
					matrix(round(minos$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	attr(bt, "description") = paste("The Rosen-Suzuki problem is a function of four variables with
three nonlinear inequality constraints on the variables. It is taken
from Problem 43 of Hock and Schittkowski (1981).")
	return(bt)
}



#----------------------------------------------------------------------------------
# Some Problems in Global Optimization
#----------------------------------------------------------------------------------


# Distribution of Electrons on a Sphere
# Given n electrons, find the equilibrium state distribution (of minimal Coulomb potential)
# of the electrons positioned on a conducting sphere. This model is from the COPS benchmarking suite.
# See http://www-unix.mcs.anl.gov/~more/cops/.

.electron = function()
{
	gofn = function(dat, n)
	{

		x = dat[1:n]
		y = dat[(n+1):(2*n)]
		z = dat[(2*n+1):(3*n)]
		ii = matrix(1:n, ncol = n, nrow = n, byrow = TRUE)
		jj = matrix(1:n, ncol = n, nrow = n)
		ij = which(ii<jj, arr.ind = TRUE)
		i = ij[,1]
		j = ij[,2]
		#  Coulomb potential
		potential = sum(1.0/sqrt((x[i]-x[j])^2 + (y[i]-y[j])^2 + (z[i]-z[j])^2))
		potential
	}

	goeqfn = function(dat, n)
	{
		x = dat[1:n]
		y = dat[(n+1):(2*n)]
		z = dat[(2*n+1):(3*n)]
		apply(cbind(x^2, y^2, z^2), 1, "sum")
	}

	n = 25
	LB = rep(-1, 3*n)
	UB = rep(1, 3*n)
	eqB = rep(1, n)
	ans = gosolnp(pars  = NULL, fixed = NULL, fun = gofn, eqfun = goeqfn, eqB = eqB, LB = LB, UB = UB,
			control = list(), distr = rep(1, length(LB)), distr.opt = list(outer.iter = 10, trace = 1),
			n.restarts = 2, n.sim = 20000, rseed = 443, n = 25)

	conopt = list()
	conopt$fn  = 243.813
	conopt$iter = 33
	conopt$nfun = NA
	conopt$elapsed = 0.041
	conopt$pars = c(-0.0117133872042326,	0.627138691757704,	-0.471025867741051,	-0.164419761338935,	-0.0315460712487934,
			-0.12718981058582,	-0.540049624346613,	0.600346770449059,	0.29796281847713,	-0.740960572770077,	0.972512148478245,
			-0.870858895858346,	0.84178885636396,	-0.182471994739506,	0.603293664844919,	0.0834172554171806,	0.51317309921937,
			0.260639996237799,	-0.0972877803105543,	-0.979882559381314,	-0.64809471648373,	-0.722351411610064,	0.847184430059647,
			0.514683899757428,	-0.574607272207711, 0.114609211815613,	-0.748168886860133,	-0.379763612890494,	0.743271243797936,
			0.846784034469756,	0.220425955966718,	0.839147591392778,	-0.613810163641104,	-0.499794531840362,	-0.199680552951248,
			-0.105141937435843,	-0.434753057357539,	-0.127562956191463,	-0.895691740627038,	0.574257349984438,	-0.967631920158332,
			-0.0243647398873149,	0.959445715727407,	-0.517406241891138,	0.197677956191858,	0.503867654081605,	0.286619450971711,
			0.522509289901788,	0.474911361600545,	-0.768915699421274, -0.993341595387613,	-0.21665728244143,	-0.796187308516752,	-0.648470508368975,
			0.531000606737778,	0.967075565826839,	-0.0646353084836963,	0.512660548728168,	-0.813279524362711,	0.641174786133487,
			0.207762590603952,	-0.229335044471304,	-0.524518077390237,	-0.405512363446903,	0.55341236880543,	0.23818066376044,
			0.857939234263016,	-0.107381147942665,	0.850191665834437,	0.0274516931378701,	0.571043453369507,	-0.629331175510656,
			-0.0962423162171412,	-0.713834491989006,	0.280348229724225)
	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			conopt =  rbind(round(conopt$fn, 5L),
					round(conopt$iter, 0L),
					round(0, 0L),
					round(conopt$nfun, 0L),
					round(conopt$elapsed, 3L),
					matrix(round(conopt$pars, 5L), ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	colnames(bt) = c("solnp", "conopt")
	attr(bt, "description") = paste("The equilibrium state distribution (of minimal Coulomb potential)\n of the electrons positioned on a conducting sphere.")
	return(bt)
}

# Permutation Problem -- Unique Solution f(x) = 0 and x(i) = i

.permutation = function()
{
	.perm = function(x, n, b){
		F = 0
		for(k in 1:n){
			S = 0
			for(i in 1:n){
				S = S + ( ( (i^k) + b ) * (( x[i]/i )^k -1))
			}
			F = F + S^2
		}
		F
	}

	ans = gosolnp(pars  = NULL, fixed = NULL, fun = .perm, eqfun = NULL, eqB = NULL, LB = rep(-4, 4), UB = rep(4, 4),
			control = list(outer.iter = 25, trace = 1, tol = 1e-9), distr = rep(1, 4), distr.opt = list(),
			n.restarts = 6, n.sim = 20000, rseed = 99, n = 4, b =0.5)


	bt = data.frame( solnp = rbind(round(ans$values[length(ans$values)], 5L),
					round(ans$outer.iter, 0L),
					round(ans$convergence, 0L),
					round(ans$nfuneval, 0L),
					round(ans$elapsed, 3L),
					matrix(round(ans$pars, 5L), ncol = 1L)),
			actual =  rbind(0,
					NA,
					round(0, 0L),
					NA,
					NA,
					matrix(1:4, ncol = 1L)) )
	rownames(bt) <- c("funcValue", "majorIter", "exitFlag", "nfunEval", "time(sec)",
			paste("par.", 1L:length(ans$pars), sep = "") )
	colnames(bt) = c("solnp", "expected")
	attr(bt, "description") = paste("Permutation Problem PERM(4,0.5).")

}

#' Retrieve Implemented Test Problems for the SOLNP Suite
#'
#' Returns a list (or a single object) of implemented test problems corresponding to a selected suite.
#' Problem functions must follow the naming convention \sQuote{problem_name_problem} and return a list
#' describing the optimization problem (e.g., objective, constraints, bounds).
#'
#' @param suite Character. The test suite to draw from. Must be one of \dQuote{Hock-Schittkowski} or \dQuote{Other}.
#'              Default is \dQuote{Hock-Schittkowski}.
#' @param number Integer or vector of integers. One or more problem numbers to retrieve. Ignored if return_all = TRUE.
#' @param return_all Logical. If TRUE, returns all implemented problems in the specified suite. Default is FALSE.
#'
#' @return If one problem is requested and implemented, the evaluated problem object is returned directly.
#'         Otherwise, an unnamed list of evaluated problem objects is returned.
#'
#' @details
#' - Problems are matched by number within the selected suite, using the table from [solnp_problems_table()].
#' - If a requested problem is valid but not yet implemented (i.e., the corresponding function does not exist),
#'   a message will inform the user.
#' - If a problem number exceeds the allowable range (e.g., > 306 for Hock-Schittkowski), an error is raised.
#'
#' @examples
#' \dontrun{
#' # Retrieve a single HS problem
#' prob <- solnp_problem_suite(number = 1)
#'
#' # Retrieve multiple HS problems
#' probs <- solnp_problem_suite(number = c(1, 2, 3))
#'
#' # Retrieve problem in "Other" suite
#' other_prob <- solnp_problem_suite(suite = "Other", number = 1)
#' }
#'
#' @seealso [solnp_problems_table()]
#' @export
solnp_problem_suite <- function(suite = "Hock-Schittkowski", number = 1, return_all = FALSE)
{
    suite <- match.arg(suite[1], c("Hock-Schittkowski","Other"))
    v_problems <- solnp_problems_table()
    v_problems <- v_problems[v_problems$Suite == suite,]
    if (!return_all) {
        number <- unique(as.integer(number))
        if (any(number < 1)) stop("Problem number must be strictly positive.")
        if (suite == "Hock-Schittkowski" & any(number > 306)) stop("Some requested problem numbers do not exist (maximum is 306 for Hock-Schittkowski).")
        known <- v_problems$Number
        not_implemented <- setdiff(number, known)
        implemented <- intersect(number, known)

        if (length(not_implemented) > 0) {
            message("The following problems are not yet implemented: ",
                    paste(not_implemented, collapse = ", "))
        }
        funs <- lapply(implemented, function(n) {
            fname <- paste0(v_problems$Problem[v_problems$Number == n], "_problem")
            if (!exists(fname, mode = "function", inherits = TRUE)) {
                warning(sprintf("Function '%s' not found in the environment.", fname))
                return(NULL)
            }
            get(fname, mode = "function", inherits = TRUE)()
        })
        funs <- funs[!sapply(funs, is.null)]
        if (length(funs) == 1) {
            return(funs[[1]])
        } else {
            return(unname(funs))
        }
    } else {
        # return_all = TRUE: Return all implemented hsXX_problem functions
        all_funs <- lapply(seq_len(nrow(v_problems)), function(i) {
            fname <- paste0(v_problems$Problem[i], "_problem")
            if (exists(fname, mode = "function", inherits = TRUE)) {
                get(fname, mode = "function", inherits = TRUE)()
            } else {
                NULL
            }
        })
        all_funs <- all_funs[!sapply(all_funs, is.null)]
        return(unname(all_funs))
    }
}

#' List of Valid Test Problems for the SOLNP Suite
#'
#' Returns a data.frame of known and registered test problems used with the SOLNP solver.
#' The list includes problems from the Hock-Schittkowski suite as well as a selection
#' of other classic optimization problems.
#'
#' @return A data.frame with the following columns:
#' \describe{
#'   \item{Suite}{A character string indicating the suite the problem belongs to.
#'                One of \dQuote{Hock-Schittkowski} or \dQuote{Other}.}
#'   \item{Problem}{The base name of the problem function (without the \sQuote{_problem} suffix).}
#'   \item{Number}{An integer identifier used to index or request problems programmatically.}
#' }
#'
#' @details
#' - All problem functions are expected to follow the naming convention
#'   \sQuote{Problem_problem} (e.g., \sQuote{hs01_problem}).
#' - For Hock-Schittkowski problems, numbers range from 1 to 64, with a few
#'   selected extras (e.g., 110, 118, 119).
#' - The \dQuote{Other} suite includes named problems like \sQuote{box},
#'   \sQuote{alkylation}, \sQuote{entropy}, \sQuote{garch}, etc.,
#' and are numbered sequentially.
#'
#' @examples
#' # View all known problems
#' tail(solnp_problems_table())
#'
#' # Filter only HS problems
#' head(subset(solnp_problems_table(), Suite == "Hock-Schittkowski"))
#'
#' @seealso [solnp_problem_suite()]
#' @export
solnp_problems_table <- function()
{
    hs_problems <- c("hs01","hs02","hs03","hs04","hs05","hs06","hs07",
                     "hs08","hs09", paste0("hs",10:64),"hs110","hs118","hs119")
    hs_number <- c(1:64,110,118,119)
    other_problems <- sort(c("alkylation","wright4","wright9","box","entropy","rosen_suzuki","powell",
                             "himmelblau5","garch"))
    other_number <- c(1:length(other_problems))
    rbind(data.frame("Suite" = "Hock-Schittkowski", "Problem" = hs_problems, "Number" = hs_number),
          data.frame("Suite" = "Other", "Problem" = other_problems, "Number" = other_number))
}

#' Standardize an Optimization Problem to NLP Standard Form
#'
#' Converts a problem specified with two-sided inequalities and nonzero equality right-hand sides
#' to the standard nonlinear programming (NLP) form.
#'
#'
#' @param prob A list specifying the problem in SOLNP-compatible format, with components
#'   \code{fn}, \code{eq_fn}, \code{eq_jac}, \code{eq_b}, \code{ineq_fn}, \code{ineq_jac},
#'   \code{ineq_lower}, \code{ineq_upper}, and others.
#'
#' @return A list with the same structure as the input, but with \code{eq_fn} and \code{ineq_fn}
#' standardized to the forms \eqn{e(x) = 0} and \eqn{g(x) \leq 0}, and with
#' \code{eq_b}, \code{ineq_lower}, and \code{ineq_upper} removed.
#'
#' @details
#' The standard form given by the following set of equations:
#'
#' \deqn{ \min_x\ f(x) }
#' \deqn{ \textrm{subject to}\quad e(x) = 0 }
#' \deqn{ \qquad\qquad\qquad g(x) \leq 0 }
#'
#' Specifically:
#' \itemize{
#'   \item All equality constraints are standardized to \eqn{e(x) = e(x) - b = 0}
#'   \item Each two-sided inequality \eqn{l \leq g(x) \leq u} is converted to one or two
#'   one-sided constraints: \eqn{l - g(x) \leq 0}, \eqn{g(x) - u \leq 0}
#' }
#'
#' The returned problem object has all equalities as \eqn{e(x) = 0}, all inequalities as \eqn{g(x) \leq 0},
#' and any right-hand side or bounds are absorbed into the standardized constraint functions.
#'
#' @examples
#' # Alkylation problem
#' p <- solnp_problem_suite(suite = "Other", number = 1)
#' ps <- solnp_standardize_problem(p)
#' ps$eq_fn(ps$start)    # standardized equalities: e(x) = 0
#' ps$ineq_fn(ps$start)  # standardized inequalities: g(x) <= 0
#'
#' @seealso \code{\link{solnp_problem_suite}}
#'
#' @export
solnp_standardize_problem <- function(prob) {
    # Objective and gradient
    fn <- prob$fn
    gr <- prob$gr

    # Equalities
    eq_fn <- prob$eq_fn
    eq_jac <- prob$eq_jac
    eq_b <- if (!is.null(prob$eq_b)) prob$eq_b else NULL

    # Standardize equalities: e(x) = eq_fn(x) - eq_b = 0
    new_eq_fn <- if (is.null(eq_fn)) {
        NULL
    } else if (is.null(eq_b) || all(eq_b == 0)) {
        eq_fn
    } else {
        function(x) eq_fn(x) - eq_b
    }

    new_eq_jac <- eq_jac  # No transformation needed if analytic jacobian is for eq_fn

    # Inequalities
    ineq_fn <- prob$ineq_fn
    ineq_jac <- prob$ineq_jac
    ineq_lower <- prob$ineq_lower
    ineq_upper <- prob$ineq_upper

    # Standardize inequalities to g(x) <= 0 form
    new_ineq_fn <- if (is.null(ineq_fn)) {
        NULL
    } else {
        function(x) {
            gx <- ineq_fn(x)
            g <- c()
            if (!is.null(ineq_lower)) {
                g <- c(g, ineq_lower - gx)
            }
            if (!is.null(ineq_upper)) {
                g <- c(g, gx - ineq_upper)
            }
            if (length(g) == 0) NULL else g
        }
    }

    # Standardize inequality jacobian if supplied
    new_ineq_jac <- if (is.null(ineq_jac)) {
        NULL
    } else {
        function(x) {
            J <- ineq_jac(x)
            Jlist <- list()
            if (!is.null(ineq_lower)) {
                for (i in seq_along(ineq_lower)) {
                    Jlist[[length(Jlist) + 1]] <- -J[i, , drop = FALSE]
                }
            }
            if (!is.null(ineq_upper)) {
                for (i in seq_along(ineq_upper)) {
                    Jlist[[length(Jlist) + 1]] <- J[i, , drop = FALSE]
                }
            }
            if (length(Jlist) == 0) NULL else do.call(rbind, Jlist)
        }
    }

    # Return standardized problem
    prob_std <- prob
    prob_std$eq_fn <- new_eq_fn
    prob_std$eq_jac <- new_eq_jac
    prob_std$ineq_fn <- new_ineq_fn
    prob_std$ineq_jac <- new_ineq_jac
    prob_std$eq_b <- NULL      # standardized out
    prob_std$ineq_lower <- NULL
    prob_std$ineq_upper <- NULL
    prob_std
}

#################################################################################
##
##   R package Rsolnp by Alexios Galanos and Stefan Theussl Copyright (C) 2009-2013
##   This file is part of the R package Rsolnp.
##
##   The R package Rsolnp is free software: you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation, either version 3 of the License, or
##   (at your option) any later version.
##
##   The R package Rsolnp is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.
##
#################################################################################

# Based on the original subnp Yinyu Ye
# http://www.stanford.edu/~yyye/Col.html
.subnp = function(pars, yy, ob, hessv, lambda, vscale, ctrl, .env, ...)
{
	.solnp_fun = get(".solnp_fun", envir = .env)
	.solnp_eqfun = get(".solnp_eqfun", envir = .env)
	.solnp_ineqfun = get(".solnp_ineqfun", envir = .env)
	ineqLB = get(".ineqLB", envir = .env)
	ineqUB = get(".ineqUB", envir = .env)
	LB = get(".LB", envir = .env)
	UB = get(".UB", envir = .env)
	#.solnp_gradfun = get(".solnp_gradfun", envir = .env)
	#.solnp_eqjac = get(".solnp_eqjac", envir = .env)
	#.solnp_ineqjac = get(".solnp_ineqjac", envir = .env)
	ind = get("ind", envir = .env)

	# pars [nineq + np]
	rho   = ctrl[ 1 ]
	maxit = ctrl[ 2 ]
	delta = ctrl[ 3 ]
	tol   = ctrl[ 4 ]
	trace = ctrl[ 5 ]
	# [1] length of pars
	# [2] has function gradient?
	# [3] has hessian?
	# [4] has ineq?
	# [5] ineq length
	# [6] has jacobian (inequality)
	# [7] has eq?
	# [8] eq length
	# [9] has jacobian (equality)
	# [10] has upper / lower bounds
	# [11] has either lower/upper bounds or ineq


	neq   = ind[ 8 ]
	nineq = ind[ 5 ]
	np    = ind[ 1 ]
	ch    = 1
	alp   = c(0,0,0)
	nc    = neq + nineq
	npic  = np + nineq
	p0    = pars

	# pb [ 2 x (nineq + np) ]
	pb    = rbind( cbind(ineqLB, ineqUB), cbind(LB,UB) )
	sob   = numeric()
	ptt   = matrix()
	sc    = numeric()

	# scale the cost, the equality constraints, the inequality constraints,
	# the parameters (inequality parameters AND actual parameters),
	# and the parameter bounds if there are any
	# Also make sure the parameters are no larger than (1-tol) times their bounds
	# ob [ 1 neq nineq]

	ob = ob / vscale[ 1:(nc + 1) ]
	# p0 [np]
	p0 = p0 / vscale[ (neq + 2):(nc + np + 1) ]
	if( ind[ 11 ] ) {

		if( !ind[ 10 ] ) {
			mm = nineq
		} else {
			mm = npic
		}

		pb = pb / cbind(vscale[ (neq + 2):(neq + mm + 1) ], vscale[ (neq + 2):(neq + mm + 1) ])
	}

	# scale the lagrange multipliers and the Hessian

	if( nc > 0) {
		# yy [total constraints = nineq + neq]
		# scale here is [tc] and dot multiplied by yy
		yy = vscale[ 2:(nc + 1) ] * yy / vscale[ 1 ]
	}
	# yy = [zeros 3x1]

	# h is [ (np+nineq) x (np+nineq) ]
	#columnvector %*% row vector (size h) then dotproduct h then dotdivide scale[1]
	hessv = hessv * (vscale[ (neq + 2):(nc + np + 1) ] %*% t(vscale[ (neq + 2):(nc + np + 1)]) ) / vscale[ 1 ]
	# h[ 8x8 eye]
	j = ob[ 1 ]

	if( ind[4] ) {

		if( !ind[7] ) {
			# [nineq x (nineq+np) ]
			a = cbind( -diag(nineq), matrix(0, ncol = np, nrow = nineq) )
		} else {
			# [ (neq+nineq) x (nineq+np)]
			a = rbind( cbind( 0 * .ones(neq, nineq), matrix(0, ncol = np, nrow = neq) ),
					cbind( -diag(nineq), matrix(0, ncol = np, nrow = nineq) ) )
		}

	}
	if( ind[7] && !ind[4] ) {
		a = .zeros(neq, np)
	}

	if( !ind[7] && !ind[4] ) {
		a = .zeros(1, np)
	}

	# gradient
	g= 0 * .ones(npic, 1)
	p = p0 [ 1:npic ]
	if( nc > 0 ) {
		# [ nc ]
		constraint = ob[ 2:(nc + 1) ]
		# constraint [5 0 11 3x1]
		# gradient routine
		for( i in 1:np ) {
			# scale the parameters (non ineq)
			p0[ nineq + i ] = p0[ nineq + i ] + delta
			tmpv = p0[ (nineq + 1):npic ] * vscale[ (nc + 2):(nc + np + 1) ]
			funv 	= .safefunx(tmpv, .solnp_fun, .env, ...)
			eqv 	= .solnp_eqfun(tmpv, ...)
			ineqv 	= .solnp_ineqfun(tmpv, ...)
			ctmp = get(".solnp_nfn", envir =  .env)
			assign(".solnp_nfn", ctmp + 1, envir = .env)

			ob = c(funv, eqv, ineqv) / vscale[ 1:(nc + 1) ]
			g[ nineq + i ]   = (ob[ 1 ] - j) / delta
			a[ , nineq + i ] = (ob[ 2:(nc + 1) ] - constraint) / delta
			p0[ nineq + i ]  = p0[ nineq + i ] - delta
		}

		if( ind[4] ) {
			constraint[ (neq + 1):(neq + nineq) ] = constraint[ (neq + 1):(neq + nineq) ] - p0[ 1:nineq ]
		}

		# solver messages
		if( .solvecond(a) > 1 / .eps ) {
			if( trace ) .subnpmsg( "m1" )
		}

		# a(matrix) x columnvector - columnvector
		# b [nc,1]
		b  = a %*% p0 - constraint
		ch = -1
		alp[ 1 ] = tol - max( abs(constraint) )

		if( alp[ 1 ] <= 0 ) {
			ch = 1

			if( !ind[11] ) {
				# a %*% t(a) gives [nc x nc]
				# t(a) %*% above gives [(np+nc) x 1]
				p0 = p0 - t(a) %*% solve(a %*% t(a), constraint)
				alp[ 1 ] = 1
			}

		}

		if( alp[ 1 ] <= 0 ) {
			# this expands p0 to [nc+np+1]
			p0[ npic + 1 ] = 1
			a  = cbind(a, -constraint)
			# cx is rowvector
			cx = cbind(.zeros(1,npic), 1)
			dx = .ones(npic + 1, 1)
			go = 1
			minit = 0

			while( go >= tol ) {
				minit = minit + 1
				# gap [(nc + np) x 2]
				gap = cbind(p0[ 1:mm ] - pb[ , 1 ], pb[ , 2 ] - p0[ 1:mm ] )
				# this sorts every row
				gap = t( apply( gap, 1, FUN=function( x ) sort(x) ) )
				dx[ 1:mm ] = gap[ , 1 ]
				# expand dx by 1
				dx[ npic + 1 ] = p0[ npic + 1 ]

				if( !ind[10] ) {
					dx[ (mm + 1):npic ] = max( c(dx[ 1:mm ], 100) ) * .ones(npic - mm, 1)
				}
				# t( a %*% diag( as.numeric(dx) ) ) gives [(np+nc + 1 (or more) x nc]
				# dx * t(cx) dot product of columnvectors
				# qr.solve returns [nc x 1]

				# TODO: Catch errors here
				y = try( qr.solve( t( a %*% diag( as.numeric(dx) , length(dx), length(dx) ) ), dx * t(cx) ), silent = TRUE)
				if(inherits(y, "try-error")){
					p = p0 * vscale[ (neq + 2):(nc + np + 1) ]  # unscale the parameter vector
					if( nc > 0 ) {
						y = 0 # unscale the lagrange multipliers
					}
					hessv = vscale[ 1 ] * hessv / (vscale[ (neq + 2):(nc + np + 1) ] %*% t(vscale[ (neq + 2):(nc + np + 1) ]) )
					ans = list(p = p, y = y, hessv = hessv, lambda = lambda)
					assign(".solnp_errors", 1, envir = .env)
					return(ans)
				}
				v = dx * ( dx *(t(cx) - t(a) %*% y) )

				if( v[ npic + 1 ] > 0 ) {
					z = p0[ npic + 1 ] / v[ npic + 1 ]

					for( i in 1:mm ) {

						if( v[ i ] < 0 ) {
							z = min(z, -(pb[ i, 2 ] - p0[ i ]) / v[ i ])
						} else if( v[ i ] > 0 ) {
							z = min( z, (p0[ i ] - pb[ i , 1 ]) / v[ i ])
						}
					}

					if( z >= p0[ npic + 1 ] / v[ npic + 1 ] ) {
						p0 = p0 - z * v
					} else {
						p0 = p0 - 0.9 * z * v
					}
					go = p0[ npic + 1 ]

					if( minit >= 10 ) {
						go = 0
					}

				} else {
					go = 0
					minit = 10
				}

			}

			if( minit >= 10 ) {
				if( trace ) .subnpmsg( "m2" )
			}

			a = matrix(a[ , 1:npic ], ncol = npic)
			b = a %*% p0[ 1:npic ]
		}

	}

	p = p0 [ 1:npic ]
	y = 0

	if( ch > 0 ) {

		tmpv = p[ (nineq + 1):npic ] * vscale[ (nc + 2):(nc + np + 1) ]
		funv = .safefunx(tmpv, .solnp_fun, .env,...)
		eqv = .solnp_eqfun(tmpv, ...)
		ineqv = .solnp_ineqfun(tmpv, ...)
		ctmp = get(".solnp_nfn", envir =  .env)
		assign(".solnp_nfn", ctmp + 1, envir = .env)
		ob = c(funv, eqv, ineqv) / vscale[ 1:(nc + 1) ]
	}

	j = ob[ 1 ]

	if( ind[4] ) {
		ob[ (neq + 2):(nc + 1) ] = ob[ (neq + 2):(nc + 1) ] - p[ 1:nineq ]

	}

	if( nc > 0 ) {
		ob[ 2:(nc + 1) ] = ob[ 2:(nc + 1) ] - a %*% p + b
		j = ob[ 1 ] - t(yy) %*% matrix(ob[ 2:(nc + 1) ], ncol=1) + rho * .vnorm(ob[ 2:(nc + 1) ]) ^ 2
	}

	minit = 0
	while( minit < maxit ) {
		minit = minit + 1

		if( ch > 0 ) {

			for (i in 1:np) {

				p[ nineq + i ] = p[ nineq + i ] + delta
				tmpv = p[ (nineq + 1):npic ] * vscale[ (nc + 2):(nc + np + 1) ]
				funv 	= .safefunx(tmpv, .solnp_fun, .env, ...)
				eqv 	= .solnp_eqfun(tmpv, ...)
				ineqv 	= .solnp_ineqfun(tmpv, ...)
				ctmp = get(".solnp_nfn", envir =  .env)
				assign(".solnp_nfn", ctmp + 1, envir = .env)
				obm = c(funv, eqv, ineqv) / vscale[ 1:(nc + 1) ]

				if( ind[4] ) {
					obm[ (neq + 2):(nc + 1)] = obm[ (neq + 2):(nc + 1) ] - p[ 1:nineq ]
				}

				if( nc > 0 ) {

					obm[ 2:(nc + 1) ] = obm[ 2:(nc + 1) ] - a %*% p + b
					obm = obm[ 1 ] - t(yy) %*% obm[ 2:(nc + 1) ] + rho * .vnorm(obm[ 2:(nc + 1 ) ]) ^ 2
				}

				g[ nineq + i ] = (obm - j) / delta
				p[ nineq + i ] = p[ nineq + i ] - delta
			}

			if ( ind[4] ) {
				g[ 1:nineq ] = 0
			}

		}
		if ( minit > 1 ) {
			yg = g - yg
			sx = p - sx
			sc[ 1 ] = t(sx) %*% hessv %*% sx
			sc[ 2 ] = t(sx) %*% yg

			if( (sc[ 1 ] * sc[ 2 ]) > 0 ) {
				sx = hessv %*% sx
				hessv  = hessv - ( sx %*% t(sx) ) / sc[ 1 ] + ( yg %*% t(yg) ) / sc[ 2 ]
			}

		}

		dx = 0.01 * .ones(npic, 1)
		if( ind[11] ) {

			gap = cbind(p[ 1:mm ] - pb[ , 1 ], pb[ , 2 ] - p[ 1:mm ])
			gap = t( apply( gap, 1, FUN = function( x ) sort(x) ) )
			gap = gap[ , 1 ] + sqrt(.eps) * .ones(mm, 1)
			dx[ 1:mm, 1 ] = .ones(mm, 1) / gap
			if( !ind[10] ){
				dx[ (mm + 1):npic, 1 ] = min (c( dx[ 1:mm, 1 ], 0.01) ) * .ones(npic - mm, 1)
			}

		}

		go = -1
		lambda = lambda / 10
		while( go <= 0 ) {
			cz = try(chol( hessv + lambda * diag( as.numeric(dx * dx), length(dx), length(dx) ) ),  silent = TRUE)
			if(inherits(cz, "try-error")){
				p = p * vscale[ (neq + 2):(nc + np + 1) ]  # unscale the parameter vector
				if( nc > 0 ) {
					y = 0 # unscale the lagrange multipliers
				}
				hessv = vscale[ 1 ] * hessv / (vscale[ (neq + 2):(nc + np + 1) ] %*% t(vscale[ (neq + 2):(nc + np + 1) ]) )
				ans = list(p = p, y = y, hessv = hessv, lambda = lambda)
				assign(".solnp_errors", 1, envir = .env)
				return(ans)
			}
			cz = try(solve(cz), silent = TRUE)
			if(inherits(cz, "try-error")){
				p = p * vscale[ (neq + 2):(nc + np + 1) ]  # unscale the parameter vector
				if( nc > 0 ) {
					y = 0 # unscale the lagrange multipliers
				}
				hessv = vscale[ 1 ] * hessv / (vscale[ (neq + 2):(nc + np + 1) ] %*% t(vscale[ (neq + 2):(nc + np + 1) ]) )
				ans = list(p = p, y = y, hessv = hessv, lambda = lambda)
				assign(".solnp_errors", 1, envir = .env)
				return(ans)
			}
			yg = t(cz) %*% g

			if( nc == 0 ) {
				u = -cz %*% yg
			} else{
				y = try( qr.solve(t(cz) %*% t(a), yg), silent = TRUE )
				if(inherits(y, "try-error")){
					p = p * vscale[ (neq + 2):(nc + np + 1) ]  # unscale the parameter vector
					if( nc > 0 ) {
						# y = vscale[ 1 ] * y / vscale[ 2:(nc + 1) ] # unscale the lagrange multipliers
						y = 0
					}
					hessv = vscale[ 1 ] * hessv / (vscale[ (neq + 2):(nc + np + 1) ] %*% t(vscale[ (neq + 2):(nc + np + 1) ]) )
					ans = list(p = p, y = y, hessv = hessv, lambda = lambda)
					assign(".solnp_errors", 1, envir = .env)
					return(ans)
				}

				u = -cz %*% (yg - ( t(cz) %*% t(a) ) %*% y)
			}

			p0 = u[ 1:npic ] + p
			if( !ind[ 11 ] ) {
				go = 1
			} else {
				go = min( c(p0[ 1:mm ] - pb[ , 1 ], pb[ , 2 ] - p0[ 1:mm ]) )
				lambda = 3 * lambda
			}

		}

		alp[ 1 ] = 0
		ob1 = ob
		ob2 = ob1
		sob[ 1 ] = j
		sob[ 2 ] = j
		ptt = cbind(p, p)
		alp[ 3 ] = 1.0
		ptt = cbind(ptt, p0)
		tmpv = ptt[ (nineq + 1):npic, 3 ] * vscale[ (nc + 2):(nc + np + 1) ]
		funv 	= .safefunx(tmpv, .solnp_fun, .env, ...)
		eqv 	= .solnp_eqfun(tmpv, ...)
		ineqv 	= .solnp_ineqfun(tmpv, ...)
		ctmp = get(".solnp_nfn", envir =  .env)
		assign(".solnp_nfn", ctmp + 1, envir = .env)

		ob3 = c(funv, eqv, ineqv) / vscale[ 1:(nc + 1) ]
		sob[ 3 ] = ob3[ 1 ]

		if( ind[4] ) {
			ob3[ (neq + 2):(nc + 1) ] = ob3[ (neq + 2):(nc + 1) ] - ptt[ 1:nineq, 3 ]
		}

		if( nc > 0 ) {
			ob3[ 2:(nc + 1) ] = ob3[ 2:(nc + 1) ] - a %*% ptt[ , 3 ] + b
			sob[ 3 ] = ob3[ 1 ] - t(yy) %*% ob3[ 2:(nc + 1) ] + rho * .vnorm(ob3[ 2:(nc + 1) ]) ^ 2
		}

		go = 1
		while( go > tol ) {
			alp[ 2 ] = (alp[ 1 ] + alp[ 3 ]) / 2
			ptt[ , 2 ] = (1 - alp[ 2 ]) * p + alp[ 2 ] * p0
			tmpv = ptt[ (nineq + 1):npic, 2 ] * vscale[ (nc + 2):(nc + np + 1) ]
			funv 	= .safefunx(tmpv, .solnp_fun, .env, ...)
			eqv 	= .solnp_eqfun(tmpv, ...)
			ineqv 	= .solnp_ineqfun(tmpv, ...)
			ctmp = get(".solnp_nfn", envir =  .env)
			assign(".solnp_nfn", ctmp + 1, envir = .env)

			ob2 = c(funv, eqv, ineqv) / vscale[ 1:(nc + 1) ]

			sob[ 2 ] = ob2[ 1 ]

			if( ind[4] ) {
				ob2[ (neq + 2):(nc + 1) ] = ob2[ (neq + 2):(nc + 1) ] - ptt[ 1:nineq , 2 ]
			}

			if( nc > 0 ) {
				ob2[ 2:(nc + 1) ] = ob2[ 2:(nc + 1) ] - a %*% ptt[ , 2 ] + b
				sob[ 2 ] = ob2[ 1 ] - t(yy) %*% ob2[ 2:(nc + 1) ] + rho * .vnorm(ob2[ 2:(nc + 1) ]) ^ 2
			}

			obm = max(sob)

			if( obm < j ) {
				obn = min(sob)
				go = tol * (obm - obn) / (j - obm)
			}

			condif1 = sob[ 2 ] >= sob[ 1 ]
			condif2 = sob[ 1 ] <= sob[ 3 ] && sob[ 2 ] < sob[ 1 ]
			condif3 = sob[ 2 ] <  sob[ 1 ] && sob[ 1 ] > sob[ 3 ]

			if( condif1 ) {
				sob[ 3 ] = sob[ 2 ]
				ob3 = ob2
				alp[ 3 ] = alp[ 2 ]
				ptt[ , 3 ] = ptt[ , 2 ]
			}

			if( condif2 ) {
				sob[ 3 ] = sob[ 2 ]
				ob3 = ob2
				alp[ 3 ] = alp[ 2 ]
				ptt[ , 3 ] = ptt[ , 2 ]
			}

			if( condif3 ) {
				sob[ 1 ] = sob[ 2 ]
				ob1 = ob2
				alp[ 1 ] = alp[ 2 ]
				ptt[ , 1 ] = ptt[ , 2 ]
			}

			if( go >= tol ) {
				go = alp[ 3 ] - alp[ 1 ]
			}

		}

		sx = p
		yg = g
		ch = 1
		obn = min(sob)
		if( j <= obn ) {
			maxit = minit
		}

		reduce = (j - obn) / ( 1 + abs(j) )

		if( reduce < tol ) {
			maxit = minit
		}

		condif1 = sob[ 1 ] <  sob[ 2 ]
		condif2 = sob[ 3 ] <  sob[ 2 ] && sob[ 1 ] >= sob[ 2 ]
		condif3 = sob[ 1 ] >= sob[ 2 ] && sob[ 3 ] >= sob[ 2 ]

		if( condif1 ) {
			j = sob[ 1 ]
			p = ptt[ , 1 ]
			ob = ob1
		}

		if( condif2 ) {
			j = sob [ 3 ]
			p = ptt[ , 3 ]
			ob = ob3
		}

		if( condif3 ) {
			j = sob[ 2 ]
			p = ptt[ , 2 ]
			ob = ob2
		}

	}

	p = p * vscale[ (neq + 2):(nc + np + 1) ]  # unscale the parameter vector

	if( nc > 0 ) {
		y = vscale[ 1 ] * y / vscale[ 2:(nc + 1) ] # unscale the lagrange multipliers
	}

	hessv = vscale[ 1 ] * hessv / (vscale[ (neq + 2):(nc + np + 1) ] %*% t(vscale[ (neq + 2):(nc + np + 1) ]) )
	if( reduce > tol ) {
		if( trace ) .subnpmsg( "m3" )
	}

	ans = list(p = p, y = y, hessv = hessv, lambda = lambda)
	return( ans )
}


.fdgrad = function(i, p, delta, np, vscale, constraint, j, nineq, npic, nc, .solnp_fun,
		.solnp_eqfun, .solnp_ineqfun, .env, ...)
{
	ans = list()
	px = p
	px[ nineq + i ] = px[ nineq + i ] + delta
	tmpv = px[ (nineq + 1):npic ] * vscale[ (nc + 2):(nc + np + 1) ]
	funv = .safefunx(tmpv, .solnp_fun, .env, ...)
	eqv 	= .solnp_eqfun(tmpv, ...)
	ineqv 	= .solnp_ineqfun(tmpv, ...)
	ctmp = get(".solnp_nfn", envir =  .env)
	assign(".solnp_nfn", ctmp + 1, envir = .env)
	ob = c(funv, eqv, ineqv) / vscale[ 1:(nc + 1) ]
	ans$p	= px
	ans$ob 	= ob
	ans$g   = (ob[ 1 ] - j) / delta
	ans$a   = (ob[ 2:(nc + 1) ] - constraint) / delta
	return( ans )
}

#' Nonlinear optimization using augmented Lagrange method (original version)
#'
#' @param pars an numeric vector of decision variables (length n).
#' @param fun the objective function (must return a scalar).
#' @param eqfun an optional function for calculating equality constraints.
#' @param eqB a vector of the equality bounds (if eq_fn provided).
#' @param ineqfun an optional function for calculating inequality constraints.
#' @param ineqLB the lower bounds for the inequality (must be finite)
#' @param ineqUB the upper bounds for the inequalitiy (must be finite)
#' @param LB lower bounds on decision variables
#' @param UB upper bounds on decision variables
#' @param control a list of solver control parameters (see details).
#' @param ... additional arguments passed to the supplied functions (common to all functions supplied).
#' @returns An list with the following slot:
#' \describe{
#'  \item{pars}{Optimal Parameters.}
#'  \item{convergence }{Indicates whether the solver has converged (0) or not (1 or 2).}
#'  \item{values}{Vector of function values during optimization with last one the
#'    value at the optimal.}
#'  \item{lagrange}{The vector of Lagrange multipliers.}
#'  \item{hessian}{The Hessian of the augmented problem at the optimal solution.}
#'  \item{ineqx0}{The estimated optimal inequality vector of slack variables used
#'    for transforming the inequality into an equality constraint.}
#'  \item{nfuneval}{The number of function evaluations.}
#'  \item{elapsed}{Time taken to compute solution.}
#'}
#' @details
#' The optimization problem solved by \code{csolnp} is formulated as:
#' \deqn{
#' \begin{aligned}
#' \min_{x \in \mathbb{R}^n} \quad & f(x) \\
#' \text{s.t.} \quad & g(x) = b \\
#' & h_l \le h(x) \le h_u\\
#' & x_l \le x \le x_u\\
#' \end{aligned}
#' }
#'
#' where \eqn{f(x)} is the objective function, \eqn{g(x)} is the vector of equality constraints
#' with target value \eqn{b}, \eqn{h(x)} is the vector of inequality constraints bounded
#' by \eqn{h_l} and \eqn{h_u}, with parameter bounds \eqn{x_l} and \eqn{x_u}. Internally,
#' inequality constraints are converted into equality constraints using slack variables
#' and solved using an augmented Lagrangian approach.
#' The control is a list with the following options:
#'\describe{
#'  \item{rho}{This is used as a penalty weighting scaler for infeasibility in the
#'  augmented objective function. The higher its value the more the weighting to
#'  bring the solution into the feasible region (default 1). However, very high
#'  values might lead to numerical ill conditioning or significantly slow down
#'  convergence.}
#'  \item{outer.iter}{Maximum number of major (outer) iterations (default 400).}
#'  \item{inner.iter}{Maximum number of minor (inner) iterations (default 800).}
#'  \item{delta}{Relative step size in forward difference evaluation (default 1.0e-7).}
#'  \item{tol}{ Relative tolerance on feasibility and optimality (default 1e-8).}
#'  \item{trace}{The value of the objective function and the parameters is printed
#'  at every major iteration (default 1).}
#'}
#' @examples
#' {
#' # From the original paper by Y.Ye
#' # see the unit tests for more....
#' # POWELL Problem
#' fn1 = function(x)
#' {
#'     exp(x[1] * x[2] * x[3] * x[4] * x[5])
#' }
#' eqn1 = function(x){
#'     z1 = x[1] * x[1] + x[2] * x[2] + x[3] * x[3] + x[4] * x[4] + x[5] * x[5]
#'     z2 = x[2] * x[3] - 5 * x[4] * x[5]
#'     z3 = x[1] * x[1] * x[1] + x[2] * x[2] * x[2]
#'     return(c(z1, z2, z3))
#' }
#' x0 = c(-2, 2, 2, -1, -1)
#' }
#' powell = solnp(x0, fun = fn1, eqfun = eqn1, eqB = c(10, 0, -1))
#' @keywords optimize
#' @rdname solnp
#' @author Alexios Galanos
#' @export
#'
solnp = function(pars, fun, eqfun = NULL, eqB = NULL, ineqfun = NULL, ineqLB = NULL, ineqUB = NULL, LB = NULL, UB = NULL, control = list(), ...)
{
	# start timer
	tic = Sys.time()
	xnames = names(pars)
	# get environment
	.solnpenv <- environment()
	assign("xnames", xnames, envir = .solnpenv)
	# initiate function count
	assign(".solnp_nfn", 0, envir = .solnpenv)
	assign(".solnp_errors", 0, envir = .solnpenv)

	# index of function indicators
	# [1] length of pars
	# [2] has function gradient?
	# [3] has hessian?
	# [4] has ineq?
	# [5] ineq length
	# [6] has jacobian (inequality)
	# [7] has eq?
	# [8] eq length
	# [9] has jacobian (equality)
	# [10] has upper / lower bounds
	# [11] has either lower/upper bounds or ineq


	ind = rep(0, 11)
	np = ind[1]  = length(pars)
	# lower parameter bounds - indicator
	# lpb[1]=1 means lower/upper bounds present
	# lpb[2]=1 means lower/upper bounds OR inequality bounds present

	# do parameter and LB/UB checks
	check1 = .checkpars(pars, LB, UB, .solnpenv)

	# .LB and .UB assigned

	.LB = get(".LB", envir = .solnpenv)
	.UB = get(".UB", envir = .solnpenv)


	if( !is.null(.LB) || !is.null(.UB) ) ind[10] = 1

	# do function checks and return starting value
	funv = .checkfun(pars, fun, .solnpenv, ...)
	#.solnp_fun assigned
	.solnp_fun = get(".solnp_fun", envir = .solnpenv)

	# Analytical Gradient Functionality not yet implemented in subnp function

	# gradient and hessian checks
	#if(!is.null(grad)){
	#	gradv = .checkgrad(pars, grad, .solnpenv, ...)
	#	ind[2] = 1
	#} else{
	#	.solnp_gradfun = function(pars, ...) .fdgrad(pars, fun = .solnp_fun, ...)
		ind[2] = 0
	#	gradv = .solnp_gradfun(pars, ...)
	#}
	# .solnp_gradfun(pars, ...) assigned

	.solnp_hessfun = NULL
	ind[3] = 0
	#hessv = NULL
	# .solnp_hessfun(pars, ...) assigned

	# do inequality checks and return starting values

	if(!is.null(ineqfun)){
		ineqv 	= .checkineq(pars, ineqfun, ineqLB, ineqUB, .solnpenv, ...)
		ind[4] 	= 1
		nineq 	= length(ineqLB)
		ind[5] 	= nineq

		# check for infinities/nans
		.ineqLBx = .ineqLB
		.ineqUBx = .ineqUB
		.ineqLBx[!is.finite(.ineqLB)] = -1e10
		.ineqUBx[!is.finite(.ineqUB)] =  1e10
		ineqx0 	= (.ineqLBx + .ineqUBx)/2
		#if(!is.null(ineqgrad)){
		#	ineqjacv = .cheqjacineq(pars, gradineq, .ineqUB, .ineqLB, .solnpenv, ...)
		#	ind[6] = 1
		#} else{
		# .solnp_ineqjac = function(pars, ...) .fdjac(pars, fun = .solnp_ineqfun, ...)
		ind[6] = 0
		#ineqjacv = .solnp_ineqjac(pars, ...)
		#}
	} else{
		.solnp_ineqfun = function(pars, ...) .emptyfun(pars, ...)
		# .solnp_ineqjac = function(pars, ...) .emptyjac(pars, ...)
		ineqv 	= NULL
		ind[4] 	= 0
		nineq 	= 0
		ind[5] 	= 0
		ind[6] 	= 0
		ineqx0 	= NULL
		.ineqLB = NULL
		.ineqUB = NULL
	}
	# .solnp_ineqfun and .solnp_ineqjac assigned
	# .ineqLB and .ineqUB assigned
	.solnp_ineqfun = get(".solnp_ineqfun", envir = .solnpenv)
	.ineqLB = get(".ineqLB", envir = .solnpenv)
	.ineqUB = get(".ineqUB", envir = .solnpenv)


	# equality checks
	if(!is.null(eqfun)){
		eqv 	= .checkeq(pars, eqfun, eqB, .solnpenv, ...)
		ind[7] 	= 1
		.eqB = get(".eqB", envir = .solnpenv)
		neq 	= length(.eqB)
		ind[8] 	= neq
		#if(!is.null(eqgrad)){
		#	eqjacv = .cheqjaceq(pars, gradeq, .solnpenv, ...)
		#	ind[9] = 1
		#} else{
		#	.solnp_eqjac = function(pars, ...) .fdjac(pars, fun = .solnp_eqfun, ...)
		#	eqjacv = .solnp_eqjac(pars, ...)
			ind[9] = 0
		#}
	} else {
		eqv = NULL
		#eqjacv = NULL
		.solnp_eqfun = function(pars, ...) .emptyfun(pars, ...)
		#.solnp_eqjac = function(pars, ...) .emptyjac(pars, ...)
		ind[7] 	= 0
		neq 	= 0
		ind[8] 	= 0
		ind[9] 	= 0
	}
	# .solnp_eqfun(pars, ...) and .solnp_eqjac(pars, ...) assigned
	# .solnp_eqB assigned
	.solnp_eqfun = get(".solnp_eqfun", envir = .solnpenv)

	if( ind[ 10 ] || ind [ 4 ]) ind[ 11 ] = 1

	# parameter bounds (pb)
	pb  = rbind( cbind(.ineqLB, .ineqUB), cbind(.LB, .UB) )

	# check control list
	ctrl  = .solnpctrl( control )
	rho   = ctrl[[ 1 ]]
	# maxit = outer iterations
	maxit = ctrl[[ 2 ]]
	# minit = inner iterations
	minit = ctrl[[ 3 ]]
	delta = ctrl[[ 4 ]]
	tol   = ctrl[[ 5 ]]
	trace = ctrl[[ 6 ]]

	# total constraints (tc) = no.inequality constraints + no.equality constraints
	tc = nineq + neq

	# initialize fn value and inequalities and set to NULL those not needed
	j  = jh = funv
	tt = 0 * .ones(3, 1)

	if( tc > 0 ) {
		# lagrange multipliers (lambda)
		lambda = 0 * .ones(tc, 1)
		# constraint vector = [1:neq 1:nineq]
		constraint = c(eqv, ineqv)
		if( ind[4] ) {
			tmpv = cbind(constraint[ (neq + 1):tc ] - .ineqLB, .ineqUB - constraint[ (neq + 1):tc ] )
			testmin = apply( tmpv, 1, FUN = function( x ) min(x[ 1 ], x[ 2 ]) )
			if( all(testmin > 0) ) ineqx0 = constraint[ (neq + 1):tc ]
			constraint[ (neq + 1):tc ] = constraint[ (neq + 1):tc ] - ineqx0
		}
		tt[ 2 ] = .vnorm(constraint)
		if( max(tt[ 2 ] - 10 * tol, nineq, na.rm = TRUE) <= 0 ) rho = 0
	} else{
		lambda = 0
	}
	# starting augmented parameter vector
	p  = c(ineqx0, pars)
	hessv  = diag(np + nineq)
	mu = np
	.solnp_iter = 0
	ob = c(funv, eqv, ineqv)

	while( .solnp_iter < maxit ){
		.solnp_iter = .solnp_iter + 1
		.subnp_ctrl = c(rho, minit, delta, tol, trace)

		# make the scale for the cost, the equality constraints, the inequality
		# constraints, and the parameters
		if( ind[7] ) {
			# [1 neq]
			vscale = c( ob[ 1 ], rep(1, neq) * max( abs(ob[ 2:(neq + 1) ]) ) )
		} else {
			vscale = 1
		}

		if( !ind[ 11 ] ) {
			vscale = c(vscale, p)
		} else {
			# [ 1 neq np]
			vscale = c(vscale, rep( 1, length.out = length(p) ) )
		}

		vscale = apply( matrix(vscale, ncol = 1), 1, FUN = function( x ) min( max( abs(x), tol ), 1/tol ) )

		res   = .subnp(pars = p, yy = lambda, ob = ob, hessv = hessv, lambda = mu, vscale = vscale,
				ctrl = .subnp_ctrl, .env = .solnpenv, ...)
		if(get(".solnp_errors", envir =  .solnpenv) == 1){
			maxit = .solnp_iter
		}
		p  = res$p
		lambda  = res$y
		hessv  = res$hessv
		mu = res$lambda
		temp = p[ (nineq + 1):(nineq + np) ]
		funv = .safefunx(temp, .solnp_fun, .env = .solnpenv, ...)
		ctmp = get(".solnp_nfn", envir =  .solnpenv)
		assign(".solnp_nfn", ctmp + 1, envir = .solnpenv)

		tempdf = cbind(temp, funv)

		if( trace ){
			.report(.solnp_iter, funv, temp)
		}

		eqv = .solnp_eqfun(temp, ...)
		ineqv = .solnp_ineqfun(temp, ...)

		ob = c(funv, eqv, ineqv)

		tt[ 1 ] = (j - ob[ 1 ]) / max(abs(ob[ 1 ]), 1)
		j = ob[ 1 ]

		if( tc > 0 ){
			constraint = ob[ 2:(tc + 1) ]

			if( ind[ 4 ] ){
				tempv = rbind( constraint[ (neq + 1):tc ] - pb[ 1:nineq, 1 ],
				              pb[ 1:nineq, 2 ] - constraint[ (neq + 1):tc ] )

				if( min(tempv) > 0 ) {
					p[ 1:nineq ] = constraint[ (neq + 1):tc ]
				}

				constraint[ (neq + 1):tc ] = constraint[ (neq + 1):tc ] - p[ 1:nineq ]
			}

			tt[ 3 ] = .vnorm(constraint)

			if( tt[ 3 ] < 10 * tol ) {
				rho = 0
				mu  = min(mu, tol)
			}

			if( tt[ 3 ] < 5 * tt[ 2 ]) {
				rho = rho/5
			}

			if( tt[ 3 ] > 10 * tt[ 2 ]) {
				rho = 5 * max( rho, sqrt(tol) )
			}

			if( max( c( tol + tt[ 1 ], tt[ 2 ] - tt[ 3 ] ) ) <= 0 ) {
				lambda = 0
				hessv = diag( diag ( hessv ) )
			}

			tt[ 2 ] = tt[ 3 ]
		}

		if( .vnorm( c(tt[ 1 ], tt[ 2 ]) ) <= tol ) {
			maxit = .solnp_iter
		}

		jh = c(jh, j)
	}

	if( ind[ 4 ] ) {
		ineqx0 = p[ 1:nineq ]
	}

	p = p[ (nineq + 1):(nineq + np) ]

	if(get(".solnp_errors", envir =  .solnpenv) == 1){
		convergence = 2
		if( trace ) cat( paste( "\nsolnp--> Solution not reliable....Problem Inverting Hessian.\n", sep="" ) )
	} else{
		if( .vnorm( c(tt[ 1 ], tt[ 2 ]) ) <= tol ) {
			convergence = 0
			if( trace ) cat( paste( "\nsolnp--> Completed in ", .solnp_iter, " iterations\n", sep="" ) )
		} else{
			convergence = 1
			if( trace ) cat( paste( "\nsolnp--> Exiting after maximum number of iterations\n",
							"Tolerance not achieved\n", sep="" ) )
		}
	}
	# end timer
	ctmp = get(".solnp_nfn", envir =  .solnpenv)
	toc = Sys.time() - tic
	names(p) = xnames
	ans = list(pars = p, convergence = convergence, values = as.numeric(jh), lagrange = lambda,
			hessian = hessv, ineqx0 = ineqx0, nfuneval = ctmp, outer.iter = .solnp_iter,
			elapsed = toc, vscale = vscale)
	return( ans )
}

# auto_scale_factors <- function(pars, fn, gr, min_scale = 1e-12, max_scale = 1e4, eps = .Machine$double.eps^0.25, ...)
# {
#   if (!is.null(gr)) {
#     g <- gr(pars, ...)
#     H_approx <- tcrossprod(g)  # g %*% t(g), outer product
#     if (!corpcor::is.positive.definite(H_approx)) H_approx <- corpcor::make.positive.definite(H_approx)
#     V_approx <- solve(H_approx)
#     diag_V <- abs(diag(V_approx))
#     scale <- sqrt(pmax(diag_V, eps))
#     D <- pmin(pmax(scale, min_scale), max_scale)
#   } else {
#     g <- grad(func = fn, x = pars, method = "Richardson", side = NULL, method.args = list(eps = eps), ...)
#     H_approx <- tcrossprod(g)  # g %*% t(g), outer product
#     if (!corpcor::is.positive.definite(H_approx)) H_approx <- corpcor::make.positive.definite(H_approx)
#     V_approx <- solve(H_approx)
#     diag_V <- abs(diag(V_approx))
#     scale <- sqrt(pmax(diag_V, eps))
#     D <- pmin(pmax(scale, min_scale), max_scale)
#   }
#   return(D)
# }


solnp_problem_setup <- function(pars, fn, gr, eq_fn, eq_b, eq_jac, ineq_fn, ineq_lower, ineq_upper, ineq_jac, lower, upper, ...)
{
  # index of function indicators
  # [1] length of pars
  # [2] has function gradient?
  # [3] 0 [empty for now]
  # [4] has ineq?
  # [5] ineq length
  # [6] has jacobian (inequality)
  # [7] has eq?
  # [8] eq length
  # [9] has jacobian (equality)
  # [10] has upper / lower bounds
  # [11] has either lower/upper bounds or ineq

  setup_index <- rep(0, 11)
  if (any(is.na(pars))) stop("\nNA found in initial pars.")
  n <- length(pars)
  setup_index[1] <- n

  # check objective function (fn)
  initial_fn_value <- try(fn(pars, ...), silent = TRUE)
  if (inherits(initial_fn_value, 'try-error')) stop("\nobjective function (fn) returned an error on evaluation with initial pars.")
  if (length(initial_fn_value) != 1L) stop("\nobjective function (fn) must return a scalar value.")
  if (is.na(initial_fn_value)) stop("\nobjective function (fn) returned NA with initial pars.")

  # check gradient function (gr)
  if (!is.null(gr)) {
    initial_gr_value <- try(gr(pars, ...), silent = TRUE)
    if (inherits(initial_gr_value, 'try-error')) stop("\ngradient (gr) returned an error on evaluation with initial pars.")
    if (length(initial_gr_value) != n) stop("\ngradient (gr) must return a vector equal to length(pars).")
    if (any(is.na(initial_gr_value))) stop("\nNA's detected in gradient (gr) evaluation with initial pars.")
    if (any(!is.finite(initial_gr_value))) stop("\nnon finite values detected in gradient (gr) evaluation with initial pars.")
    setup_index[2] <- 1
  }

  # check inequality function (ineq_fn) and jacobian
  if (!is.null(ineq_fn)) {
    initial_ineq_value <- try(ineq_fn(pars, ...), silent = TRUE)
    if (inherits(initial_ineq_value, 'try-error')) stop("\nineq_fn returned an error on evaluation with initial pars.")
    nineq <- length(initial_ineq_value)
    setup_index[4] <- 1
    setup_index[5] <- nineq
    if (is.null(ineq_lower) | is.null(ineq_upper)) stop("\nineq_lower and ineq_upper cannot be NULL when ineq_fn provided.")
    if (length(ineq_lower) != nineq) stop(paste0("\nineq_lower must be of length : ", nineq))
    if (length(ineq_upper) != nineq) stop(paste0("\nineq_upper must be of length : ", nineq))
    if (any(is.na(ineq_upper)) | any(!is.finite(ineq_upper))) stop("\nineq_upper must not contain any NA or non finite values.")
    if (any(is.na(ineq_lower)) | any(!is.finite(ineq_lower))) stop("\nineq_lower must not contain any NA or non finite values.")
    if (any((ineq_upper - initial_ineq_value) < 0)) warning("\nupper inequality values violated with initial pars.")
    if (any((initial_ineq_value - ineq_lower) < 0)) warning("\nlower inequality values violated with initial pars.")
    if (!is.null(ineq_jac)) {
      initial_ineq_jac_value <- try(ineq_jac(pars, ...), silent = TRUE)
      if (inherits(initial_ineq_jac_value, 'try-error')) stop("\nineq_jac returned an error on evaluation with initial pars.")
      if (!is.matrix(initial_ineq_jac_value)) stop(paste0("\nineq_jac must return a matrix of dimensions : ", nineq, " x ", n))
      if (any(is.na(initial_ineq_jac_value))) stop("\nNA's detected in ineq_jac evaluation with initial pars.")
      if (any(!is.finite(initial_ineq_jac_value))) stop("\nnon-finite values detected in ineq_jac evaluation with initial pars.")
      setup_index[6] <- 1
    }
  }

  # check equality function (eq_fn) and jacobian
  if (!is.null(eq_fn)) {
    initial_eq_value <- try(eq_fn(pars, ...), silent = TRUE)
    if (inherits(initial_eq_value, 'try-error')) stop("\neq_fn returned an error on evaluation with initial pars.")
    neq <- length(initial_eq_value)
    setup_index[7] <- 1
    setup_index[8] <- neq
    if (is.null(eq_b)) stop("\neq_b cannot be NULL with eq_fn provided.")
    if (any(is.na(eq_b)) | any(!is.finite(eq_b))) stop("\neq_b must not contain any NA or non finite values.")
    if (!is.null(eq_jac)) {
      initial_eq_jac_value <- try(eq_jac(pars, ...), silent = TRUE)
      if (inherits(initial_eq_jac_value, 'try-error')) stop("\neq_jac returned an error on evaluation with initial pars.")
      if (!is.matrix(initial_eq_jac_value)) stop(paste0("\neq_jac must return a matrix of dimensions : ", neq, " x ", n))
      if (any(is.na(initial_eq_jac_value))) stop("\nNA's detected in eq_jac evaluation with initial pars.")
      if (any(!is.finite(initial_eq_jac_value))) stop("\nnon-finite values detected in eq_jac evaluation with initial pars.")
      setup_index[9] <- 1
    }
  }

  # check bounds

  if (!is.null(lower) || !is.null(upper)) {
    setup_index[10] <- 1
    if (length(lower) != n) stop(paste0("\nlower must be of length : ", n))
    if (length(upper) != n) stop(paste0("\nupper must be of length : ", n))
    if (any(is.na(lower)) | any(!is.finite(lower))) stop("\nlower must not contain any NA or non finite values.")
    if (any(is.na(upper)) | any(!is.finite(upper))) stop("\nupper must not contain any NA or non finite values.")
    if (any((upper - pars) < 0)) warning("\nupper bound values violated with initial pars.")
    if (any((pars - lower) < 0)) warning("\nlower bound values violated with initial pars.")
  }
  setup_index[11] <- max(setup_index[10], setup_index[4])
  return(setup_index)
}


solnp_objective_wrapper <- function(fn, gr, n, ...)
{
  f <- function(pars)
  {
    if (any(is.na(pars))) stop("\nNA's detected in parameters")
    out <- try(fn(pars, ...), silent = TRUE)
    if (inherits(out, 'try-error')) {
      warning("\nerror in objective function. Returning 1e10")
      return(1e10)
    } else {
      if (is.na(out) | !is.finite(out)) return(1e10)
      return(out)
    }
  }

  if (!is.null(gr)) {
    g <- function(pars)
    {
      if (any(is.na(pars))) stop("\nNA's detected in parameters")
      out <- try(gr(pars, ...), silent = TRUE)
      if (inherits(out, 'try-error')) {
        warning("\nerror in gradient function. Returning 1e10")
        return(1e10)
      } else {
        if (any(is.na(out)) | any(!is.finite(out))) return(rep(1e5, n))
        return(out)
      }
    }
  } else {
    g <- function(pars)
    {
      if (any(is.na(pars))) stop("\nNA's detected in parameters")
      out <- try(grad(func = f, x = pars), silent = TRUE)
      if (inherits(out, 'try-error')) {
        warning("\nerror in gradient function. Returning 1e10")
        return(1e10)
      } else {
        if (any(is.na(out)) | any(!is.finite(out))) return(rep(1e5, n))
        return(out)
      }
    }
  }
  return(list(objective_fun = f, gradient_fun = g))
}


# solnp_scaled_objective_wrapper <- function(fn, gr, n, scale_vector, ...)
# {
#   f <- function(pars)
#   {
#     if (any(is.na(pars))) stop("\nNA's detected in parameters")
#     out <- try(fn(pars * scale_vector, ...), silent = TRUE)
#     if (inherits(out, 'try-error')) {
#       warning("\nerror in objective function. Returning 1e10")
#       return(1e10)
#     } else {
#       if (is.na(out) | !is.finite(out)) return(1e10)
#       return(out)
#     }
#   }
#
#   if (!is.null(gr)) {
#     g <- function(pars)
#     {
#       if (any(is.na(pars))) stop("\nNA's detected in parameters")
#       out <- try(gr(pars * scale_vector, ...) * scale_vector, silent = TRUE)
#       if (inherits(out, 'try-error')) {
#         warning("\nerror in gradient function. Returning 1e10")
#         return(1e10)
#       } else {
#         if (any(is.na(out)) | any(!is.finite(out))) return(rep(1e5, n))
#         return(out)
#       }
#     }
#   } else {
#     g <- function(pars)
#     {
#       if (any(is.na(pars))) stop("\nNA's detected in parameters")
#       out <- try(grad(func = f, x = pars * scaled_vector) * scale_vector, silent = TRUE)
#       if (inherits(out, 'try-error')) {
#         warning("\nerror in gradient function. Returning 1e10")
#         return(1e10)
#       } else {
#         if (any(is.na(out)) | any(!is.finite(out))) return(rep(1e5, n))
#         return(out)
#       }
#     }
#   }
#   return(list(objective_fun = f, gradient_fun = g))
# }


solnp_ineq_wrapper <- function(ineq_fn, ineq_jac, n_ineq, n, ...)
{
  f <- j <- NULL
  if (!is.null(ineq_fn)) {
    f <- function(pars)
    {
      if (any(is.na(pars))) stop("\nNA's detected in parameters")
      # Apply scaling factor if needed
      out <- try(ineq_fn(pars, ...), silent = TRUE)
      if (inherits(out, 'try-error')) {
        warning("\nerror in inequality function evaluation. Returning large violation.")
        return(rep(1e10, n_ineq))
      } else {
        if (any(is.na(out)) | any(!is.finite(out))) {
          out[is.na(out)] <- 1e10
          out[!is.finite(out)] <- 1e10
        }
        return(out)
      }
    }

    j <- NULL
    if (!is.null(ineq_jac)) {
      # If user provides an analytical Jacobian
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to inequality Jacobian function wrapper.")
        }
        jac_out <- try(ineq_jac(pars, ...), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in user-provided analytical inequality Jacobian function. Returning zeros.")
          return(.zeros(n_ineq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in user-provided analytical inequality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    } else {
      # If no analytical Jacobian is provided, use numerical differentiation (numDeriv)
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to numerical inequality Jacobian function wrapper.")
        }
        jac_out <- try(jacobian(func = f, x = pars, method = "Richardson"), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in numerical inequality Jacobian calculation. Returning zeros.")
          return(.zeros(n_ineq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in numerical inequality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    }
  } else {
    f <- function(pars) return(NULL)
    j <- function(pars) return(NULL)
  }
  return(list(ineq_f = f, ineq_j = j))
}

solnp_scaled_ineq_wrapper <- function(ineq_fn, ineq_jac, n_ineq, n, scale_vector, ...)
{
  f <- j <- NULL
  if (!is.null(ineq_fn)) {
    f <- function(pars)
    {
      if (any(is.na(pars))) stop("\nNA's detected in parameters")
      # Apply scaling factor if needed
      out <- try(ineq_fn(pars * scale_vector, ...), silent = TRUE)
      if (inherits(out, 'try-error')) {
        warning("\nerror in inequality function evaluation. Returning large violation.")
        return(rep(1e10, n_ineq))
      } else {
        if (any(is.na(out)) | any(!is.finite(out))) {
          out[is.na(out)] <- 1e10
          out[!is.finite(out)] <- 1e10
        }
        return(out)
      }
    }

    j <- NULL
    if (!is.null(ineq_jac)) {
      # If user provides an analytical Jacobian
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to inequality Jacobian function wrapper.")
        }
        jac_out <- try(ineq_jac(pars * scale_vector, ...) %*% diag(scale_vector), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in user-provided analytical inequality Jacobian function. Returning zeros.")
          return(.zeros(n_ineq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in user-provided analytical inequality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    } else {
      # If no analytical Jacobian is provided, use numerical differentiation (numDeriv)
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to numerical inequality Jacobian function wrapper.")
        }
        jac_out <- try(jacobian(func = f, x = pars * scale_vector, method = "Richardson") %*% diag(scale_vector), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in numerical inequality Jacobian calculation. Returning zeros.")
          return(.zeros(n_ineq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in numerical inequality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    }
  } else {
    f <- function(pars) return(NULL)
    j <- function(pars) return(NULL)
  }
  return(list(ineq_f = f, ineq_j = j))
}


solnp_eq_wrapper <- function(eq_fn, eq_b, eq_jac, n_eq, n , ...)
{
  f <- j <- NULL
  if (!is.null(eq_fn)) {
    f <- function(pars)
    {
      if (any(is.na(pars))) stop("\nNA's detected in parameters")
      # Apply scaling factor if needed
      out <- try(eq_fn(pars, ...) - eq_b, silent = TRUE)
      if (inherits(out, 'try-error')) {
        warning("\nerror in equality function evaluation. Returning large violation.")
        return(rep(1e10, n_eq))
      } else {
        if (any(is.na(out)) | any(!is.finite(out))) {
          out[is.na(out)] <- 1e10
          out[!is.finite(out)] <- 1e10
        }
        return(out)
      }
    }
    j <- NULL
    if (!is.null(eq_jac)) {
      # If user provides an analytical Jacobian
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to equality Jacobian function wrapper.")
        }
        jac_out <- try(eq_jac(pars, ...), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in user-provided analytical inequality Jacobian function. Returning zeros.")
          return(.zeros(n_eq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in user-provided analytical equality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    } else {
      # If no analytical Jacobian is provided, use numerical differentiation (numDeriv)
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to numerical inequality Jacobian function wrapper.")
        }
        jac_out <- try(jacobian(func = f, x = pars, method = "Richardson"), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in numerical inequality Jacobian calculation. Returning zeros.")
          return(.zeros(n_eq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in numerical inequality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    }
  }  else {
    f <- function(pars) return(NULL)
    j <- function(pars) return(NULL)
  }
  return(list(eq_f = f, eq_j = j))
}

solnp_scaled_eq_wrapper <- function(eq_fn, eq_b, eq_jac, n_eq, n, scale_vector, ...)
{
  f <- j <- NULL
  if (!is.null(eq_fn)) {
    f <- function(pars)
    {
      if (any(is.na(pars))) stop("\nNA's detected in parameters")
      # Apply scaling factor if needed
      out <- try(eq_fn(pars * scale_vector, ...) - eq_b, silent = TRUE)
      if (inherits(out, 'try-error')) {
        warning("\nerror in equality function evaluation. Returning large violation.")
        return(rep(1e10, n_eq))
      } else {
        if (any(is.na(out)) | any(!is.finite(out))) {
          out[is.na(out)] <- 1e10
          out[!is.finite(out)] <- 1e10
        }
        return(out)
      }
    }
    j <- NULL
    if (!is.null(eq_jac)) {
      # If user provides an analytical Jacobian
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to equality Jacobian function wrapper.")
        }
        jac_out <- try(eq_jac(pars * scale_vector, ...) %*% diag(scale_vector), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in user-provided analytical inequality Jacobian function. Returning zeros.")
          return(.zeros(n_eq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in user-provided analytical equality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    } else {
      # If no analytical Jacobian is provided, use numerical differentiation (numDeriv)
      j <- function(pars) {
        if (any(is.na(pars))) {
          stop("\nNA's detected in parameters passed to numerical inequality Jacobian function wrapper.")
        }
        jac_out <- try(jacobian(func = f, x = pars * scale_vector, method = "Richardson") %*% diag(scale_vector), silent = TRUE)
        if (inherits(jac_out, 'try-error')) {
          warning("\nError in numerical inequality Jacobian calculation. Returning zeros.")
          return(.zeros(n_eq, n)) # Return a matrix of zeros on error
        } else {
          if (any(is.na(jac_out)) | any(!is.finite(jac_out))) {
            warning("\nNA/Inf detected in numerical inequality Jacobian. Converting to zeros.")
            jac_out[is.na(jac_out)] <- 0
            jac_out[!is.finite(jac_out)] <- 0
          }
          return(jac_out)
        }
      }
    }
  }  else {
    f <- function(pars) return(NULL)
    j <- function(pars) return(NULL)
  }
  return(list(eq_f = f, eq_j = j))
}


augmented_function_gradient <- function(gr, n, n_ineq)
{
  f <- function(x) {
    # Extract original parameters from the augmented vector (augmented_pars = [s, P])
    original_parameters <- x[(n_ineq + 1):(n_ineq + n)]
    # Call the user's gradient function
    user_grad <- gr(original_parameters)
    # Construct the full augmented gradient: [ 0 | d(obj)/d(P) ]
    # Objective function does not depend on slack variables, so components for s are zero.
    augmented_grad <- c(.zeros(n_ineq, 1), user_grad)
    return(augmented_grad)
  }
  return(f)
}

augmented_equality_jacobian <- function(eq_jac, n_eq, n_ineq, n)
{
  f <- function(x) {
    original_parameters <- x[(n_ineq + 1):(n_ineq + n)]
    user_eq_jac <- eq_jac(original_parameters)
    # Construct the full augmented Jacobian: [ 0 | d(eq)/d(P) ]
    # Equality constraints do not depend on slack variables, so columns for s are zero.
    augmented_jac <- cbind(.zeros(n_eq, n_ineq), user_eq_jac)
    return(augmented_jac)
  }
  return(f)
}

augmented_inequality_jacobian <- function(ineq_jac, n_ineq, n)
{
  f <- function(x) {
    original_parameters <- x[(n_ineq + 1):(n_ineq + n)]
    user_ineq_jac <- ineq_jac(original_parameters)
    # Construct the full augmented Jacobian: [ d(ineq)/d(P) | d(ineq)/d(s) ]
    # where d(ineq)/d(s) is -I because the internal constraint is G_j(P,s) = ineq_j(P) - s_j = 0
    augmented_jac <- cbind(user_ineq_jac, -diag(n_ineq))
    return(augmented_jac)
  }
  return(f)
}


.solnp_ctrl <- function(control_list) {
  # Set default control parameters
  default_control <- list(
    rho = 1,        # Penalty parameter
    max_iter = 400,     # Maximum number of major iterations
    min_iter = 800,     # Maximum number of minor iterations
    tol = 1.0e-8,   # Tolerance on feasibility and optimality
    trace = FALSE   # Trace optimization progress
  )
  # Update defaults with user-provided control parameters
  for (name in names(control_list)) {
    if (name %in% names(default_control)) {
      default_control[[name]] <- control_list[[name]]
    } else {
      warning(paste("Unknown control parameter:", name))
    }
  }
  return(default_control)
}

.solver_warnings <- function(message_indicator)
{
  m1 <- paste0("\nsolnp: Redundant constraints were found.")
  m2 <- paste0("\nLinearized problem has no feasible solution.The problem may not be feasible.")
  m3 <- paste0("\nMinor optimization routine did not converge in the specified number of minor iterations.")
  ans <- switch(message_indicator,
               "M1" = m1,
               "M2" = m2,
               "M3" = m3)
  warning(ans)
}

.eps = .Machine$double.eps

.subnpmsg = function(m){
	g1 = c("solnp-->")
	m1 = paste("\n",g1, "Redundant constraints were found. Poor\n",
			g1, "intermediate results may result.Suggest that you\n",
			g1, "remove redundant constraints and re-OPTIMIZE\n", sep = "")
	m2 = paste("\n",g1, "The linearized problem has no feasible\n",
			g1, "solution.  The problem may not be feasible.\n", sep = "")
	m3 = paste("\n",g1, "Minor optimization routine did not converge in the \n",
			g1, "specified number of minor iterations.  You may need\n",
			g1, "to increase the number of minor iterations.        \n", sep = "")
	ans = switch(m,
			m1 = m1,
			m2 = m2,
			m3 = m3)
	cat(ans)
}



.checkpars = function(pars, LB, UB, .env)
{
	if(is.null(pars))
		stop("\nsolnp-->error: must supply starting parameters\n", call. = FALSE)
	if(!is.null(LB)){
		if(length(pars) != length(LB))
			stop("\nsolnp-->error: LB length not equal to parameter length\n", call. = FALSE)
		if(is.null(UB)) UB = rep(.Machine$double.xmax/2, length(LB))
	} else{
		LB = NULL
	}
	if(!is.null(UB)){
		if(length(pars) != length(UB))
			stop("\nsolnp-->error: UB length not equal to parameter length\n", call. = FALSE)
		if(is.null(LB)) LB = rep(-.Machine$double.xmax/2, length(UB))
	} else{
		UB = NULL
	}
	if(!is.null(UB) && any(LB > UB))
		stop("\nsolnp-->error: UB must be greater than LB\n", call. = FALSE)

	if(!is.null(UB) && any(LB == UB))
		warning("\nsolnp-->warning: Equal Lower/Upper Bounds Found. Consider\n
						excluding fixed parameters.\n", call. = FALSE)
	# deal with infinite values as these are not accepted by solve.QP
	if(!is.null(LB) && !any(is.finite(LB))){
		idx = which(!is.finite(LB))
		LB[idx] = sign(LB[idx])*.Machine$double.xmax/2
	}
	if(!is.null(UB) && !any(is.finite(UB))){
		idx = which(!is.finite(UB))
		UB[idx] = sign(UB[idx])*.Machine$double.xmax/2
	}
	assign(".LB", LB, envir = .env)
	assign(".UB", UB, envir = .env)
	return(1)
}

.checkfun = function(pars, fun, .env, ...)
{
	if(!is.function(fun)) stop("\nsolnp-->error: fun does not appear to be a function\n", call. = FALSE)
	val = fun(pars, ...)
	if(length(val) != 1)  stop("\nsolnp-->error: objective function returns value of length greater than 1!\n", call. = FALSE)

	assign(".solnp_fun", fun, envir = .env)
	ctmp = get(".solnp_nfn", envir =  .env)
	assign(".solnp_nfn", ctmp + 1, envir = .env)
	return(val)
}

# Might eventually use this, but really the user must take care of such problems
# in their own function/setup
.safefunx = function(pars, fun, .env, ...){
	xnames = get("xnames", envir = .env)
	names(pars) = xnames
	v  = fun(pars, ...)
	if(is.na(v) | !is.finite(v) | is.nan(v)) {
		warning(paste("\nsolnp-->warning: ", v , " detected in function call...check your function\n", sep = ""), immediate. = FALSE)
		v = 1e24
	}
	as.numeric(v)
}

.checkgrad = function(pars, fun, .env, ...)
{
	n = length(pars)
	val = fun(pars, ...)
	if(length(val)!=n)
		stop("\nsolnp-->error: gradient vector length must be equal to length(pars)\n", call. = FALSE)
	assign(".solnp_gradfun", fun, envir = .env)
	return(val)
}

.checkhess = function(pars, fun, .env, ...)
{
	n = length(pars)
	val = fun(pars, ...)
	if(length(as.vector(val)) != (n*n))
		stop("\nsolnp-->error: hessian must be of length length(pars) x length(pars)\n", call. = FALSE)
	assign(".solnp_hessfun", fun, envir = .env)
	return(val)
}

.checkineq = function(pars, fun, ineqLB, ineqUB, .env, ...)
{
	xnames = get("xnames", envir = .env)
	val = fun(pars, ...)
	n = length(val)
	if(!is.null(ineqLB)){
		if(length(ineqLB) != n)
			stop("\nsolnp-->error: inequality function returns vector of different length to
							inequality lower bounds\n", call. = FALSE)
	} else{
		stop("\nsolnp-->error: inequality function given without lower bounds\n", call. = FALSE)
	}
	if(!is.null(ineqUB)){
		if(length(ineqUB) != n)
			stop("\nsolnp-->error: inequality function returns vector of different length to
							inequality upper bounds\n", call. = FALSE)
	} else{
		stop("\nsolnp-->error: inequality function given without upper bounds\n", call. = FALSE)
	}
	if(any(ineqLB > ineqUB))
		stop("\nsolnp-->error: ineqUB must be greater than ineqLB\n", call. = FALSE)

	assign(".ineqLB", ineqLB, envir = .env)
	assign(".ineqUB", ineqUB, envir = .env)
	.solnp_ineqfun = function(x, ...){
		names(x) = xnames
		fun(x, ...)
	}
	assign(".solnp_ineqfun", .solnp_ineqfun, envir = .env)
	return(val)
}


.checkeq = function(pars, fun, eqB, .env, ...)
{
	xnames = get("xnames", envir = .env)
	n = length(eqB)
	val = fun(pars, ...) - eqB
	if(length(val)!=n)
		stop("\nsolnp-->error: equality function returns vector of different length
						to equality value\n", call. = FALSE)
	.eqB = eqB
	assign(".eqB", .eqB, envir = .env)
	.solnp_eqfun = function(x, ...){
		names(x) = xnames
		fun(x, ...) - .eqB
	}
	assign(".solnp_eqfun", .solnp_eqfun, envir = .env)
	return(val)
}


# check the jacobian of inequality
.cheqjacineq = function(pars, fun, .env,  ...)
{
	# must be a matrix -> nrows = no.inequalities, ncol = length(pars)
	val = fun(pars, ...)
	.ineqLB = get(".ineqLB", envir = .env)
	.ineqUB = get(".ineqUB", envir = .env)
	if(!is.matrix(val))
		stop("\nsolnp-->error: Jacobian of Inequality must return a matrix type object\n", call. = FALSE)
	nd = dim(val)
	if(nd[2] != length(pars))
		stop("\nsolnp-->error: Jacobian of Inequality column dimension must be equal to length
						of parameters\n", call. = FALSE)
	if(nd[1] != length(.ineqUB))
		stop("\nsolnp-->error: Jacobian of Inequality row dimension must be equal to length
						of inequality bounds vector\n", call. = FALSE)
	# as in inequality function, transforms from a 2 sided inequality to a one sided inequality
	# (for the jacobian).
	.solnp_ineqjac = function(x, ...) { retval = fun(x, ...); rbind( retval ) }
	assign(".solnp_ineqjac", .solnp_ineqjac, envir = .env)
	return(val)
}

# check the jacobian of equality
.cheqjaceq = function(pars, fun, .env, ...)
{
	# must be a matrix -> nrows = no.equalities, ncol = length(pars)
	val = fun(pars, ...)
	.eqB = get(".eqB", envir = .env)
	if(!is.matrix(val))
		stop("\nsolnp-->error: Jacobian of Equality must return a matrix type object\n", call. = FALSE)
	nd = dim(val)
	if(nd[2] != length(pars))
		stop("\nsolnp-->error: Jacobian of Equality column dimension must be equal to length of parameters\n", call. = FALSE)
	if(nd[1] != length(.eqB))
		stop("\nsolnp-->error: Jacobian of Equality row dimension must be equal to length of equality bounds vector\n", call. = FALSE)
	assign(".solnp_eqjac", fun, envir = .env)
	return(val)
}

# reporting function
.report = function(iter, funv, pars)
{
	cat( paste( "\nIter: ", iter ," fn: ", format(funv, digits = 4, scientific = 5, nsmall = 4, zero.print = TRUE), "\t Pars: ", sep=""),
			format(pars, digits = 4, scientific = 6, nsmall = 5, zero.print = TRUE) )
}

# finite difference gradient
.fdgrad = function(pars, fun, ...)
{
	if(!is.null(fun)){

		y0 = fun(pars, ...)
		nx = length(pars)
		grd = rep(0, nx)
		deltax = sqrt(.eps)
		for(i in 1:nx)
		{
			init = pars[i]
			pars[i]= pars[i] + deltax
			grd[i] = (fun(pars, ...) - y0) / deltax
			pars[i] = init
		}
	}
	else
	{
		grd = 0
	}
	return(grd)
}

# finite difference jacobian
.fdjac = function(pars, fun, ...)
{
	nx = length(pars)
	if(!is.null(fun))
	{
		y0 = fun(pars, ...)
		nf = length (y0)
		jac = matrix(0, nrow = nf, ncol= nx)
		deltax = sqrt (.eps)
		for(i  in 1:nx)
		{
			init = pars[i]
			pars[i]= pars[i] + deltax
			jac[,i] = (fun(pars, ...) - y0) / deltax
			pars[i] = init
		}
	} else{
		jac = rep(0, nx)
	}
	return(jac)
}

.emptygrad = function(pars, ...)
{
	matrix(0, nrow = 0, ncol = 1)
}

.emptyjac = function(pars, ...)
{
	#matrix(0, nrow = 0, ncol = length(pars))
	NULL
}

.emptyfun = function(pars, ...)
{
	NULL
}

.ineqlbfun = function(pars, .env, ...)
{
	LB = get(".solnp_LB", envir = .env)
	UB = get(".solnp_UB", envir = .env)
	.solnp_ineqfun = get(".solnp_ineqfun", envir = .env)
	res = c(pars - LB,  UB - pars)
	if(!is.null(.solnp_ineqfun)) res = c(.solnp_ineqfun(pars, ...), res)
	res
}

.ineqlbjac = function(pars, .env, ...)
{
	.solnp_ineqjac = get(".solnp_ineqjac", envir = .env)
	n = length(pars)
	res = rbind(diag(n), -diag(n))
	if(!is.null(.solnp_ineqjac)) res = rbind(.solnp_ineqjac(pars, ...), res)
	res
}

.solnpctrl = function(control){
	# parameters check is now case independent
	ans = list()
	params = unlist(control)
	if(is.null(params)) {
		ans$rho = 1
		ans$outer.iter = 400
		ans$inner.iter = 800
		ans$delta = 1.0e-7
		ans$tol = 1.0e-8
		ans$trace = 1
	} else{
		npar = tolower(names(unlist(control)))
		names(params) = npar
		if(any(substr(npar, 1, 3) == "rho")) ans$rho = as.numeric(params["rho"]) else ans$rho = 1
		if(any(substr(npar, 1, 10) == "outer.iter")) ans$outer.iter = as.numeric(params["outer.iter"]) else ans$outer.iter = 400
		if(any(substr(npar, 1, 10) == "inner.iter")) ans$inner.iter = as.numeric(params["inner.iter"]) else ans$inner.iter = 800
		if(any(substr(npar, 1, 5) == "delta")) ans$delta = as.numeric(params["delta"]) else ans$delta = 1.0e-7
		if(any(substr(npar, 1, 3) == "tol")) ans$tol = as.numeric(params["tol"]) else ans$tol = 1.0e-8
		if(any(substr(npar, 1, 5) == "trace")) ans$trace = as.numeric(params["trace"]) else ans$trace = 1
	}
	return(ans)
}

.zeros = function( n = 1, m = 1)
{
	if(missing(m)) m = n
	sol = matrix(0, nrow = n, ncol = m)
	return(sol)
}

.ones <- function(n = 1, m = 1)
{
	if(missing(m)) m = n
	sol = matrix(1, nrow = n, ncol = m)
	return(sol)
}

.vnorm <- function(x)
{
	sum((x)^2)^(1/2)
}

.solvecond = function(x)
{
	z = svd(x)$d
	if(any( z == 0 )) ret = Inf else ret = max( z ) / min( z )
	return(ret)
}

#' Multi-start version of csolnp
#'
#' Runs the csolnp solver from multiple diverse feasible starting values and
#' returns the best solution found.
#'
#' This function automates the process of generating multiple feasible starting
#' points (using lower, upper, and constraint information), runs `csolnp` from
#' each, and returns the solution with the lowest objective value. It is useful
#' for problems where local minima are a concern or the objective surface is
#' challenging.
#'
#' @param fn the objective function (must return a scalar).
#' @param gr an optional function for computing the analytic gradient of the
#' function (must return a vector of length n).
#' @param eq_fn an optional function for calculating equality constraints.
#' @param eq_b a vector of the equality bounds (if eq_fn provided).
#' @param eq_jac an optional function for computing the analytic Jacobian of the
#' equality function (a matrix with number of columns n and number of rows
#' equal to the number of equalities).
#' @param ineq_fn an optional function for calculating inequality constraints.
#' @param ineq_lower the lower bounds for the inequality constraints (must be finite).
#' @param ineq_upper the upper bounds for the inequality constraints (must be finite).
#' @param ineq_jac an optional function for computing the analytic Jacobian of the
#' inequality function (a matrix with number of columns n and number of rows equal
#' to the number of inequalities).
#' @param lower lower bounds for the parameters. This is strictly required.
#' @param upper upper bounds for the parameters. This is strictly required.
#' @param control a list of solver control parameters (see details).
#' @param n_candidates integer. The number of initial feasible candidate points
#' to generate for multi-start optimization. Default is 20.
#' @param penalty numeric. The penalty parameter used when projecting to feasibility
#' for candidate generation. Default is 1e4.
#' @param eq_tol Numeric. Tolerance for equality constraint violation
#' (default is 1e-6). Candidate solutions with \code{kkt_diagnostics\$eq_violation}
#' less than or equal to this value are considered feasible with respect
#' to equality constraints.
#' @param ineq_tol Numeric. Tolerance for inequality constraint violation
#' (default is 1e-6). Candidate solutions with \code{kkt_diagnostics\$ineq_violation}
#' less than or equal to this value are considered feasible with respect to
#' inequality constraints.
#' @param seed an optional random seed used to initialize the random number
#' generator for the random samples.
#' @param return_all logical. Whether to return all solutions as a list. This
#' may be useful for debugging.
#' @param ... additional arguments passed to the supplied functions
#' (common to all functions supplied).
#' @details
#' \strong{Candidate Generation:}
#'
#' The \code{generate_feasible_starts} approach creates a diverse set of initial
#' parameter vectors (candidates) that are feasible with respect to box and
#' (optionally) nonlinear constraints. The process is as follows:
#'
#' \enumerate{
#'   \item For each candidate, a random point is sampled inside the parameter
#'   box constraints (\code{lower} and \code{upper}) but a small distance away
#'   from the boundaries, to avoid numerical issues. This is achieved by a helper
#'   function that applies a user-specified buffer (\code{eps}).
#'   \item If nonlinear inequality constraints (\code{ineq_fn}) are provided,
#'   each sampled point is projected towards the feasible region using a
#'   fast penalized minimization. This step does not solve the feasibility problem
#'   exactly, but quickly produces a point that satisfies the constraints to
#'   within a specified tolerance, making it suitable as a starting point for
#'   optimization.
#'   \item If only box constraints are present, the sampled point is used
#'   directly as a feasible candidate.
#'   \item The set of feasible candidates is ranked by the objective with lower
#'   values considered better. This allows prioritization of candidates that start
#'   closer to optimality.
#' }
#'
#' This method efficiently creates a diverse set of robust initial values, improving
#' the chances that multi-start optimization will identify the global or a
#' high-quality local solution, especially in the presence of non-convexities or
#' challenging constraint boundaries.

#' \strong{Solution Selection:}
#' For each candidate starting point, \code{csolnp_ms} runs the \code{csolnp} solver
#' and collects the resulting solutions and their associated KKT diagnostics. If
#' equality or inequality constraints are present, candidate solutions are first
#' filtered to retain only those for which the maximum violation of
#' equality (\code{kkt_diagnostics\$eq_violation}) and/or
#' inequality (\code{kkt_diagnostics\$ineq_violation}) constraints are less than
#' or equal to user-specified tolerances (\code{eq_tol} and \code{ineq_tol}).
#' Among the feasible solutions (those satisfying all constraints within tolerance),
#' the solution with the lowest objective value is selected and returned as the
#' best result. If no candidate fully satisfies the constraints, the solution with
#' the smallest total constraint violation is returned, with a warning issued to
#' indicate that strict feasibility was not achieved. This two-stage selection
#' process ensures that the final result is both feasible (when possible) and
#' optimally minimizes the objective function among all feasible candidates.
#' @return
#' A list containing the best solution found by multi-start, with elements analogous
#' to those returned by \code{csolnp}. If \code{return_all} is TRUE, then a list of
#' all solutions is returned instead.
#' @seealso \code{\link{csolnp}}
#' @rdname csolnp_ms
#' @author Alexios Galanos
#' @export
csolnp_ms <- function(fn, gr = NULL, eq_fn = NULL, eq_b = NULL, eq_jac = NULL,
                      ineq_fn = NULL, ineq_lower = NULL, ineq_upper = NULL, ineq_jac = NULL,
                      lower = NULL, upper = NULL, control = list(), n_candidates = 20,
                      penalty = 1e4, eq_tol = 1e-6, ineq_tol = 1e-6, seed = NULL,
                      return_all = FALSE, ...)
{
    init <- generate_feasible_starts(n_candidates, fn, lower, upper, ineq_fn, ineq_lower, ineq_upper, eq_fn, eq_b,
                                     maxit = 100, penalty = penalty, eps = 1e-4, seed = seed)
    n <- NROW(init)
    sol <- future_lapply(1:n, function(i) {
        out <- try(csolnp(init[i,], fn = fn, gr = gr, eq_fn = eq_fn, eq_b = eq_b, eq_jac = eq_jac,
                          ineq_fn = ineq_fn, ineq_lower = ineq_lower, ineq_upper = ineq_upper, ineq_jac = ineq_jac,
                          lower = lower, upper = upper, control = control, ...), silent = TRUE)
        if (inherits(out, 'try-error')) return(NULL)
        return(out)
    }, future.seed = TRUE, future.packages = "Rsolnp")
    sol <- eval(sol)
    if (return_all) {
        return(sol)
    } else {
        # Remove NULL solutions up front
        sol <- sol[!sapply(sol, is.null)]
        if (length(sol) == 0) stop("No successful csolnp runs.")

        # Convergence filter as before
        convergence <- sapply(sol, function(x) if (is.null(x)) 10 else as.numeric(x$convergence))
        if (!any(convergence < 2)) {
            warning("No solution achieved convergence < 2. Results may be unreliable.")
        } else {
            sol <- sol[which(convergence < 2)]
        }
        # Get constraint violations
        has_eq <- !is.null(eq_fn)
        has_ineq <- !is.null(ineq_fn)

        eq_violation <- sapply(sol, function(x)
            if (!is.null(x$kkt_diagnostics$eq_violation)) as.numeric(x$kkt_diagnostics$eq_violation) else NA_real_)
        ineq_violation <- sapply(sol, function(x)
            if (!is.null(x$kkt_diagnostics$ineq_violation)) as.numeric(x$kkt_diagnostics$ineq_violation) else NA_real_)

        # Identify feasible candidates
        if (has_eq && has_ineq) {
            is_feasible <- (eq_violation <= eq_tol) & (ineq_violation <= ineq_tol)
        } else if (has_eq) {
            is_feasible <- (eq_violation <= eq_tol)
        } else if (has_ineq) {
            is_feasible <- (ineq_violation <= ineq_tol)
        } else {
            is_feasible <- rep(TRUE, length(sol)) # No constraints, all feasible
        }

        # Among feasible, choose lowest objective
        obj <- sapply(sol, function(x) if (is.null(x)) Inf else as.numeric(x$objective))

        if (any(is_feasible, na.rm=TRUE)) {
            idx <- which(is_feasible)
            min_index <- idx[which.min(obj[idx])]
            sol_best <- sol[[min_index]]
        } else {
            # No feasible solution, pick one with smallest total violation
            warning("No candidate fully satisfies constraints within tolerance. Returning least infeasible solution.")
            total_violation <- eq_violation
            if (has_eq && has_ineq) total_violation <- eq_violation + ineq_violation
            if (!has_eq && has_ineq) total_violation <- ineq_violation
            if (has_eq && !has_ineq) total_violation <- eq_violation
            min_index <- which.min(total_violation)
            sol_best <- sol[[min_index]]
        }
        return(sol_best)
    }
}

#' @rawNamespace useDynLib(Rsolnp, .registration=TRUE)
#' @keywords internal
#' @importFrom stats rnorm runif
#' @importFrom parallel parLapply clusterExport clusterEvalQ
#' @importFrom numDeriv grad jacobian hessian
#' @importFrom truncnorm rtruncnorm
#' @importFrom Rcpp evalCpp
#' @importFrom utils tail
#' @importFrom stats optim
#' @importFrom future.apply future_lapply
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

.rsolnp <- function(pars, fn, gr = NULL, eq_fn = NULL, eq_b = NULL, eq_jac = NULL,
         ineq_fn = NULL, ineq_lower = NULL, ineq_upper = NULL,
         ineq_jac = NULL, lower = NULL, upper = NULL, control = list(), ...)
{
    # Start timer for performance measurement
    start_time <- Sys.time()
    # Store original parameter names
    parameter_names <- names(pars)

    if (is.null(lower)) {
        lower <- -1000 * abs(pars)
        warning("\nlower values are NULL. Setting to -1000 x abs(pars).")
    }

    if (is.null(upper)) {
        upper <- 1000 * abs(pars)
        warning("\nupper values are NULL. Setting to 1000 x abs(pars).")
    }

    nudge_factor <- 1e-5

    # Clamp parameters at or below lower bound
    idx_lower <- which(pars <= lower)
    if (length(idx_lower) > 0) {
        tmp_p <- lower[idx_lower]
        # For strict zero bounds, nudge slightly up
        tmp_p[tmp_p == 0] <- 1e-12
        # Nudge up by a tiny fraction
        tmp_p <- tmp_p + abs(tmp_p) * nudge_factor
        pars[idx_lower] <- tmp_p
        warning("\nsome parameters at lower bounds. Adjusting by a factor of 1e-5.")
    }

    # Clamp parameters at or above upper bound
    idx_upper <- which(pars >= upper)
    if (length(idx_upper) > 0) {
        tmp_p <- upper[idx_upper]
        # For strict zero upper bounds (rare), nudge slightly down
        tmp_p[tmp_p == 0] <- -1e-12
        # Nudge down by a tiny fraction
        tmp_p <- tmp_p - abs(tmp_p) * nudge_factor
        pars[idx_upper] <- tmp_p
        warning("\nsome parameters at upper bounds. Adjusting by a factor of 1e-5.")
    }

    # Create a dedicated environment for csolnp's internal state
    .solnp_environment <- environment()
    assign("parameter_names", parameter_names, envir = .solnp_environment)
    # Initialize function call and error counters
    assign(".solnp_nfn", 0, envir = .solnp_environment)
    assign(".solnp_errors", 0, envir = .solnp_environment)
    # Indicator vector for problem characteristics
    # [1] Number of parameters (np)
    # [4] Has inequality constraints
    # [5] Number of inequality constraints
    # [7] Has equality constraints
    # [8] Number of equality constraints
    # [10] Has parameter upper/lower bounds
    # [11] Has any bounds (parameter or inequality)
    problem_indicators <- solnp_problem_setup(pars, fn, gr, eq_fn, eq_b, eq_jac, ineq_fn, ineq_lower, ineq_upper, ineq_jac, lower, upper, ...)
    num_parameters <- problem_indicators[1]
    objective_list <- solnp_objective_wrapper(fn = fn, gr = gr, n = num_parameters, ...)
    inequality_list <- solnp_ineq_wrapper(ineq_fn, ineq_jac, n_ineq = problem_indicators[5], n = problem_indicators[1], ...)
    equality_list <- solnp_eq_wrapper(eq_fn, eq_b, eq_jac, n_eq = problem_indicators[8], n = problem_indicators[1], ...)
    lower_tmp <- lower
    upper_tmp <- upper
    pars_tmp <- pars
    objective_fun <- objective_list$objective_fun
    gradient_fun <- objective_list$gradient_fun

    # Objective function checks and initial evaluation
    ineq_f <- inequality_list$ineq_f
    ineq_j <- inequality_list$ineq_j
    eq_f <- equality_list$eq_f
    eq_j <- equality_list$eq_j
    aug_fn_gr <- augmented_function_gradient(gr = objective_list$gradient_fun, n_ineq = problem_indicators[5], n = problem_indicators[1])
    aug_ineq_jac <- aug_eq_jac <- NULL
    if (!is.null(ineq_fn)) {
        aug_ineq_jac <- augmented_inequality_jacobian(ineq_jac = inequality_list$ineq_j, n_ineq = problem_indicators[5], n = problem_indicators[1])
    } else {
        aug_ineq_jac <- function(pars) {return(NULL)}
    }
    if (!is.null(eq_fn)) {
        aug_eq_jac <- augmented_equality_jacobian(eq_jac = equality_list$eq_j, n_eq = problem_indicators[8], n_ineq = problem_indicators[5], n = problem_indicators[1])
    } else {
        aug_eq_jac <- function(pars) {return(NULL)}
    }
    # Inequality constraint checks and initial evaluation
    initial_ineq_values <- NULL
    initial_ineq_guess <- NULL

    if (!is.null(ineq_f)) {
        initial_ineq_values <- ineq_f(pars_tmp)
        initial_ineq_guess <- (ineq_lower + ineq_upper) / 2
    }
    # Equality constraint checks and initial evaluation
    initial_eq_values <- NULL
    if (!is.null(eq_f)) {
        initial_eq_values <- eq_f(pars_tmp)
    }

    initial_objective_value <- objective_fun(pars_tmp)
    # Combine inequality and parameter bounds into a single matrix
    all_bounds <- rbind(cbind(ineq_lower, ineq_upper),
                        cbind(lower_tmp, upper_tmp))

    # Parse and apply control parameters
    control_params <- .solnp_ctrl(control)
    penalty_param <- control_params$rho
    max_major_iterations <- control_params$max_iter
    max_minor_iterations <- control_params$min_iter
    tolerance <- control_params$tol
    trace_progress <- control_params$trace
    # Calculate total number of constraints
    n_ineq <- problem_indicators[5]
    n_eq <- problem_indicators[8]
    total_constraints <- n_ineq + n_eq

    # Initialize objective function values and internal status vector
    current_objective_value <- previous_objective_value <- initial_objective_value
    status_vector <- 0 * .ones(3, 1) # [obj_change, constraint_norm_prev, constraint_norm_curr]


    # Initialize Lagrange multipliers and constraints
    if (total_constraints > 0) {
        lagrange_mults <- 0 * .ones(total_constraints, 1)
        # Concatenate equality and inequality constraint values
        constraint_values <- c(initial_eq_values, initial_ineq_values)
        # Adjust inequality constraint values based on initial guess
        if (problem_indicators[4]) {
            # Calculate slack for inequality constraints relative to bounds
            temp_ineq_slack_lb <- constraint_values[(n_eq + 1):total_constraints] - ineq_lower
            temp_ineq_slack_ub <- ineq_upper - constraint_values[(n_eq + 1):total_constraints]
            # Check if all initial inequalities are within bounds (positive slack)
            min_slack <- apply(cbind(temp_ineq_slack_lb, temp_ineq_slack_ub), 1, min)
            if (all(min_slack > 0)) {
                initial_ineq_guess <- constraint_values[(n_eq + 1):total_constraints]
            }
            # Adjust inequality constraint values by subtracting the initial guess
            constraint_values[(n_eq + 1):total_constraints] <-
                constraint_values[(n_eq + 1):total_constraints] - initial_ineq_guess
        }
        # Store the initial norm of the constraint violations
        status_vector[2] <- .vnorm(constraint_values)

        # If initial constraints are "small" (within 10*tolerance), set penalty to 0
        if (max(status_vector[2] - 10 * tolerance, n_ineq, na.rm = TRUE) <= 0) {
            penalty_param <- 0
        }
    } else {
        lagrange_mults <- 0
    }

    # Starting augmented parameter vector (initial_ineq_guess for inequalities, then original parameters)
    augmented_parameters <- c(initial_ineq_guess, pars_tmp)

    # Initialize Hessian approximation (identity matrix for augmented parameters)
    augmented_hessian <- diag(num_parameters + n_ineq)

    # Initialize mu (related to optimality conditions, often small)
    lambda <- num_parameters # Or some other initial value

    # Store initial objective and constraint values
    scaled_eval <- c(initial_objective_value, initial_eq_values, initial_ineq_values)

    # Initialize major iteration counter
    major_iteration_count <- 0

    # Store historical objective values
    historical_objective_values <- c(initial_objective_value)

    penalty_param <- control_params$rho
    tol <- control_params$tol
    min_iter <- control_params$min_iter
    trace <- control_params$trace
    # Main optimization loop (Major Iterations)
    while (major_iteration_count < max_major_iterations) {
        ctrl <- list(rho = penalty_param, min_iter = min_iter, tol = tol, trace = trace)
        major_iteration_count <- major_iteration_count + 1
        assign(".solnp_iter", major_iteration_count, envir = .solnp_environment) # Update global iteration count
        # Control parameters for the inner optimization (.subnp)
        #
        # Determine scaling factors for objective, constraints, and parameters
        objective_and_eq_scale <- NULL
        if (problem_indicators[7]) { # If equality constraints exist
            # Scale objective by its value, and equality constraints by max absolute value
            objective_and_eq_scale <- c(scaled_eval[1], rep(1, n_eq) * max(abs(scaled_eval[2:(n_eq + 1)])))
        } else {
            objective_and_eq_scale <- 1 # No scaling for equality constraints
        }

        if (!problem_indicators[11]) { # If no parameter or inequality bounds
            scaling_factors <- c(objective_and_eq_scale, augmented_parameters)
        } else {
            # If bounds exist, scale parameters by 1 (or a default value)
            scaling_factors <- c(objective_and_eq_scale, rep(1, length(augmented_parameters)))
        }

        # Ensure scaling factors are not too small or too large
        scaling_factors <- apply(matrix(scaling_factors, ncol = 1), 1,
                                 FUN = function(x) min(max(abs(x), tolerance), 1/tolerance))
        L <- list()
        L$objective_fun <- objective_fun
        L$equality_fun <- eq_f
        L$inequality_fun <- ineq_f
        L$augmented_gradient <- aug_fn_gr
        L$augmented_inequality_jac <- aug_ineq_jac
        L$augmented_equality_jac <- aug_eq_jac
        L$gradient_fun <- gradient_fun
        L$ineq_jac <- ineq_j
        L$eq_jac <- eq_j
        L$lower <- lower_tmp
        L$upper <- upper_tmp
        L$inequality_lower <- ineq_lower
        L$inequality_upper <- ineq_upper
        L$problem_indicators <- problem_indicators
        .env <- new.env()
        subnp_results <- .rsubnp(pars = augmented_parameters, lagrange_mults = lagrange_mults, scaled_eval = scaled_eval,
                                augmented_hessian = augmented_hessian, lambda = lambda, scaling_factors = scaling_factors,
                                ctrl = ctrl, fun_list = L, .env = .solnp_environment)
        if (get(".solnp_errors", envir = .solnp_environment) == 1) {
            max_major_iterations <- major_iteration_count # Terminate major iterations
        }

        # Update parameters, Lagrange multipliers, Hessian, and mu from subnp results
        augmented_parameters <- subnp_results$p
        lagrange_mults <- subnp_results$y
        augmented_hessian <- subnp_results$augmented_hessian
        lambda <- subnp_results$lambda # mu is returned as lambda from subnp

        # Extract original parameters from augmented parameters
        current_parameters <- augmented_parameters[(n_ineq + 1):(n_ineq + num_parameters)]

        # Re-evaluate objective function with updated parameters
        current_objective_value <- objective_fun(current_parameters)

        # Report progress if tracing is enabled
        if (trace_progress) {
            .report(major_iteration_count, current_objective_value, current_parameters)
        }

        # Re-evaluate constraint functions with updated parameters
        current_eq_values <- eq_f(current_parameters)
        current_ineq_values <- ineq_f(current_parameters)

        # Update combined objective and constraint values
        scaled_eval <- c(current_objective_value, current_eq_values, current_ineq_values)

        # Calculate relative change in objective function
        status_vector[1] <- (previous_objective_value - scaled_eval[1]) / max(abs(scaled_eval[1]), 1)
        previous_objective_value <- scaled_eval[1]

        # Update constraint violation norms and penalty parameter
        if (total_constraints > 0) {
            current_constraint_violations <- scaled_eval[2:(total_constraints + 1)]

            if (problem_indicators[4]) {
                # Recalculate slack for inequality constraints relative to bounds
                temp_ineq_slack_lb <- current_constraint_violations[(n_eq + 1):total_constraints] - all_bounds[1:n_ineq, 1]
                temp_ineq_slack_ub <- all_bounds[1:n_ineq, 2] - current_constraint_violations[(n_eq + 1):total_constraints]
                # If inequalities are within bounds, update the initial guess for inequalities in augmented parameters
                if (min(c(temp_ineq_slack_lb, temp_ineq_slack_ub)) > 0) {
                    augmented_parameters[1:n_ineq] <- current_constraint_violations[(n_eq + 1):total_constraints]
                }
                # Re-adjust inequality constraint values by subtracting the current guess for inequalities
                current_constraint_violations[(n_eq + 1):total_constraints] <-
                    current_constraint_violations[(n_eq + 1):total_constraints] - augmented_parameters[1:n_ineq]
            }
            # Calculate current norm of constraint violations
            status_vector[3] <- .vnorm(current_constraint_violations)

            # Adjust penalty parameter based on constraint violation norms
            if (status_vector[3] < 10 * tolerance) {
                penalty_param <- 0
                lambda <- min(lambda, tolerance)
            }

            if (status_vector[3] < 5 * status_vector[2]) {
                penalty_param <- penalty_param / 5
            }

            if (status_vector[3] > 10 * status_vector[2]) {
                penalty_param <- 5 * max(penalty_param, sqrt(tolerance))
            }

            # If changes in objective and constraints are small, reset Lagrange multipliers and Hessian
            if (max(c(tolerance + status_vector[1], status_vector[2] - status_vector[3])) <= 0) {
                lagrange_mults <- 0
                augmented_hessian <- diag(diag(augmented_hessian)) # Reset Hessian to diagonal
            }
            status_vector[2] <- status_vector[3] # Update previous constraint norm
        }

        # Check for convergence based on combined norms of objective change and constraint violations
        if (.vnorm(c(status_vector[1], status_vector[2])) <= tolerance) {
            max_major_iterations <- major_iteration_count # Terminate major iterations
        }

        # Store current objective value for historical tracking
        historical_objective_values <- c(historical_objective_values, current_objective_value)
    }

    # Final adjustments and results preparation
    if (problem_indicators[4]) { # If inequality constraints were present
        initial_ineq_guess <- augmented_parameters[1:n_ineq]
    }

    # Extract the optimized parameters (excluding initial inequality guess)
    optimized_parameters <- augmented_parameters[(n_ineq + 1):(n_ineq + num_parameters)]
    # Determine convergence status
    convergence_status <- 0
    if (get(".solnp_errors", envir = .solnp_environment) == 1) {
        convergence_status <- 2 # Solution not reliable due to Hessian inversion problems
        if (trace_progress) {
            cat("\nsolnp: Solution not reliable....Problem Inverting Hessian.\n")
        }
    } else {
        if (.vnorm(c(status_vector[1], status_vector[2])) <= tolerance) {
            convergence_status <- 0 # Converged successfully
            if (trace_progress) {
                cat(paste("\nsolnp: Completed in ", major_iteration_count, " iterations\n", sep = ""))
            }
        } else {
            convergence_status <- 1 # Exited after maximum iterations, tolerance not achieved
            if (trace_progress) {
                cat(paste("\nsolnp: Exiting after maximum number of iterations. Tolerance not achieved\n", sep = ""))
            }
        }
    }
    # Calculate elapsed time and retrieve final function evaluation count
    final_function_evaluations <- get(".solnp_nfn", envir = .solnp_environment)
    elapsed_time <- Sys.time() - start_time

    # Prepare results list
    results <- list(
        pars = optimized_parameters,
        convergence = convergence_status,
        objective = tail(as.numeric(historical_objective_values), 1),
        objective_history = as.numeric(historical_objective_values),
        lagrange = lagrange_mults,
        hessian = augmented_hessian,
        ineq_initial_values = initial_ineq_guess,
        n_eval = final_function_evaluations,
        outer_iterations = major_iteration_count,
        elapsed_time = elapsed_time,
        vscale = scaling_factors
    )
    return(results)
}

.rsubnp <- function(pars, lagrange_mults, scaled_eval, augmented_hessian, lambda, scaling_factors, ctrl, fun_list, .env, ...) {
    .solnp_fun <- fun_list$objective_fun      # J(P)
    .solnp_eqfun <- fun_list$equality_fun       # EC(P)
    .solnp_ineqfun <- fun_list$inequality_fun   # IC(P)
    .solnp_gradfun <- fun_list$gradient_fun
    .solnp_eqjac   <- fun_list$eq_jac
    .solnp_ineqjac <- fun_list$ineq_jac
    lower <- fun_list$lower
    upper <- fun_list$upper
    ineq_lower <- fun_list$inequality_lower
    ineq_upper <- fun_list$inequality_upper
    setup <- fun_list$problem_indicators

    # pars [nineq + np]
    rho <- ctrl$rho
    maxit <- ctrl$min_iter
    tol <- ctrl$tol
    trace <- ctrl$trace
    # [1] length of pars
    # [2] has function gradient?
    # [3] has hessian?
    # [4] has ineq?
    # [5] ineq length
    # [6] has jacobian (inequality)
    # [7] has eq?
    # [8] eq length
    # [9] has jacobian (equality)
    # [10] has upper / lower bounds
    # [11] has either lower/upper bounds or ineq


    n_eq <- setup[8]
    n_ineq <- setup[ 5 ]
    n_pars <- setup[ 1 ]
    status_flag <- 1
    line_search_steps <- c(0,0,0)
    n_constraints <- n_eq + n_ineq
    n_pic <- n_pars + n_ineq
    working_params <- pars
    # param_bounds [ 2 x (nineq + np) ]
    param_bounds <- rbind(cbind(ineq_lower, ineq_upper), cbind(lower, upper))
    step_objectives <- numeric()
    param_trials <- matrix()
    bfgs_scaling_factors <- numeric()

    if (n_constraints == 0) {
        augmented_jacobian <- matrix(0, 0, n_pars)  # avoid fake constraint
    }
    # scale the cost, the equality constraints, the inequality constraints,
    # the parameters (inequality parameters AND actual parameters),
    # and the parameter bounds if there are any
    # Also make sure the parameters are no larger than (1-tol) times their bounds
    # scaled_eval [ 1 neq nineq]

    scaled_eval <- scaled_eval / scaling_factors[1:(n_constraints + 1)]
    # working_params [np]
    working_params <- working_params / scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]
    if (setup[11] > 0) {
        if (setup[10] == 0) {
            mm <- n_ineq
        } else {
            mm <- n_pic
        }
        param_bounds <- param_bounds / cbind(scaling_factors[(n_eq + 2):(n_eq + mm + 1)], scaling_factors[(n_eq + 2):(n_eq + mm + 1)])
    }

    # scale the lagrange multipliers and the Hessian
    if (n_constraints > 0) {
        # lagrange_mults [total constraints = nineq + neq]
        # scale here is [tc] and dot multiplied by lagrange_mults
        lagrange_mults <- scaling_factors[2:(n_constraints + 1)] * lagrange_mults / scaling_factors[1]
    }
    # lagrange_mults = [zeros 3x1]

    # h is [ (np+nineq) x (np+nineq) ]
    #columnvector %*% row vector (size h) then dotproduct h then dotdivide scale[1]
    augmented_hessian <- augmented_hessian * (scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)] %*% t(scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)])) / scaling_factors[1]
    # h[ 8x8 eye]
    j <- scaled_eval[1]
    if (setup[7] > 0 && setup[4] > 0) {
        # Both equality and inequality constraints
        Aeq_block   <- cbind(matrix(0, nrow = n_eq, ncol = n_ineq), matrix(0, nrow = n_eq, ncol = n_pars))
        Aineq_block <- cbind(-diag(n_ineq), matrix(0, nrow = n_ineq, ncol = n_pars))
        augmented_jacobian <- rbind(Aeq_block, Aineq_block)
    } else if (setup[4] > 0) {
        # Only inequality constraints
        augmented_jacobian <- cbind(-diag(n_ineq), matrix(0, nrow = n_ineq, ncol = n_pars))
    } else if (setup[7] > 0) {
        # Only equality constraints
        augmented_jacobian <- cbind(matrix(0, nrow = n_eq, ncol = n_ineq), matrix(0, nrow = n_eq, ncol = n_pars))
    } else {
        # No constraints — explicitly handle this
        augmented_jacobian <- matrix(0, 0, n_ineq + n_pars)
    }
    # gradient
    grad_augmented_obj <- 0 * .ones(n_pic, 1)
    p <- working_params[1:n_pic]
    if (n_constraints > 0) {
        # [ nc ]
        constraint <- scaled_eval[2:(n_constraints + 1)]
        # constraint [5 0 11 3x1]
        # gradient routine
        x <- working_params[(n_ineq + 1):n_pic] * scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)]
        if (setup[7] > 0) {
            Aeq <- .solnp_eqjac(x)
            Aeq <- Aeq * outer(1 / scaling_factors[2:(n_eq + 1)], scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)])
            augmented_jacobian[1:n_eq, (n_ineq + 1):n_pic] <- Aeq
        }
        if (setup[4] > 0) {
            Aineq <- .solnp_ineqjac(x)
            Aineq <- Aineq * outer(1 / scaling_factors[(n_eq + 2):(n_constraints + 1)], scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)])
            augmented_jacobian[(n_eq + 1):n_constraints, (n_ineq + 1):n_pic] <- Aineq
            augmented_jacobian[(n_eq + 1):n_constraints, 1:n_ineq] <- -diag(n_ineq)
        }
        g_user <- .solnp_gradfun(x)
        g_user <- g_user / scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)] / scaling_factors[1]
        grad_augmented_obj[(n_ineq + 1):n_pic, 1] <- g_user


        if (setup[4] > 0) {
            constraint[(n_eq + 1):(n_eq + n_ineq)] <- constraint[(n_eq + 1):(n_eq + n_ineq)] - working_params[1:n_ineq]
        }
        # solver messages
        if (.solvecond(augmented_jacobian) > 1 / .eps ) {
            if (trace) .solver_warnings("M1")
        }

        # augmented_jacobian(matrix) x columnvector - columnvector
        # constraint_offset [nc,1]
        constraint_offset <- augmented_jacobian %*% working_params - constraint
        status_flag <- -1
        line_search_steps[1] = tol - max(abs(constraint))

        if (line_search_steps[1] <= 0) {
            status_flag <- 1
            if (setup[11] == 0) {
                # augmented_jacobian %*% t(augmented_jacobian) gives [nc x nc]
                # t(augmented_jacobian) %*% above gives [(np+nc) x 1]
                working_params <- working_params - t(augmented_jacobian) %*% solve(augmented_jacobian %*% t(augmented_jacobian), constraint)
                line_search_steps[1] <- 1
            }
        }

        if (line_search_steps[1] <= 0) {
            # this expands working_params to [nc+np+1]
            working_params[n_pic + 1] <- 1
            augmented_jacobian <- cbind(augmented_jacobian, -constraint)
            # cx is rowvector
            cx <- cbind(.zeros(1, n_pic), 1)
            dx <- .ones(n_pic + 1, 1)
            step_interval_width <- 1
            minit <- 0
            while (step_interval_width >= tol) {
                minit <- minit + 1
                # gap [(nc + np) x 2]
                gap <- cbind(working_params[1:mm] - param_bounds[, 1], param_bounds[, 2] - working_params[1:mm])
                # this sorts every row
                gap <- t(apply(gap, 1, FUN = function(x) sort(x)))
                dx[1:mm] <- gap[, 1]
                # expand dx by 1
                dx[n_pic + 1] <- working_params[n_pic + 1]
                if (setup[10] == 0) {
                    dx[(mm + 1):n_pic] <- max(c(dx[1:mm], 100)) * .ones(n_pic - mm, 1)
                }
                # t( augmented_jacobian %*% diag( as.numeric(dx) ) ) gives [(np+nc + 1 (or more) x nc]
                # dx * t(cx) dot product of columnvectors
                # qr.solve returns [nc x 1]
                y <- try(qr.solve(t(augmented_jacobian %*% diag(as.numeric(dx), length(dx), length(dx))), dx * t(cx)), silent = TRUE)
                if (inherits(y, "try-error")) {
                    p <- working_params * scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]  # unscale the parameter vector
                    if (n_constraints > 0) {
                        y <- 0 # unscale the lagrange multipliers
                    }
                    augmented_hessian <- scaling_factors[1] * augmented_hessian / (scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)] %*% t(scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1) ]))
                    ans = list(p = p, y = y, augmented_hessian = augmented_hessian, lambda = lambda)
                    assign(".solnp_errors", 1, envir = .env)
                    return(ans)
                }
                    v <- dx * (dx * (t(cx) - t(augmented_jacobian) %*% y))
                if (v[n_pic + 1] > 0 ) {
                    z <- working_params[n_pic + 1] / v[n_pic + 1]
                    for (i in 1:mm) {
                        if (v[i] < 0) {
                            z <- min(z, -(param_bounds[i, 2] - working_params[i]) / v[i])
                        } else if (v[i] > 0) {
                            z <- min(z, (working_params[i] - param_bounds[i , 1]) / v[i])
                        }
                    }
                    if (z >= working_params[n_pic + 1] / v[n_pic + 1]) {
                        working_params <- working_params - z * v
                    } else {
                        working_params <- working_params - 0.9 * z * v
                    }
                    step_interval_width <- working_params[n_pic + 1]
                    if (minit >= 10) {
                        step_interval_width <- 0
                    }
                } else {
                    step_interval_width <- 0
                    minit <- 10
                }
            }
            if (minit >= 10) {
                if (trace) .solver_warnings("M2")
            }
            augmented_jacobian <- matrix(augmented_jacobian[, 1:n_pic], ncol = n_pic)
            constraint_offset <- augmented_jacobian %*% working_params[1:n_pic]
        }
    }

    p <- working_params[1:n_pic]
    y <- 0

    if (status_flag > 0) {
        tmp_value <- p[(n_ineq + 1):n_pic] * scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)]
        fun_value <- .solnp_fun(tmp_value)
        eq_value <- .solnp_eqfun(tmp_value)
        ineq_value <- .solnp_ineqfun(tmp_value)
        ctmp <- get(".solnp_nfn", envir =  .env)
        assign(".solnp_nfn", ctmp + 1, envir = .env)
        scaled_eval <- c(fun_value, eq_value, ineq_value) / scaling_factors[1:(n_constraints + 1)]
    }
    j <- scaled_eval[1]

    if (setup[4] > 0) {
        scaled_eval[(n_eq + 2):(n_constraints + 1)] <- scaled_eval[(n_eq + 2):(n_constraints + 1)] - p[1:n_ineq]
    }

    if (n_constraints > 0) {
        scaled_eval[2:(n_constraints + 1)] <- scaled_eval[2:(n_constraints + 1)] - augmented_jacobian %*% p + constraint_offset
        j <- scaled_eval[1] - t(lagrange_mults) %*% matrix(scaled_eval[2:(n_constraints + 1)], ncol = 1) + rho * .vnorm(scaled_eval[2:(n_constraints + 1)])^2
    }

    minit <- 0
    delta_position <- rep(0, n_pic)   # <- NEW
    delta_gradient <- rep(0, n_pic)   # <- NEW
    reduce <- Inf
    while (minit < maxit) {
        minit <- minit + 1
        if (status_flag > 0) {
            # Unscale actual parameters
            Alin_x <- augmented_jacobian[ , (n_ineq + 1):(n_ineq + n_pars), drop = FALSE]
            x <- p[(n_ineq + 1):n_pic] * scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)]
            # Evaluate f(x) gradient
            fun_value <- .solnp_fun(x)
            ineq_value <- .solnp_ineqfun(x)
            eq_value <- .solnp_eqfun(x)
            grad_f <- .solnp_gradfun(x)
            grad_f.p <- grad_f * scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)]
            # Evaluate Jacobians (they may return NULL)
            obm <- c(fun_value, eq_value, ineq_value) / scaling_factors[1:(n_constraints + 1)]

            Aeq.j   <- matrix(0, 0, n_pars)
            Aineq.j <- matrix(0, 0, n_pars)
            gradx <- grad_f.p / scaling_factors[1]   # only ∇f term
            grads <- numeric(0)

            # Scale Jacobians
            if (n_eq > 0) {
                Aeq.raw <- .solnp_eqjac(x)
                if (!is.null(Aeq.raw)) {
                    #Aeq <- Aeq * outer(1 / scaling_factors[2:(n_eq + 1)], scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)])
                    Aeq.p <- sweep(Aeq.raw, 2, scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)], '*')
                    Aeq.j <- sweep(Aeq.p, 1, scaling_factors[2:(n_eq + 1)], '/')
                }
            }

            if (n_ineq) {
                Aineq.raw <- .solnp_ineqjac(x)
                if (!is.null(Aineq.raw)) {
                    Aineq.p <- sweep(Aineq.raw, 2, scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)], '*')
                    Aineq.j <- sweep(Aineq.p, 1, scaling_factors[(n_eq + 2):(n_constraints + 1)], '/')
                }
            }

            J <- rbind(Aeq.j, Aineq.j) - Alin_x

            if (n_constraints > 0) {
                ## residual vector r  (length n_constraints)
                r <- obm[2:(n_constraints + 1)]

                if (n_ineq > 0) {
                    r[(n_eq + 1):n_constraints] <- r[(n_eq + 1):n_constraints] - p[1:n_ineq]
                }
                r <- r - augmented_jacobian %*% p + constraint_offset

                ## multiplier & penalty pieces (only if J has rows)
                if (nrow(J) > 0) {
                    JTy <- drop(crossprod(J, lagrange_mults))   # length n_pars
                    JTr <- drop(crossprod(J, r))                # length n_pars
                } else {
                    JTy <- 0
                    JTr <- 0
                }
                gradx <- gradx - JTy + 2 * rho * JTr
                ## slack variables
                if (n_ineq > 0) {
                    idx   <- (n_eq + 1):n_constraints
                    rg    <- r[idx]                   # g̃ − s
                    grads <- lagrange_mults[idx] - 2 * rho * rg
                }
            }
            if (n_ineq > 0)  grad_augmented_obj[1:n_ineq] <- grads
            grad_augmented_obj[(n_ineq + 1):(n_ineq + n_pars)] <- gradx
        }
        if (setup[4] > 0) {
            grad_augmented_obj[1:n_ineq] <- 0
        }

        if (minit > 1) {
            delta_gradient <- grad_augmented_obj - delta_gradient
            delta_position <- p - delta_position
            bfgs_scaling_factors[1] <- t(delta_position) %*% augmented_hessian %*% delta_position
            bfgs_scaling_factors[2] <- t(delta_position) %*% delta_gradient
            if ((bfgs_scaling_factors[1] * bfgs_scaling_factors[2]) > 0) {
                delta_position <- augmented_hessian %*% delta_position
                augmented_hessian <- augmented_hessian - (delta_position %*% t(delta_position)) / bfgs_scaling_factors[1] + (delta_gradient %*% t(delta_gradient)) / bfgs_scaling_factors[2]
            }
        }

        dx <- 0.01 * .ones(n_pic, 1)
        if (setup[11] > 0) {
            gap <- cbind(p[1:mm] - param_bounds[, 1], param_bounds[, 2] - p[1:mm])
            gap <- t(apply(gap, 1, FUN = function(x) sort(x)))
            gap <- gap[, 1] + sqrt(.eps) * .ones(mm, 1)
            dx[1:mm, 1] <- .ones(mm, 1) / gap
            if (setup[10] == 0) {
                dx[(mm + 1):n_pic, 1] <- min(c(dx[1:mm, 1], 0.01)) * .ones(n_pic - mm, 1)
            }
        }
        step_interval_width <- -1
        lambda <- lambda / 10

        while (step_interval_width <= 0) {
            cz <- try(chol(augmented_hessian + lambda * diag(as.numeric(dx * dx), length(dx), length(dx))), silent = TRUE)
            if (inherits(cz, "try-error")) {
                p <- p * scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]  # unscale the parameter vector
                if (n_constraints > 0) {
                    y <- 0 # unscale the lagrange multipliers
                }
                augmented_hessian <- scaling_factors[1] * augmented_hessian / (scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)] %*% t(scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]))
                ans <- list(p = p, y = y, augmented_hessian = augmented_hessian, lambda = lambda)
                assign(".solnp_errors", 1, envir = .env)
                return(ans)
            }
            cz <- try(solve(cz), silent = TRUE)
            if (inherits(cz, "try-error")) {
                p <- p * scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]  # unscale the parameter vector
                if (n_constraints > 0) {
                    y <- 0 # unscale the lagrange multipliers
                }
                augmented_hessian <- scaling_factors[1] * augmented_hessian / (scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)] %*% t(scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]))
                ans <- list(p = p, y = y, augmented_hessian = augmented_hessian, lambda = lambda)
                assign(".solnp_errors", 1, envir = .env)
                return(ans)
            }
            delta_gradient <- t(cz) %*% grad_augmented_obj

            if (n_constraints == 0 || nrow(augmented_jacobian) == 0) {
                u <- -cz %*% delta_gradient
            } else {
                y <- try(qr.solve(t(cz) %*% t(augmented_jacobian), delta_gradient), silent = TRUE)
                if (inherits(y, "try-error")) {
                    p <- p * scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]  # unscale the parameter vector
                    if (n_constraints > 0) {
                        # y = scaling_factors[ 1 ] * y / scaling_factors[ 2:(nc + 1) ] # unscale the lagrange multipliers
                        y <- 0
                    }
                    augmented_hessian <- scaling_factors[1] * augmented_hessian / (scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)] %*% t(scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]))
                    ans <- list(p = p, y = y, augmented_hessian = augmented_hessian, lambda = lambda)
                    assign(".solnp_errors", 1, envir = .env)
                    return(ans)
                }
                u <- -cz %*% (delta_gradient - (t(cz) %*% t(augmented_jacobian)) %*% y)
            }
            working_params <- u[1:n_pic] + p

            if (setup[11] == 0) {
                step_interval_width <- 1
            } else {
                step_interval_width <- min(c(working_params[1:mm] - param_bounds[, 1], param_bounds[, 2] - working_params[1:mm]))
                lambda <- 3 * lambda
            }
        }

        line_search_steps[1] <- 0
        scaled_eval_1 <- scaled_eval
        scaled_eval_2 <- scaled_eval_1
        step_objectives[1] <- j
        step_objectives[2] <- j
        param_trials <- cbind(p, p)
        line_search_steps[3] <- 1.0
        param_trials <- cbind(param_trials, working_params)
        tmpv <- param_trials[(n_ineq + 1):n_pic, 3 ] * scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)]
        ### CONTINUE HERE
        fun_value <- .solnp_fun(tmpv)
        eq_value <- .solnp_eqfun(tmpv)
        ineq_value <- .solnp_ineqfun(tmpv)
        ctmp <- get(".solnp_nfn", envir =  .env)
        assign(".solnp_nfn", ctmp + 1, envir = .env)
        scaled_eval_3 <- c(fun_value, eq_value, ineq_value) / scaling_factors[1:(n_constraints + 1)]
        step_objectives[3] <- scaled_eval_3[1]

        if ( setup[4] > 0) {
            scaled_eval_3[(n_eq + 2):(n_constraints + 1)] <- scaled_eval_3[(n_eq + 2):(n_constraints + 1)] - param_trials[1:n_ineq, 3]
        }

        if (n_constraints > 0) {
            scaled_eval_3[2:(n_constraints + 1)] <- scaled_eval_3[2:(n_constraints + 1)] - augmented_jacobian %*% param_trials[, 3] + constraint_offset
            step_objectives[3] <- scaled_eval_3[1] - t(lagrange_mults) %*% scaled_eval_3[2:(n_constraints + 1)] + rho * .vnorm(scaled_eval_3[2:(n_constraints + 1) ])^2
        }

        step_interval_width <- 1
        while (step_interval_width > tol) {
            line_search_steps[2] <- (line_search_steps[1] + line_search_steps[3]) / 2
            param_trials[, 2] <- (1 - line_search_steps[2]) * p + line_search_steps[2] * working_params
            tmpv <- param_trials[(n_ineq + 1):n_pic, 2] * scaling_factors[(n_constraints + 2):(n_constraints + n_pars + 1)]
            fun_value <- .solnp_fun(tmpv)
            eq_value <- .solnp_eqfun(tmpv)
            ineq_value <- .solnp_ineqfun(tmpv)
            ctmp <- get(".solnp_nfn", envir =  .env)
            assign(".solnp_nfn", ctmp + 1, envir = .env)
            scaled_eval_2 <- c(fun_value, eq_value, ineq_value) / scaling_factors[1:(n_constraints + 1)]
            step_objectives[2] <- scaled_eval_2[1]
            if (setup[4] > 0) {
                scaled_eval_2[(n_eq + 2):(n_constraints + 1)] <- scaled_eval_2[(n_eq + 2):(n_constraints + 1)] - param_trials[1:n_ineq , 2]
            }
            if (n_constraints > 0) {
                scaled_eval_2[2:(n_constraints + 1)] <- scaled_eval_2[2:(n_constraints + 1)] - augmented_jacobian %*% param_trials[, 2] + constraint_offset
                step_objectives[2] <- scaled_eval_2[1] - t(lagrange_mults) %*% scaled_eval_2[2:(n_constraints + 1)] + rho * .vnorm(scaled_eval_2[2:(n_constraints + 1) ])^2
            }
            max_step_obj <- max(step_objectives)

            if (max_step_obj < j) {
                min_step_obj <- min(step_objectives)
                step_interval_width <- tol * (max_step_obj - min_step_obj) / (j - max_step_obj)
            }

            condif1 <- step_objectives[2] >= step_objectives[1]
            condif2 <- step_objectives[1] <= step_objectives[3] && step_objectives[2] < step_objectives[1]
            condif3 <- step_objectives[2] <  step_objectives[1] && step_objectives[1] > step_objectives[3]

            if (condif1) {
                step_objectives[3] <- step_objectives[2]
                scaled_eval_3 <- scaled_eval_2
                line_search_steps[3] <- line_search_steps[2]
                param_trials[, 3] <- param_trials[, 2]
            }

            if (condif2) {
                step_objectives[3] <- step_objectives[2]
                scaled_eval_3 <- scaled_eval_2
                line_search_steps[3] <- line_search_steps[2]
                param_trials[, 3] <- param_trials[, 2]
            }

            if (condif3) {
                step_objectives[1] <- step_objectives[2]
                scaled_eval_1 <- scaled_eval_2
                line_search_steps[1] <- line_search_steps[2]
                param_trials[, 1] <- param_trials[, 2]
            }

            if (step_interval_width >= tol) {
                step_interval_width <- line_search_steps[3] - line_search_steps[1]
            }

        }

        delta_position <- p
        delta_gradient <- grad_augmented_obj
        status_flag <- 1
        min_step_obj <- min(step_objectives)
        if (j <= min_step_obj) {
            maxit <- minit
        }
        reduce <- (j - min_step_obj) / (1 + abs(j))

        if (reduce < tol) {
            maxit <- minit
        }

        condif1 <- step_objectives[1] <  step_objectives[2]
        condif2 <- step_objectives[3] <  step_objectives[2] && step_objectives[1] >= step_objectives[2]
        condif3 <- step_objectives[1] >= step_objectives[2] && step_objectives[3] >= step_objectives[2]

        if (condif1) {
            j <- step_objectives[1]
            p <- param_trials[, 1]
            scaled_eval <- scaled_eval_1
        }

        if (condif2) {
            j <- step_objectives[3]
            p <- param_trials[, 3]
            scaled_eval <- scaled_eval_3
        }

        if (condif3) {
            j <- step_objectives[2]
            p <- param_trials[, 2]
            scaled_eval <- scaled_eval_2
        }
    }

    p <- p * scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]  # unscale the parameter vector
    if (n_constraints > 0) {
        y <- scaling_factors[1] * y / scaling_factors[2:(n_constraints + 1)] # unscale the lagrange multipliers
    }
    augmented_hessian <- scaling_factors[1] * augmented_hessian / (scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1) ] %*% t(scaling_factors[(n_eq + 2):(n_constraints + n_pars + 1)]))

    if (reduce > tol) {
        if (trace) .solver_warnings("M3")
    }

    ans = list(p = p, y = y, augmented_hessian = augmented_hessian, lambda = lambda)
    return(ans)
}

#' Nonlinear optimization using augmented Lagrange method (C++ version)
#'
#' @param pars an numeric vector of decision variables (length n).
#' @param fn the objective function (must return a scalar).
#' @param gr an optional function for computing the analytic gradient of the function (must return
#' a vector of length n).
#' @param eq_fn an optional function for calculating equality constraints.
#' @param eq_b a vector of the equality bounds (if eq_fn provided).
#' @param eq_jac an optional function for computing the analytic jacobian of the equality.
#' function (a matrix with number of columns n and number of rows the same length as the number of equalities).
#' @param ineq_fn an optional function for calculating inequality constraints.
#' @param ineq_lower the lower bounds for the inequality (must be finite)
#' @param ineq_upper the upper bounds for the inequalitiy (must be finite)
#' @param ineq_jac an optional function for computing the analytic jacobian of the inequality (a matrix
#' with number of columns n and number of rows the same length as the number of inequalities).
#' @param lower lower bounds for the parameters. This is strictly required.
#' @param upper upper bounds for the parameters. This is strictly required.
#' @param control a list of solver control parameters (see details).
#' @param ... additional arguments passed to the supplied functions (common to all functions supplied).
#' @param use_r_version (logical) used for debugging and validation. Uses the R version of the solver
#' rather than the C++ version. Will be deprecated in future releases.
#' @returns A list with the following slot:
#' \describe{
#'   \item{pars}{ The parameters at the optimal solution found.}
#'   \item{objective}{ The value of the objective at the optimal solution found.}
#'   \item{objective_history}{ A vector of objective values obtained at each outer iteration.}
#'   \item{out_iterations}{The number of outer iterations used to arrive at the solution.}
#'   \item{convergence}{The convergence code (0 = converged).}
#'   \item{message}{The convergence message.}
#'   \item{kkt_diagnostics}{A list of optimal solution diagnostics.}
#'   \item{lagrange}{ The vector of Lagrange multipliers at the optimal solution found.}
#'   \item{n_eval}{ The number of function evaluations.}
#'   \item{elapsed}{ The time taken to find a solution.}
#'   \item{hessian}{ The Hessian at the optimal solution.}
#' }
#' @details
#' The optimization problem solved by \code{csolnp} is formulated as:
#' \deqn{
#' \begin{aligned}
#' \min_{x \in \mathbb{R}^n} \quad & f(x) \\
#' \text{s.t.} \quad & g(x) = b \\
#' & h_l \le h(x) \le h_u\\
#' & x_l \le x \le x_u\\
#' \end{aligned}
#' }
#'
#' where \eqn{f(x)} is the objective function, \eqn{g(x)} is the vector of equality constraints
#' with target value \eqn{b}, \eqn{h(x)} is the vector of inequality constraints bounded
#' by \eqn{h_l} and \eqn{h_u}, with parameter bounds \eqn{x_l} and \eqn{x_u}. Internally,
#' inequality constraints are converted into equality constraints using slack variables
#' and solved using an augmented Lagrangian approach.
#' This function is based on the original R code, but converted to C++, making use of
#' \code{Rcpp} and \code{RcppArmadillo}.
#' Additionally, it allows the user to pass in analytic gradient and Jacobians, else
#' finite differences using functions from the \code{numDeriv} package are used.
#'
#' The control list consists of the following options:
#' \describe{
#'   \item{rho}{Numeric. Initial penalty parameter for the augmented Lagrangian. Controls the weight given to constraint violation in the objective. Default is \code{1}.}
#'   \item{max_iter}{Integer. Maximum number of major (outer) iterations allowed. Default is \code{400}.}
#'   \item{min_iter}{Integer. Maximum number of minor (inner) iterations (per major iteration) for the quadratic subproblem solver. Default is \code{800}.}
#'   \item{tol}{Numeric. Convergence tolerance for both feasibility (constraint violation) and optimality (change in objective). The algorithm terminates when changes fall below this threshold. Default is \code{1e-8}.}
#'   \item{trace}{Integer If \code{1}, prints progress, \code{2} includes diagnostic information during optimization. Default is \code{0}.}
#' }
#'
#' Tracing information provides the following:
#' \describe{
#'   \item{Iter}{The current major iteration number.}
#'   \item{Obj}{The value of the objective function \eqn{f(x)} at the current iterate.}
#'   \item{||Constr||}{The norm of the current constraint violation, summarizing how well all constraints (equality and inequality) are satisfied. Typically the Euclidean or infinity norm.}
#'   \item{RelObj}{The relative change in the objective function value compared to the previous iteration, i.e., \eqn{|f_k - f_{k-1}| / max(1, |f_{k-1}|)}.}
#'   \item{Step}{The norm of the parameter update taken in this iteration, i.e., \eqn{||x_k - x_{k-1}||}.}
#'   \item{Penalty}{The current value of the penalty parameter (\eqn{\rho}) in the augmented Lagrangian. This parameter is adaptively updated to balance objective minimization and constraint satisfaction.}
#' }
#' @rdname csolnp
#' @author Alexios Galanos
#' @export
csolnp <- function(pars, fn, gr = NULL, eq_fn = NULL, eq_b = NULL, eq_jac = NULL,
                   ineq_fn = NULL, ineq_lower = NULL, ineq_upper = NULL,
                   ineq_jac = NULL, lower = NULL, upper = NULL,
                   control = list(), use_r_version = FALSE, ...)
{
  # Start timer for performance measurement
  if (use_r_version) {
    return(.rsolnp(pars, fn = fn, gr = gr, eq_fn = eq_fn, eq_b = eq_b, eq_jac = eq_jac,
                   ineq_fn = ineq_fn, ineq_lower = ineq_lower, ineq_upper = ineq_upper,
                   ineq_jac = ineq_jac, lower = lower, upper = upper, control = control, ...))
  }
  start_time <- Sys.time()
  # Store original parameter names
  parameter_names <- names(pars)
  # Indicator vector for problem characteristics
  # [1] Number of parameters (np)
  # [4] Has inequality constraints
  # [5] Number of inequality constraints
  # [7] Has equality constraints
  # [8] Number of equality constraints
  # [10] Has parameter upper/lower bounds
  # [11] Has any bounds (parameter or inequality)
  #
  if (is.null(lower)) {
    lower <- -1000 * abs(pars)
    warning("\nlower values are NULL. Setting to -1000 x abs(pars).")
  }

  if (is.null(upper)) {
    upper <- 1000 * abs(pars)
    warning("\nupper values are NULL. Setting to 1000 x abs(pars).")
  }

  # if pars are at their limits adjust
  nudge_factor <- 1e-5
  # Clamp parameters at or below lower bound
  idx_lower <- which(pars <= lower)
  if (length(idx_lower) > 0) {
    tmp_p <- lower[idx_lower]
    # For strict zero bounds, nudge slightly up
    tmp_p[tmp_p == 0] <- 1e-12
    # Nudge up by a tiny fraction
    tmp_p <- tmp_p + abs(tmp_p) * nudge_factor
    pars[idx_lower] <- tmp_p
    warning("\nsome parameters at lower bounds. Adjusting by a factor of 1e-5.")
  }

  # Clamp parameters at or above upper bound
  idx_upper <- which(pars >= upper)
  if (length(idx_upper) > 0) {
    tmp_p <- upper[idx_upper]
    # For strict zero upper bounds (rare), nudge slightly down
    tmp_p[tmp_p == 0] <- -1e-12
    # Nudge down by a tiny fraction
    tmp_p <- tmp_p - abs(tmp_p) * nudge_factor
    pars[idx_upper] <- tmp_p
    warning("\nsome parameters at upper bounds. Adjusting by a factor of 1e-5.")
  }

  problem_indicators <- solnp_problem_setup(pars, fn, gr, eq_fn, eq_b, eq_jac, ineq_fn, ineq_lower, ineq_upper, ineq_jac, lower, upper, ...)
  num_parameters <- problem_indicators[1]
  objective_list <- solnp_objective_wrapper(fn = fn, gr = gr, n = num_parameters, ...)
  inequality_list <- solnp_ineq_wrapper(ineq_fn, ineq_jac, n_ineq = problem_indicators[5], n = problem_indicators[1], ...)
  equality_list <- solnp_eq_wrapper(eq_fn, eq_b, eq_jac, n_eq = problem_indicators[8], n = problem_indicators[1], ...)
  lower_tmp <- lower
  upper_tmp <- upper
  pars_tmp <- pars
  objective_fun <- objective_list$objective_fun
  gradient_fun <- objective_list$gradient_fun

  # Objective function checks and initial evaluation
  ineq_f <- inequality_list$ineq_f
  ineq_j <- inequality_list$ineq_j
  eq_f <- equality_list$eq_f
  eq_j <- equality_list$eq_j
  aug_fn_gr <- augmented_function_gradient(gr = objective_list$gradient_fun, n_ineq = problem_indicators[5], n = problem_indicators[1])
  aug_ineq_jac <- aug_eq_jac <- NULL
  if (!is.null(ineq_fn)) {
    aug_ineq_jac <- augmented_inequality_jacobian(ineq_jac = inequality_list$ineq_j, n_ineq = problem_indicators[5], n = problem_indicators[1])
  } else {
    aug_ineq_jac <- function(pars) {return(NULL)}
  }
  if (!is.null(eq_fn)) {
    aug_eq_jac <- augmented_equality_jacobian(eq_jac = equality_list$eq_j, n_eq = problem_indicators[8], n_ineq = problem_indicators[5], n = problem_indicators[1])
  } else {
    aug_eq_jac <- function(pars) {return(NULL)}
  }
  # Inequality constraint checks and initial evaluation
  initial_ineq_values <- NULL
  initial_ineq_guess <- NULL

  if (!is.null(ineq_f)) {
    initial_ineq_values <- ineq_f(pars_tmp)
    initial_ineq_guess <- (ineq_lower + ineq_upper) / 2
  }
  # Equality constraint checks and initial evaluation
  initial_eq_values <- NULL
  if (!is.null(eq_f)) {
    initial_eq_values <- eq_f(pars_tmp)
  }

  initial_objective_value <- objective_fun(pars_tmp)
  # Combine inequality and parameter bounds into a single matrix
  all_bounds <- rbind(cbind(ineq_lower, ineq_upper),
                      cbind(lower_tmp, upper_tmp))

  # Parse and apply control parameters
  control_params <- .solnp_ctrl(control)
  penalty_param <- control_params$rho
  max_major_iterations <- control_params$max_iter
  max_minor_iterations <- control_params$min_iter
  tolerance <- control_params$tol
  trace_progress <- control_params$trace
  # Calculate total number of constraints
  n_ineq <- problem_indicators[5]
  n_eq <- problem_indicators[8]
  total_constraints <- n_ineq + n_eq

  # Initialize objective function values and internal status vector
  current_objective_value <- previous_objective_value <- initial_objective_value
  status_vector <- 0 * .ones(3, 1) # [obj_change, constraint_norm_prev, constraint_norm_curr]


  # Initialize Lagrange multipliers and constraints
  if (total_constraints > 0) {
    lagrange_mults <- 0 * .ones(total_constraints, 1)
    # Concatenate equality and inequality constraint values
    constraint_values <- c(initial_eq_values, initial_ineq_values)
    # Adjust inequality constraint values based on initial guess
    if (problem_indicators[4]) {
      # Calculate slack for inequality constraints relative to bounds
      temp_ineq_slack_lb <- constraint_values[(n_eq + 1):total_constraints] - ineq_lower
      temp_ineq_slack_ub <- ineq_upper - constraint_values[(n_eq + 1):total_constraints]
      # Check if all initial inequalities are within bounds (positive slack)
      min_slack <- apply(cbind(temp_ineq_slack_lb, temp_ineq_slack_ub), 1, min)
      if (all(min_slack > 0)) {
        initial_ineq_guess <- constraint_values[(n_eq + 1):total_constraints]
      }
      # Adjust inequality constraint values by subtracting the initial guess
      constraint_values[(n_eq + 1):total_constraints] <-
        constraint_values[(n_eq + 1):total_constraints] - initial_ineq_guess
    }
    # Store the initial norm of the constraint violations
    status_vector[2] <- .vnorm(constraint_values)

    # If initial constraints are "small" (within 10*tolerance), set penalty to 0
    if (max(status_vector[2] - 10 * tolerance, n_ineq, na.rm = TRUE) <= 0) {
      penalty_param <- 0
    }
  } else {
    lagrange_mults <- 0
  }

  # Starting augmented parameter vector (initial_ineq_guess for inequalities, then original parameters)
  augmented_parameters <- c(initial_ineq_guess, pars_tmp)

  # Initialize Hessian approximation (identity matrix for augmented parameters)
  augmented_hessian <- diag(num_parameters + n_ineq)

  # Initialize mu (related to optimality conditions, often small)
  lambda <- num_parameters # Or some other initial value

  # Store initial objective and constraint values
  scaled_eval <- c(initial_objective_value, initial_eq_values, initial_ineq_values)

  # Initialize major iteration counter
  major_iteration_count <- 0

  # Store historical objective values
  historical_objective_values <- c(initial_objective_value)

  penalty_param <- control_params$rho
  tol <- control_params$tol
  min_iter <- control_params$min_iter
  trace <- control_params$trace

  opt_state <- list()
  opt_state$major_iteration_count <- major_iteration_count
  opt_state$max_major_iterations <- max_major_iterations
  opt_state$penalty_param <- penalty_param
  opt_state$min_iter <- max_minor_iterations
  opt_state$tol <- tol
  opt_state$ftol <- 1e-8
  opt_state$trace <- trace
  opt_state$n_eq <- n_eq
  opt_state$n_ineq <- n_ineq
  opt_state$num_parameters <- num_parameters
  opt_state$total_constraints <- total_constraints
  opt_state$problem_indicators <- problem_indicators
  opt_state$scaled_eval <- scaled_eval
  opt_state$augmented_parameters <- augmented_parameters
  opt_state$lagrange_mults <- lagrange_mults
  opt_state$augmented_hessian <- augmented_hessian
  opt_state$lambda <- lambda
  opt_state$lower_tmp <- lower
  opt_state$upper_tmp <- upper
  opt_state$ineq_lower <- ineq_lower
  opt_state$ineq_upper <- ineq_upper
  opt_state$all_bounds <- all_bounds
  opt_state$solnp_fun <- objective_fun
  opt_state$solnp_gradfun <- gradient_fun
  opt_state$solnp_eqfun <- eq_f
  opt_state$solnp_ineqfun <- ineq_f
  opt_state$solnp_eqjac <- eq_j
  opt_state$solnp_ineqjac <- ineq_j
  opt_state$previous_objective_value <- previous_objective_value
  opt_state$status_vector <- status_vector
  opt_state$historical_objective_values <- historical_objective_values
  result <- .csolnp(opt_state)
  # Calculate elapsed time and retrieve final function evaluation count
  elapsed_time <- Sys.time() - start_time

  optimized_parameters <- result$parameters

  if (problem_indicators[4] == 1) {
    initial_ineq_guess <- optimized_parameters[1:n_ineq]
  }
  optimized_parameters <- optimized_parameters[(n_ineq + 1):(n_ineq + num_parameters)]
  lagrange_mults <- result$lagrange_mults
  augmented_hessian <- result$augmented_hessian
  lambda <- result$lambda
  status_vector <- result$status_vector
  historical_objective_values <- result$historical_objective_values
  # Assign original names to optimized parameters
  names(optimized_parameters) <- parameter_names
  n_function_eval <- result$n_fun_evaluations
  major_iteration_count <- result$major_iteration_count
  convergence <- result$convergence
  convergence_message <- switch(as.character(convergence),
                                "0" = "Solution converged within tolerance.",
                                "1" = "Exiting after maximum number of iterations, tolerance not achieved.",
                                "2" = "Solution not reliable.")
  if (trace > 1) {
      print(convergence_message)
  }

  # Prepare results list
  results <- list(
    pars = optimized_parameters,
    convergence = result$convergence,
    message = convergence_message,
    objective = result$best_objective,
    objective_history = as.numeric(historical_objective_values),
    lagrange = lagrange_mults,
    hessian = augmented_hessian,
    ineq_initial_values = initial_ineq_guess,
    n_eval = n_function_eval,
    outer_iterations = major_iteration_count,
    kkt_diagnostics = result$kkt_diagnostics,
    elapsed_time = elapsed_time
  )
  return(results)
}


#' Summarize KKT Condition Diagnostics
#'
#' Given a list of KKT diagnostic statistics (stationarity, feasibility, complementarity, etc.),
#' this function prints a clear summary indicating which KKT conditions are satisfied at a specified tolerance.
#'
#' @param kkt A named list containing numeric entries for KKT diagnostics.
#'        Required names are \code{"kkt_stationarity"}, \code{"eq_violation"},
#'        \code{"ineq_violation"}, \code{"dual_feas_violation"}, \code{"compl_slackness"}.
#' @param tol Numeric tolerance for considering a condition as "satisfied". Default is \code{1e-8}.
#'
#' @return An object of class "solnp_kkt_summary" (a data.frame with columns: condition, value, status, tol).
#' @examples
#' kkt <- list(
#'   kkt_stationarity = 5.828909e-06,
#'   eq_violation = 0,
#'   ineq_violation = 0,
#'   dual_feas_violation = 0.4380053,
#'   compl_slackness = 0
#' )
#' kkt_diagnose(kkt, tol = 1e-8)
#'
#' @export
kkt_diagnose <- function(kkt, tol = 1e-8) {
  stopifnot(is.list(kkt))
  required_names <- c(
    "kkt_stationarity",
    "eq_violation",
    "ineq_violation",
    "dual_feas_violation",
    "compl_slackness"
  )
  missing <- setdiff(required_names, names(kkt))
  if (length(missing) > 0)
    stop("KKT diagnostic list is missing required element(s): ", paste(missing, collapse = ", "))

  nulls <- vapply(kkt[required_names], is.null, logical(1))
  if (any(nulls))
    stop("KKT diagnostic contains NULL for: ", paste(required_names[nulls], collapse = ", "))

  vals <- as.numeric(unlist(kkt[required_names]))
  na_idx <- which(is.na(vals))
  if (length(na_idx) > 0)
    stop("KKT diagnostic contains NA for: ", paste(required_names[na_idx], collapse = ", "))

  res <- data.frame(
    condition = c("stationarity", "equality_violation", "inequality_violation", "dual_feasibility", "complementarity"),
    value = as.numeric(unlist(kkt[required_names])),
    status = ifelse(abs(as.numeric(unlist(kkt[required_names]))) <= tol, "OK", "FAIL"),
    stringsAsFactors = FALSE
  )
  attr(res, "tolerance") <- tol
  class(res) <- c("solnp_kkt_summary", class(res))
  return(res)
}

#' @export
print.solnp_kkt_summary <- function(x, ...) {
  cat("KKT summary (tol =", attr(x, "tolerance"), "):\n")
  for (i in seq_len(nrow(x))) {
    cat(" -", x$condition[i], ":",
        if (x$status[i] == "OK") "OK" else "FAIL",
        sprintf("(|value| = %.2e)", abs(x$value[i])), "\n")
  }
  invisible(x)
}

quick_stationary_feasible <- function(x0, lower, upper, ineq_fn = NULL, ineq_lower = NULL, ineq_upper = NULL,
                                      eq_fn = NULL, eq_b = NULL, penalty = 1e4, maxit = 10)
{
    # Inequality constraints (may be empty)
    standard_form_ineq_fn <- function(x) {
        if (is.null(ineq_fn) || is.null(ineq_lower) || is.null(ineq_upper)) return(numeric(0))
        h <- ineq_fn(x)
        c(ineq_lower - h, h - ineq_upper)
    }
    # Equality constraints (may be empty)
    standard_form_eq_fn <- function(x) {
        if (is.null(eq_fn) || is.null(eq_b)) return(numeric(0))
        eq_fn(x) - eq_b
    }
    penalized <- function(x) {
        ineq_val <- standard_form_ineq_fn(x)
        eq_val <- standard_form_eq_fn(x)
        penalty_ineq <- if (length(ineq_val) == 0) 0 else sum(pmax(ineq_val, 0)^2)
        penalty_eq   <- if (length(eq_val)   == 0) 0 else sum(eq_val^2)
        norm_term <- sum((x - x0)^2)
        norm_term + penalty * (penalty_ineq + penalty_eq)
    }
    res <- optim(x0, penalized, method = "L-BFGS-B", lower = lower, upper = upper, control = list(maxit = maxit))
    res$par
}

sample_box_feasible <- function(lower, upper, eps = 1e-4) {
    effective_lower <- lower + eps * (upper > lower + 2*eps)
    effective_upper <- upper - eps * (upper > lower + 2*eps)
    is_too_narrow <- effective_lower >= effective_upper
    res <- numeric(length(lower))
    res[is_too_narrow] <- (lower[is_too_narrow] + upper[is_too_narrow]) / 2
    res[!is_too_narrow] <- effective_lower[!is_too_narrow] +
        (effective_upper[!is_too_narrow] - effective_lower[!is_too_narrow]) * runif(sum(!is_too_narrow))
    res
}

generate_feasible_starts <- function(n, fn = NULL, lower, upper, ineq_fn, ineq_lower, ineq_upper, eq_fn, eq_b, maxit = 100, penalty = 1e4, eps = 1e-4, seed = NULL)
{
    if (!is.null(seed)) set.seed(seed)
    candidates <- vector("list", n)
    for (i in seq_len(n)) {
        x0 <- sample_box_feasible(lower, upper, eps = eps)
        if (!is.null(ineq_fn) | !is.null(eq_fn)) {
            feasible <- quick_stationary_feasible(x0, lower, upper, ineq_fn, ineq_lower, ineq_upper, eq_fn, eq_b, maxit = maxit, penalty = penalty)
        } else {
            feasible <- x0
        }
        candidates[[i]] <- feasible
    }
    # Optionally, rank by objective value
    if (!is.null(fn)) {
        obj_vals <- sapply(candidates, fn)
        ord <- order(obj_vals, decreasing = FALSE)
        candidates <- candidates[ord]
    }
    out <- do.call(rbind, candidates)
    if (!is.null(names(lower))) colnames(out) <- names(lower)
    return(out)
}

hs01_problem <- function()
{
    fn <- function(x) {
        100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2
    }
    gr <- function(x) {
        g2 <- 200 * (x[2] - x[1]^2)
        g1 <- -2 * (x[1] * (g2 - 1) + 1)
        c(g1, g2)
    }
    lower <- c(-1000, -1.5)
    upper <- c(1000, 1000)
    start <- c(-2, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- c(1, 1)
    list(
        name = "hs01",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs02_problem <- function()
{
    fn <- function(x) {
        100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2
    }
    gr <- function(x) {
        dfdx1 <- -400 * x[1] * (x[2] - x[1]^2) - 2 * (1 - x[1])
        dfdx2 <- 200 * (x[2] - x[1]^2)
        c(dfdx1, dfdx2)
    }
    lower <- c(-1000, 1.5)
    upper <- c( 1000,  1.5)
    start <- c(-2, 1)
    # Known solution from Fortran (XEX and FEX)
    w1 <- sqrt(598/1200)
    best_par <- c(2 * w1 * cos(acos(2.5e-3 / w1^3) / 3), 1.5)
    best_fn <- 100 * (best_par[2] - best_par[1]^2)^2 + (1 - best_par[1])^2
    list(
        name = "hs02",
        fn = fn,
        gr = gr,
        eq_fn = NULL,
        eq_b = NULL,
        eq_jac = NULL,
        ineq_fn = NULL,
        ineq_jac = NULL,
        ineq_lower = NULL,
        ineq_upper = NULL,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}



hs03_problem <- function()
{
    fn <- function(x) {
        x[2] + 0.00001 * (x[2] - x[1])^2
    }
    gr <- function(x) {
        dfdx1 <- -2e-5 * (x[2] - x[1])
        dfdx2 <- 1 + 2e-5 * (x[2] - x[1])
        c(dfdx1, dfdx2)
    }
    lower <- c(-1000, 0)
    upper <- c( 1000, 1000)
    start <- c(10, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs03",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 0,
        best_par = c(0, 0)
    )
}

hs04_problem <- function()
{
    fn <- function(x) {
        ((x[1] + 1)^3) / 3 + x[2]
    }
    gr <- function(x) {
        dfdx1 <- (x[1] + 1)^2
        dfdx2 <- 1
        c(dfdx1, dfdx2)
    }
    lower <- c(1, 0)
    upper <- c(1000, 1000)
    start <- c(1.125, 0.125)
    eq_fn <- NULL
    eq_jac <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs04",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 8/3,
        best_par = c(1, 0)
    )
}

hs05_problem <- function()
{
    fn <- function(x) {
        sin(x[1] + x[2]) + (x[1] - x[2])^2 - 1.5 * x[1] + 2.5 * x[2] + 1
    }
    gr <- function(x) {
        dfdx1 <- cos(x[1] + x[2]) + 2 * (x[1] - x[2]) - 1.5
        dfdx2 <- cos(x[1] + x[2]) - 2 * (x[1] - x[2]) + 2.5
        c(dfdx1, dfdx2)
    }
    lower <- c(-1.5, -3)
    upper <- c(4, 3)
    start <- c(0, 0)
    eq_fn <- NULL
    eq_jac <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs05",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -(sqrt(3)/2 + pi/3),
        best_par = c(0.5 - pi/3, -0.5 - pi/3)
    )
}

hs06_problem <- function()
{
    fn <- function(x) {
        (1 - x[1])^2
    }
    gr <- function(x) {
        dfdx1 <- -2 * (1 - x[1])
        dfdx2 <- 0
        c(dfdx1, dfdx2)
    }
    eq_fn <- function(x) {
        10 * (x[2] - x[1]^2)
    }
    eq_jac <- function(x) {
        matrix(c(-20 * x[1], 10), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c( 1000,  1000)
    start <- c(-1.2, 1)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs06",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 0,
        best_par = c(1, 1)
    )
}

hs07_problem <- function()
{
    fn <- function(x) {
        log(1 + x[1]^2) - x[2]
    }
    gr <- function(x) {
        dfdx1 <- 2 * x[1] / (1 + x[1]^2)
        dfdx2 <- -1
        c(dfdx1, dfdx2)
    }
    eq_fn <- function(x) {
        (1 + x[1]^2)^2 + x[2]^2 - 4
    }
    eq_jac <- function(x) {
        matrix(c(4 * x[1] * (1 + x[1]^2), 2 * x[2]), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(2, 2)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs07",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -sqrt(3),
        best_par = c(0, sqrt(3))
    )
}

s08_problem <- function()
{
    fn <- function(x) {
        -1
    }
    gr <- function(x) {
        c(0, 0)
    }
    eq_fn <- function(x) {
        c(x[1]^2 + x[2]^2 - 25, x[1] * x[2] - 9)
    }
    eq_jac <- function(x) {
        matrix(c(2 * x[1], 2 * x[2], x[2], x[1]), nrow = 2, byrow = TRUE)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(2, 1)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    # List all four solutions as best_par if you want:
    A <- sqrt((25 + sqrt(301))/2)
    B <- sqrt((25 - sqrt(301))/2)
    best_par <- list(
        c( A,  9/A),
        c(-A, -9/A),
        c( B,  9/B),
        c(-B, -9/B)
    )
    list(
        name = "hs08",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -1,
        best_par = NULL
    )
}

hs09_problem <- function()
{
    fn <- function(x) {
        sin(pi * x[1] / 12) * cos(pi * x[2] / 16)
    }
    gr <- function(x) {
        v3 <- pi / 12
        v4 <- pi / 16
        v1 <- v3 * x[1]
        v2 <- v4 * x[2]
        dfdx1 <- v3 * cos(v1) * cos(v2)
        dfdx2 <- -v4 * sin(v1) * sin(v2)
        c(dfdx1, dfdx2)
    }
    eq_fn <- function(x) {
        4 * x[1] - 3 * x[2]
    }
    eq_jac <- function(x) {
        matrix(c(4, -3), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(0, 0)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs09",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -0.5,
        best_par = c(-3, -4)
    )
}

hs10_problem <- function()
{
    fn <- function(x) {
        x[1] - x[2]
    }
    gr <- function(x) {
        c(1, -1)
    }
    ineq_fn <- function(x) {
        -3 * x[1]^2 + 2 * x[1] * x[2] - x[2]^2 + 1
    }
    ineq_jac <- function(x) {
        matrix(c(-6 * x[1] + 2 * x[2], 2 * (x[1] - x[2])), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(-10, 10)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    list(
        name = "hs10",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -1,
        best_par = c(0, 1)
    )
}

hs11_problem <- function()
{
    fn <- function(x) {
        (x[1] - 5)^2 + x[2]^2 - 25
    }
    gr <- function(x) {
        c(2 * (x[1] - 5), 2 * x[2])
    }
    ineq_fn <- function(x) {
        -x[1]^2 + x[2]
    }
    ineq_jac <- function(x) {
        matrix(c(-2 * x[1], 1), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(4.9, 0.1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e10
    # Solution per Fortran code
    AEX <- 7.5 * sqrt(6)
    AW <- (sqrt(AEX^2 + 1) + AEX)^(1/3)
    QAW <- AW^2
    x1sol <- (AW - 1/AW) / sqrt(6)
    x2sol <- (QAW - 2 + 1/QAW) / 6
    best_par <- c(x1sol, x2sol)
    best_fn <- (x1sol - 5)^2 + x2sol^2 - 25
    list(
        name = "hs11",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs12_problem <- function()
{
    fn <- function(x) {
        0.5 * x[1]^2 + x[2]^2 - x[1] * x[2] - 7 * x[1] - 7 * x[2]
    }
    gr <- function(x) {
        c(x[1] - x[2] - 7, 2 * x[2] - x[1] - 7)
    }
    ineq_fn <- function(x) {
        25 - 4 * x[1]^2 - x[2]^2
    }
    ineq_jac <- function(x) {
        matrix(c(-8 * x[1], -2 * x[2]), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(0, 0)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    list(
        name = "hs12",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -30,
        best_par = c(2, 3)
    )
}

hs13_problem <- function()
{
    fn <- function(x) {
        (x[1] - 2)^2 + x[2]^2
    }
    gr <- function(x) {
        c(2 * (x[1] - 2), 2 * x[2])
    }
    ineq_fn <- function(x) {
        (1 - x[1])^3 - x[2]
    }
    ineq_jac <- function(x) {
        matrix(c(-3 * (1 - x[1])^2, -1), nrow = 1)
    }
    lower <- c(0, 0)
    upper <- c(1000, 1000)
    start <- c(0, 0)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    list(
        name = "hs13",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 1,
        best_par = c(1, 0)
    )
}
hs14_problem <- function()
{
    fn <- function(x) {
        (x[1] - 2)^2 + (x[2] - 1)^2
    }
    gr <- function(x) {
        c(2 * (x[1] - 2), 2 * (x[2] - 1))
    }
    ineq_fn <- function(x) {
        1 - 0.25 * x[1]^2 - x[2]^2
    }
    ineq_jac <- function(x) {
        matrix(c(-0.5 * x[1], -2 * x[2]), nrow = 1)
    }
    eq_fn <- function(x) {
        x[1] - 2 * x[2] + 1
    }
    eq_jac <- function(x) {
        matrix(c(1, -2), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(2, 2)
    ineq_lower <- 0
    ineq_upper <- 1e8
    eq_b <- 0
    w7 <- sqrt(7)
    x1sol <- (w7 - 1) * 0.5
    x2sol <- (w7 + 1) * 0.25
    best_par <- c(x1sol, x2sol)
    best_fn <- 9 - 23 * w7 / 8
    list(
        name = "hs14",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs15_problem <- function()
{
    fn <- function(x) {
        (x[2] - x[1]^2)^2 + 0.01 * (1 - x[1])^2
    }
    gr <- function(x) {
        g2 <- 2 * (x[2] - x[1]^2)
        g1 <- -0.02 * (x[1] * (g2 - 1) + 1)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        c(x[1] * x[2] - 1, x[2]^2 + x[1])
    }
    ineq_jac <- function(x) {
        matrix(c(x[2], x[1], 1,2 * x[2]), nrow = 2, byrow = TRUE)
    }
    lower <- c(-1000, -1000)
    upper <- c(0.5, 1000)
    start <- c(-2, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    list(
        name = "hs15",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 3.065,
        best_par = c(0.5, 2.0)
    )
}

hs16_problem <- function()
{
    fn <- function(x) {
        100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2
    }
    gr <- function(x) {
        g2 <- 200 * (x[2] - x[1]^2)
        g1 <- -2 * (x[1] * (g2 - 1) + 1)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        c(x[2]^2 + x[1], x[1]^2 + x[2])
    }
    ineq_jac <- function(x) {
        matrix(c(1, 2 * x[2], 2 * x[1], 1), nrow = 2, byrow = TRUE)
    }
    lower <- c(-0.5, -1000)
    upper <- c(0.5, 1)
    start <- c(0, 0.5)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    list(
        name = "hs16",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 0.25,
        best_par = c(0.5, 0.25)
    )
}

hs17_problem <- function()
{
    fn <- function(x) {
        100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2
    }
    gr <- function(x) {
        g2 <- 200 * (x[2] - x[1]^2)
        g1 <- -2 * (x[1] * (g2 - 1) + 1)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        c(x[2]^2 - x[1], x[1]^2 - x[2])
    }
    ineq_jac <- function(x) {
        matrix(c(-1, 2 * x[2], 2 * x[1], -1), nrow = 2, byrow = TRUE)
    }

    lower <- c(-2, -1000)
    upper <- c(0.5, 1)
    start <- c(-1, 0)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    list(
        name = "hs17",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 1,
        best_par = c(0, 0)
    )
}

hs18_problem <- function()
{
    fn <- function(x) {
        0.01 * x[1]^2 + x[2]^2
    }
    gr <- function(x) {
        c(0.02 * x[1], 2 * x[2])
    }
    ineq_fn <- function(x) {
        c(x[1] * x[2] - 25, x[1]^2 + x[2]^2 - 25)
    }
    ineq_jac <- function(x) {
        matrix(c(x[2], x[1], 2 * x[1], 2 * x[2]), nrow = 2, byrow = TRUE)
    }
    lower <- c(2, 0)
    upper <- c(50, 50)
    start <- c(10, 2)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    x1sol <- sqrt(250)
    x2sol <- 0.1 * x1sol
    best_par <- c(x1sol, x2sol)
    best_fn <- 5
    list(
        name = "hs18",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}


hs19_problem <- function()
{
    fn <- function(x) {
        (x[1] - 10)^3 + (x[2] - 20)^3
    }
    gr <- function(x) {
        c(3 * (x[1] - 10)^2, 3 * (x[2] - 20)^2)
    }
    ineq_fn <- function(x) {
        c((x[1] - 5)^2 + (x[2] - 5)^2 - 100, 82.81 - (x[1] - 6)^2 - (x[2] - 5)^2)
    }
    ineq_jac <- function(x) {
        matrix(c(2 * (x[1] - 5), 2 * (x[2] - 5), -2 * (x[1] - 6), -2 * (x[2] - 5)), nrow = 2, byrow = TRUE)
    }
    lower <- c(13, 0)
    upper <- c(100, 100)
    start <- c(20.1, 5.84)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    saex <- 17.280975
    aex <- sqrt(saex)
    x1sol <- 14.095
    x2sol <- 5 - aex
    best_par <- c(x1sol, x2sol)
    best_fn <- (4.095)^3 - (15 + aex)^3
    list(
        name = "hs19",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs20_problem <- function()
{
    fn <- function(x) {
        100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2
    }
    gr <- function(x) {
        g2 <- 200 * (x[2] - x[1]^2)
        g1 <- -2 * (x[1] * (g2 - 1) + 1)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        c(x[2]^2 + x[1], x[1]^2 + x[2], x[1]^2 + x[2]^2 - 1)
    }
    ineq_jac <- function(x) {
        matrix(c(1, 2 * x[2], 2 * x[1], 1, 2 * x[1], 2 * x[2]), nrow = 3, byrow = TRUE)
    }
    lower <- c(-0.5, -1000)
    upper <- c(0.5, 1000)
    start <- c(0.1, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0, 0)
    ineq_upper <- c(1e8, 1e8, 1e8)
    x1sol <- 0.5
    x2sol <- sqrt(3) * 0.5
    best_par <- c(x1sol, x2sol)
    best_fn <- 81.5 - 25 * sqrt(3)
    list(
        name = "hs20",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs21_problem <- function()
{
    fn <- function(x) {
        0.01 * x[1]^2 + x[2]^2 - 100
    }
    gr <- function(x) {
        c(0.02 * x[1], 2 * x[2])
    }
    ineq_fn <- function(x) {
        10 * x[1] - x[2] - 10
    }
    ineq_jac <- function(x) {
        matrix(c(10, -1), nrow = 1)
    }
    lower <- c(2, -50)
    upper <- c(50, 50)
    start <- c(4, -1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- -99.96
    best_par <- c(2, 0)
    list(
        name = "hs21",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs22_problem <- function()
{
    fn <- function(x) {
        (x[1] - 2)^2 + (x[2] - 1)^2
    }
    gr <- function(x) {
        c(2 * (x[1] - 2), 2 * (x[2] - 1))
    }
    ineq_fn <- function(x) {
        2 - x[1] - x[2]
    }
    ineq_jac <- function(x) {
        matrix(c(-1, -1), nrow = 1)
    }
    eq_fn <- function(x) {
        x[2] - x[1]^2
    }
    eq_jac <- function(x) {
        matrix(c(-2 * x[1], 1), nrow = 1)
    }
    lower <- c(-1000, -1000)
    upper <- c(1000, 1000)
    start <- c(2, 2)
    eq_b <- 0
    ineq_lower <- 0
    ineq_upper <- 1e8
    list(
        name = "hs22",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 1,
        best_par = c(1, 1)
    )
}

hs23_problem <- function()
{
    fn <- function(x) {
        x[1]^2 + x[2]^2
    }
    gr <- function(x) {
        c(2 * x[1], 2 * x[2])
    }
    ineq_fn <- function(x) {
        c(x[1] + x[2] - 1, x[1]^2 + x[2]^2 - 1, 9 * x[1]^2 + x[2]^2 - 9, x[1]^2 - x[2], x[2]^2 - x[1])
    }
    ineq_jac <- function(x) {
        matrix(c(1, 1, 2 * x[1],  2 * x[2], 18 * x[1], 2 * x[2], 2 * x[1], -1, -1, 2 * x[2]), nrow = 5, byrow = TRUE)
    }
    lower <- c(-50, -50)
    upper <- c(50, 50)
    start <- c(3, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- rep(0, 5)
    ineq_upper <- rep(1e8, 5)
    list(
        name = "hs23",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 2,
        best_par = c(1, 1)
    )
}

hs24_problem <- function()
{
    fn <- function(x) {
        A <- sqrt(3)
        ((x[1] - 3)^2 - 9) * x[2]^3 / (27 * A)
    }
    gr <- function(x) {
        A <- sqrt(3)
        g1 <- 2 * (x[1] - 3) * x[2]^3 / (27 * A)
        g2 <- ((x[1] - 3)^2 - 9) * x[2]^2 / (9 * A)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        A <- sqrt(3)
        c(x[1] / A - x[2], x[1] + x[2] * A, 6 - x[2] * A - x[1])
    }
    ineq_jac <- function(x) {
        A <- sqrt(3)
        matrix(c(1/A, -1, 1, A, -1, -A), nrow = 3, byrow = TRUE)
    }
    lower <- c(0, 0)
    upper <- c(1000, 1000)
    start <- c(500, 500)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- rep(0, 3)
    ineq_upper <- rep(1e8, 3)
    A <- sqrt(3)
    best_par <- c(3, A)
    best_fn <- -1
    list(
        name = "hs24",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs25_problem <- function()
{
    fn <- function(x) {
        s <- 1:99
        u <- 25 + (-50 * log(s / 100))^(2/3)
        z <- u - x[2]
        if (any(z <= 0)) {
            return(sum((x - 5)^2))
        }
        exp_term <- exp(-(z^x[3]) / x[1])
        sum((-s / 100 + exp_term)^2)
    }
    gr <- function(x) {
        s <- 1:99
        u <- 25 + (-50 * log(s / 100))^(2/3)
        z <- u - x[2]
        if (any(z <= 0)) {
            return(2 * (x - 5))
        }
        e <- exp(-(z^x[3]) / x[1])
        r <- -s / 100 + e

        de_dx1 <-  (z^x[3]) / (x[1]^2) * e
        de_dx2 <- -x[3] * z^(x[3] - 1) / x[1] * e
        de_dx3 <- -log(z) * z^x[3] / x[1] * e

        grad1 <- sum(2 * r * de_dx1)
        grad2 <- sum(2 * r * de_dx2)
        grad3 <- sum(2 * r * de_dx3)

        c(grad1, grad2, grad3)
    }
    lower <- c(0.1, 1e-5, 1e-5)
    upper <- c(100, 25.6, 5)
    start <- c(100, 12.5, 3)
    eq_fn <- NULL
    eq_jac <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs25",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 0,
        best_par = c(50, 25, 1.5)
    )
}

hs26_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[2] - x[3])^4
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - x[2])
        g3 <- -4 * (x[2] - x[3])^3
        g2 <- -g1 - g3
        c(g1, g2, g3)
    }
    eq_fn <- function(x) {
        x[1] * (1 + x[2]^2) + x[3]^4 - 3
    }
    eq_jac <- function(x) {
        matrix(c(1 + x[2]^2, 2 * x[1] * x[2], 4 * x[3]^3), nrow = 1)
    }
    lower <- rep(-1000, 3)
    upper <- rep(1000, 3)
    start <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- 0
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs26",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 0,
        best_par = c(1, 1, 1)
    )
}

hs27_problem <- function()
{
    fn <- function(x) {
        (x[1] - 1)^2 + 100 * (x[2] - x[1]^2)^2
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - 1) - 400 * (x[2] - x[1]^2) * x[1]
        g2 <- 200 * (x[2] - x[1]^2)
        g3 <- 0
        c(g1, g2, g3)
    }
    eq_fn <- function(x) {
        x[1] + x[3]^2 + 1
    }
    eq_jac <- function(x) {
        matrix(c(1, 0, 2 * x[3]), nrow = 1)
    }
    lower <- rep(-1000, 3)
    upper <- rep(1000, 3)
    start <- c(2, 2, 2)
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- 0
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs27",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 4,
        best_par = c(-1, 1, 0)
    )
}

hs28_problem <- function()
{
    fn <- function(x) {
        (x[1] + x[2])^2 + (x[2] + x[3])^2
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] + x[2])
        g3 <- 2 * (x[2] + x[3])
        g2 <- g1 + g3
        c(g1, g2, g3)
    }
    eq_fn <- function(x) {
        x[1] + 2 * x[2] + 3 * x[3] - 1
    }
    eq_jac <- function(x) {
        matrix(c(1, 2, 3), nrow = 1)
    }
    lower <- rep(-1000, 3)
    upper <- rep(1000, 3)
    start <- c(-4, 1, 1)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs28",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 0,
        best_par = c(0.5, -0.5, 0.5)
    )
}


hs29_problem <- function()
{
    fn <- function(x) {
        -x[1] * x[2] * x[3]
    }
    gr <- function(x) {
        c(-x[2] * x[3],
          -x[1] * x[3],
          -x[1] * x[2])
    }
    ineq_fn <- function(x) {
        48 - x[1]^2 - 2 * x[2]^2 - 4 * x[3]^2
    }
    ineq_jac <- function(x) {
        matrix(c(-2 * x[1], -4 * x[2], -8 * x[3]), nrow = 1)
    }
    lower <- rep(-1000, 3)
    upper <- rep(1000, 3)
    start <- c(1, 1, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- -16 * sqrt(2)
    best_par <- c(4, 2 * sqrt(2), 2)
    list(
        name = "hs29",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs30_problem <- function()
{
    fn <- function(x) {
        x[1]^2 + x[2]^2 + x[3]^2
    }
    gr <- function(x) {
        c(2 * x[1], 2 * x[2], 2 * x[3])
    }
    ineq_fn <- function(x) {
        x[1]^2 + x[2]^2 - 1
    }
    ineq_jac <- function(x) {
        matrix(c(2 * x[1], 2 * x[2], 0), nrow = 1)
    }
    lower <- c(1, -10, -10)
    upper <- c(10, 10, 10)
    start <- c(1, 1, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- 1
    best_par <- c(1, 0, 0)
    list(
        name = "hs30",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}


hs31_problem <- function()
{
    fn <- function(x) {
        9 * x[1]^2 + x[2]^2 + 9 * x[3]^2
    }
    gr <- function(x) {
        c(18 * x[1], 2 * x[2], 18 * x[3])
    }
    ineq_fn <- function(x) {
        x[1] * x[2] - 1
    }
    ineq_jac <- function(x) {
        matrix(c(x[2], x[1], 0), nrow = 1)
    }
    lower <- c(-10, 1, -10)
    upper <- c(10, 10, 1)
    start <- c(1, 1, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- 6
    best_par <- c(1 / sqrt(3), sqrt(3), 0)
    list(
        name = "hs31",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs32_problem <- function()
{
    fn <- function(x) {
        (x[1] + 3 * x[2] + x[3])^2 + 4 * (x[1] - x[2])^2
    }
    gr <- function(x) {
        g1 <- 10 * x[1] - 2 * x[2] + 2 * x[3]
        g2 <- -2 * x[1] + 26 * x[2] + 6 * x[3]
        g3 <- 2 * (x[1] + 3 * x[2] + x[3])
        c(g1, g2, g3)
    }
    ineq_fn <- function(x) {
        -x[1]^3 + 6 * x[2] + 4 * x[3] - 3
    }
    ineq_jac <- function(x) {
        matrix(c(-3 * x[1]^2, 6, 4), nrow = 1)
    }
    eq_fn <- function(x) {
        1 - x[1] - x[2] - x[3]
    }
    eq_jac <- function(x) {
        matrix(c(-1, -1, -1), nrow = 1)
    }
    lower <- rep(0, 3)
    upper <- rep(1000, 3)
    start <- c(0.1, 0.7, 0.2)
    eq_b <- 0
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- 1
    best_par <- c(0, 0, 1)
    list(
        name = "hs32",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs33_problem <- function()
{
    fn <- function(x) {
        (x[1] - 1) * (x[1] - 2) * (x[1] - 3) + x[3]
    }
    gr <- function(x) {
        g1 <- 3 * x[1]^2 - 12 * x[1] + 11
        g2 <- 0
        g3 <- 1
        c(g1, g2, g3)
    }
    ineq_fn <- function(x) {
        c(x[3]^2 - x[1]^2 - x[2]^2, x[1]^2 + x[2]^2 + x[3]^2 - 4)
    }

    ineq_jac <- function(x) {
        matrix(c(-2 * x[1], -2 * x[2],  2 * x[3], 2 * x[1],  2 * x[2],  2 * x[3]), nrow = 2, byrow = TRUE)
    }
    lower <- c(0, 0, 0)
    upper <- c(1000, 1000, 5)
    start <- c(1, 1, 3)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    best_fn <- sqrt(2) - 6
    best_par <- c(0, sqrt(2), sqrt(2))
    list(
        name = "hs33",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs34_problem <- function()
{
    fn <- function(x) {
        -x[1]
    }
    gr <- function(x) {
        c(-1, 0, 0)
    }
    ineq_fn <- function(x) {
        c(x[2] - exp(x[1]), x[3] - exp(x[2]))
    }
    ineq_jac <- function(x) {
        matrix(c(-exp(x[1]), 1, 0, 0, -exp(x[2]), 1), nrow = 2, byrow = TRUE)
    }
    lower <- c(0, 0, 0)
    upper <- c(100, 100, 10)
    start <- c(1, 1.05, 2.9)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    best_par <- c(log(log(10)), log(10), 10)
    best_fn <- -log(log(10))
    list(
        name = "hs34",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs35_problem <- function()
{
    fn <- function(x) {
        9 - 8 * x[1] - 6 * x[2] - 4 * x[3] + 2 * x[1]^2 + 2 * x[2]^2 + x[3]^2 + 2 * x[1] * x[2] + 2 * x[1] * x[3]
    }
    gr <- function(x) {
        c(-8 + 4 * x[1] + 2 * x[2] + 2 * x[3], -6 + 4 * x[2] + 2 * x[1], -4 + 2 * x[3] + 2 * x[1])
    }
    ineq_fn <- function(x) {
        -x[1] - x[2] - 2 * x[3] + 3
    }
    ineq_jac <- function(x) {
        matrix(c(-1, -1, -2), nrow = 1)
    }
    lower <- c(0, 0, 0)
    upper <- rep(1000, 3)
    start <- rep(0.5, 3)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- 1/9
    best_par <- c(4/3, 7/9, 4/9)
    list(
        name = "hs35",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs36_problem <- function()
{
    fn <- function(x) {
        -x[1] * x[2] * x[3]
    }
    gr <- function(x) {
        c(-x[2] * x[3], -x[1] * x[3], -x[1] * x[2])
    }
    ineq_fn <- function(x) {
        72 - x[1] - 2 * x[2] - 2 * x[3]
    }
    ineq_jac <- function(x) {
        matrix(c(-1, -2, -2), nrow = 1)
    }
    lower <- c(0, 0, 0)
    upper <- c(20, 11, 42)
    start <- c(10, 10, 10)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- -3300
    best_par <- c(20, 11, 15)
    list(
        name = "hs36",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs37_problem <- function()
{
    fn <- function(x) {
        -x[1] * x[2] * x[3]
    }
    gr <- function(x) {
        c(-x[2] * x[3], -x[1] * x[3], -x[1] * x[2])
    }
    ineq_fn <- function(x) {
        c(72 - x[1] - 2 * x[2] - 2 * x[3], x[1] + 2 * x[2] + 2 * x[3])
    }
    ineq_jac <- function(x) {
        matrix(c(-1, -2, -2, 1,  2,  2), nrow = 2, byrow = TRUE)
    }
    lower <- rep(0, 3)
    upper <- rep(42, 3)
    start <- rep(10, 3)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- c(0, 0)
    ineq_upper <- c(1e8, 1e8)
    best_fn <- -3456
    best_par <- c(24, 12, 12)
    list(
        name = "hs37",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs38_problem <- function()
{
    fn <- function(x) {
        (100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2 +
             90 * (x[4] - x[3]^2)^2 + (1 - x[3])^2 +
             10.1 * ((x[2] - 1)^2 + (x[4] - 1)^2) +
             19.8 * (x[2] - 1) * (x[4] - 1))
    }
    gr <- function(x) {
        c(
            -400 * x[1] * (x[2] - x[1]^2) - 2 * (1 - x[1]),
            200 * (x[2] - x[1]^2) + 20.2 * (x[2] - 1) + 19.8 * (x[4] - 1),
            -360 * x[3] * (x[4] - x[3]^2) - 2 * (1 - x[3]),
            180 * (x[4] - x[3]^2) + 20.2 * (x[4] - 1) + 19.8 * (x[2] - 1)
        )
    }
    lower <- rep(-10, 4)
    upper <- rep(10, 4)
    start <- c(-3, -1, -3, -1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- c(1, 1, 1, 1)
    list(
        name = "hs38",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs39_problem <- function()
{
    fn <- function(x) {
        -x[1]
    }
    gr <- function(x) {
        c(-1, 0, 0, 0)
    }
    eq_fn <- function(x) {
        c(x[2] - x[1]^3 - x[3]^2, x[1]^2 - x[2] - x[4]^2)
    }
    eq_jac <- function(x) {
        matrix(c(-3 * x[1]^2, 1, -2 * x[3], 0, 2 * x[1], -1, 0, -2 * x[4]), nrow = 2, byrow = TRUE)
    }
    lower <- rep(-1000, 4)
    upper <- rep(1000, 4)
    start <- rep(2, 4)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- -1
    best_par <- c(1, 1, 0, 0)
    list(
        name = "hs39",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs40_problem <- function()
{
    fn <- function(x) {
        -x[1] * x[2] * x[3] * x[4]
    }
    gr <- function(x) {
        c(-x[2] * x[3] * x[4], -x[1] * x[3] * x[4], -x[1] * x[2] * x[4], -x[1] * x[2] * x[3])
    }
    eq_fn <- function(x) {
        c(x[1]^3 + x[2]^2 - 1, x[1]^2 * x[4] - x[3], x[4]^2 - x[2])
    }
    eq_jac <- function(x) {
        matrix(c(3 * x[1]^2, 2 * x[2], 0, 0,
                 2 * x[1] * x[4], 0, -1, x[1]^2,
                 0, -1, 0, 2 * x[4]), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-1000, 4)
    upper <- rep(1000, 4)
    start <- rep(0.8, 4)
    eq_b <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- -0.25
    best_par <- c(2^(-1/3), 2^(-0.5), 2^(-11/12), 2^(-0.25))
    list(
        name = "hs40",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs41_problem <- function()
{
    fn <- function(x) {
        2 - x[1] * x[2] * x[3]
    }
    gr <- function(x) {
        c(-x[2] * x[3], -x[1] * x[3], -x[1] * x[2], 0)
    }
    eq_fn <- function(x) {
        x[1] + 2 * x[2] + 2 * x[3] - x[4]
    }
    eq_jac <- function(x) {
        matrix(c(1, 2, 2, -1), nrow = 1)
    }
    lower <- c(0, 0, 0, 0)
    upper <- c(1, 1, 1, 2)
    start <- rep(1, 4)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 52 / 27
    best_par <- c(2/3, 1/3, 1/3, 2)
    list(
        name = "hs41",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs42_problem <- function()
{
    fn <- function(x) {
        (x[1] - 1)^2 + (x[2] - 2)^2 + (x[3] - 3)^2 + (x[4] - 4)^2
    }
    gr <- function(x) {
        2 * (x - 1:4)
    }
    eq_fn <- function(x) {
        c(x[1] - 2,  x[3]^2 + x[4]^2 - 2)
    }
    eq_jac <- function(x) {
        matrix(c(1, 0, 0, 0, 0, 0, 2 * x[3], 2 * x[4]), nrow = 2, byrow = TRUE)
    }
    lower <- rep(-1000, 4)
    upper <- rep(1000, 4)
    start <- rep(1, 4)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 28 - 10 * sqrt(2)
    best_par <- c(2, 2, sqrt(0.72), sqrt(1.28))
    list(
        name = "hs42",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs43_problem <- function()
{
    fn <- function(x) {
        x[1]^2 + x[2]^2 + 2 * x[3]^2 + x[4]^2 -
            5 * x[1] - 5 * x[2] - 21 * x[3] + 7 * x[4]
    }
    gr <- function(x) {
        c(2 * x[1] - 5, 2 * x[2] - 5, 4 * x[3] - 21, 2 * x[4] + 7)
    }
    ineq_fn <- function(x) {
        c(-x[1]^2 - x[2]^2 - x[3]^2 - x[4]^2 - x[1] + x[2] - x[3] + x[4] + 8,
          -x[1]^2 - 2 * x[2]^2 - x[3]^2 - 2 * x[4]^2 + x[1] + x[4] + 10,
          -2 * x[1]^2 - x[2]^2 - x[3]^2 - 2 * x[1] + x[2] + x[4] + 5)
    }
    ineq_jac <- function(x) {
        matrix(c(-2 * x[1] - 1, -2 * x[2] + 1, -2 * x[3] - 1, -2 * x[4] + 1,
                 -2 * x[1] + 1, -4 * x[2], -2 * x[3], -4 * x[4] + 1,
                 -4 * x[1] - 2, -2 * x[2] + 1, -2 * x[3], 1), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-1000, 4)
    upper <- rep(1000, 4)
    start <- c(0, 0, 0, 0)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- rep(0, 3)
    ineq_upper <- rep(1e8, 3)
    best_fn <- -44
    best_par <- c(0, 1, 2, -1)
    list(
        name = "hs43",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs44_problem <- function()
{
    fn <- function(x) {
        x[1] - x[2] - x[3] - x[1] * x[3] + x[1] * x[4] + x[2] * x[3] - x[2] * x[4]
    }
    gr <- function(x) {
        c(1 - x[3] + x[4], -1 + x[3] - x[4], -1 - x[1] + x[2], x[1] - x[2])
    }
    ineq_fn <- function(x) {
        c(8 - x[1] - 2 * x[2], 12 - 4 * x[1] - x[2], 12 - 3 * x[1] - 4 * x[2], 8 - 2 * x[3] - x[4],
          8 - x[3] - 2 * x[4], 5 - x[3] - x[4])
    }
    ineq_jac <- function(x) {
        matrix(c(-1, -2,  0,  0,
                 -4, -1,  0,  0,
                 -3, -4,  0,  0,
                 0,  0, -2, -1,
                 0,  0, -1, -2,
                 0,  0, -1, -1), nrow = 6, byrow = TRUE)
    }
    lower <- rep(0, 4)
    upper <- rep(1000, 4)
    start <- c(1, 1, 1, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- rep(0, 6)
    ineq_upper <- rep(1e8, 6)
    best_fn <- -15
    best_par <- c(0, 3, 0, 4)
    list(
        name = "hs44",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs45_problem <- function()
{
    fn <- function(x) {
        2 - x[1] * x[2] * x[3] * x[4] * x[5] / 120
    }
    gr <- function(x) {
        c(-x[2] * x[3] * x[4] * x[5] / 120,
          -x[1] * x[3] * x[4] * x[5] / 120,
          -x[1] * x[2] * x[4] * x[5] / 120,
          -x[1] * x[2] * x[3] * x[5] / 120,
          -x[1] * x[2] * x[3] * x[4] / 120)
    }
    lower <- rep(0, 5)
    upper <- 1:5
    start <- rep(1, 5)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 1
    best_par <- 1:5
    list(
        name = "hs45",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs46_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[3] - 1)^2 + (x[4] - 1)^4 + (x[5] - 1)^6
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - x[2])
        g2 <- -g1
        g3 <- 2 * (x[3] - 1)
        g4 <- 4 * (x[4] - 1)^3
        g5 <- 6 * (x[5] - 1)^5
        c(g1, g2, g3, g4, g5)
    }
    eq_fn <- function(x) {
        c(x[1]^2 * x[4] + sin(x[4] - x[5]) - 1, x[2] + x[3]^4 * x[4]^2 - 2)
    }
    eq_jac <- function(x) {
        matrix(c(2 * x[1] * x[4], 0, 0, x[1]^2 + cos(x[4] - x[5]), -cos(x[4] - x[5]),
                 0, 1, 4 * x[3]^3 * x[4]^2, 2 * x[3]^4 * x[4], 0), nrow = 2, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- c(0.5 * sqrt(2), 1.75, 0.5, 2, 2)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- rep(1, 5)
    list(
        name = "hs46",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs47_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[2] - x[3])^2 + (x[3] - x[4])^4 + (x[4] - x[5])^4
    }
    gr <- function(x) {
        v1 <- 2 * (x[1] - x[2])
        v2 <- 2 * (x[2] - x[3])
        v3 <- 4 * (x[3] - x[4])^3
        v4 <- 4 * (x[4] - x[5])^3
        c(v1, -v1 + v2, -v2 + v3, -v3 + v4, -v4)
    }
    eq_fn <- function(x) {
        c(x[1] + x[2]^2 + x[3]^3 - 3, x[2] - x[3]^2 + x[4] - 1, x[1] * x[5] - 1)
    }
    eq_jac <- function(x) {
        matrix(c(1, 2 * x[2], 3 * x[3]^2, 0, 0,
                 0, 1, -2 * x[3], 1, 0,
                x[5], 0, 0, 0, x[1]), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- c(2, sqrt(2), -1, 2 - sqrt(2), 0.5)
    eq_b <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- rep(1, 5)
    list(
        name = "hs47",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs48_problem <- function()
{
    fn <- function(x) {
        (x[1] - 1)^2 + (x[2] - x[3])^2 + (x[4] - x[5])^2
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - 1)
        g2 <- 2 * (x[2] - x[3])
        g3 <- -g2
        g4 <- 2 * (x[4] - x[5])
        g5 <- -g4
        c(g1, g2, g3, g4, g5)
    }
    eq_fn <- function(x) {
        c(sum(x) - 5, x[3] - 2 * (x[4] + x[5]) + 3)
    }
    eq_jac <- function(x) {
        matrix(c(1, 1, 1, 1, 1,
                 0, 0, 1, -2, -2), nrow = 2, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- c(3, 5, -3, 2, -2)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- rep(1, 5)
    list(
        name = "hs48",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs49_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[3] - 1)^2 + (x[4] - 1)^4 + (x[5] - 1)^6
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - x[2])
        g2 <- -g1
        g3 <- 2 * (x[3] - 1)
        g4 <- 4 * (x[4] - 1)^3
        g5 <- 6 * (x[5] - 1)^5
        c(g1, g2, g3, g4, g5)
    }
    eq_fn <- function(x) {
        c(x[1] + x[2] + x[3] + 4 * x[4] - 7, x[3] + 5 * x[5] - 6)
    }
    eq_jac <- function(x) {
        matrix(c(1, 1, 1, 4, 0,
                 0, 0, 1, 0, 5), nrow = 2, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- c(10, 7, 2, -3, 0.8)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- rep(1, 5)
    list(
        name = "hs49",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs50_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[2] - x[3])^2 + (x[3] - x[4])^4 + (x[4] - x[5])^4
    }
    gr <- function(x) {
        v1 <- 2 * (x[1] - x[2])
        v2 <- 2 * (x[2] - x[3])
        v3 <- 4 * (x[3] - x[4])^3
        v4 <- 4 * (x[4] - x[5])^3
        c(v1, -v1 + v2, -v2 + v3, -v3 + v4, -v4)
    }
    eq_fn <- function(x) {
        c(x[1] + 2 * x[2] + 3 * x[3] - 6,
          x[2] + 2 * x[3] + 3 * x[4] - 6,
          x[3] + 2 * x[4] + 3 * x[5] - 6
        )
    }
    eq_jac <- function(x) {
        matrix(c(1, 2, 3, 0, 0,
                 0, 1, 2, 3, 0,
                 0, 0, 1, 2, 3), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- c(35, -31, 11, 5, -5)
    eq_b <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- rep(1, 5)
    list(
        name = "hs50",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs51_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[2] + x[3] - 2)^2 + (x[4] - 1)^2 + (x[5] - 1)^2
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - x[2])
        g3 <- 2 * (x[2] + x[3] - 2)
        g2 <- g3 - g1
        g4 <- 2 * (x[4] - 1)
        g5 <- 2 * (x[5] - 1)
        c(g1, g2, g3, g4, g5)
    }
    eq_fn <- function(x) {
        c(x[1] + 3 * x[2] - 4, x[3] + x[4] - 2 * x[5], x[2] - x[5])
    }
    eq_jac <- function(x) {
        matrix(c(1, 3, 0, 0, 0,
                 0, 0, 1, 1, -2,
                 0, 1, 0, 0, -1), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- c(2.5, 0.5, 2, -1, 0.5)
    eq_b <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0
    best_par <- rep(1, 5)
    list(
        name = "hs51",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs52_problem <- function()
{
    fn <- function(x) {
        (4 * x[1] - x[2])^2 + (x[2] + x[3] - 2)^2 + (x[4] - 1)^2 + (x[5] - 1)^2
    }
    gr <- function(x) {
        g1 <- 8 * (4 * x[1] - x[2])
        g3 <- 2 * (x[2] + x[3] - 2)
        g2 <- -0.25 * g1 + g3
        g4 <- 2 * (x[4] - 1)
        g5 <- 2 * (x[5] - 1)
        c(g1, g2, g3, g4, g5)
    }
    eq_fn <- function(x) {
        c(x[1] + 3 * x[2], x[3] + x[4] - 2 * x[5], x[2] - x[5])
    }
    eq_jac <- function(x) {
        matrix(c(1, 3, 0, 0, 0,
                 0, 0, 1, 1, -2,
                 0, 1, 0, 0, -1), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-1000, 5)
    upper <- rep(1000, 5)
    start <- rep(2, 5)
    eq_b <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 1859/349
    best_par <- c(-33/349, 11/349, 180/349, -158/349, 11/349)
    list(
        name = "hs52",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs53_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + (x[2] + x[3] - 2)^2 + (x[4] - 1)^2 + (x[5] - 1)^2
    }
    gr <- function(x) {
        g1 <- 2 * (x[1] - x[2])
        g3 <- 2 * (x[2] + x[3] - 2)
        g2 <- g3 - g1
        g4 <- 2 * (x[4] - 1)
        g5 <- 2 * (x[5] - 1)
        c(g1, g2, g3, g4, g5)
    }
    eq_fn <- function(x) {
        c(x[1] + 3 * x[2], x[3] + x[4] - 2 * x[5], x[2] - x[5])
    }
    eq_jac <- function(x) {
        matrix(c(1, 3, 0, 0, 0,
                 0, 0, 1, 1, -2,
                 0, 1, 0, 0, -1), nrow = 3, byrow = TRUE)
    }
    lower <- rep(-10, 5)
    upper <- rep(10, 5)
    start <- rep(2, 5)
    eq_b <- c(0, 0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 176/43
    best_par <- c(-33/43, 11/43, 27/43, -5/43, 11/43)
    list(
        name = "hs53",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs54_problem <- function()
{
    fn <- function(x) {
        v1 <- x[1] - 1e4
        v2 <- x[2] - 1
        v3 <- x[3] - 2e6
        v4 <- x[4] - 10
        v5 <- x[5] - 1e-3
        v6 <- x[6] - 1e8
        v7 <- 1 / 0.96
        v8 <- 1 / 4.9e13
        v9 <- 1 / 2.45e13
        q <- (1.5625e-8 * v1^2 + 5e-5 * v1 * v2 + v2^2) * v7 + v3^2 * v8 +
            4e-4 * v4^2 + 4e2 * v5^2 + 4e-18 * v6^2
        -exp(-0.5 * q)
    }
    gr <- function(x) {
        v1 <- x[1] - 1e4
        v2 <- x[2] - 1
        v3 <- x[3] - 2e6
        v4 <- x[4] - 10
        v5 <- x[5] - 1e-3
        v6 <- x[6] - 1e8
        v7 <- 1 / 0.96
        v8 <- 1 / 4.9e13
        v9 <- 1 / 2.45e13
        q <- (1.5625e-8 * v1^2 + 5e-5 * v1 * v2 + v2^2) * v7 + v3^2 * v8 +
            4e-4 * v4^2 + 4e2 * v5^2 + 4e-18 * v6^2
        dq1 <- (3.125e-8 * v1 + 5e-5 * v2) * v7
        dq2 <- (5e-5 * v1 + 2 * v2) * v7
        dq3 <- v3 * v9
        dq4 <- 8e-4 * v4
        dq5 <- 800 * v5
        dq6 <- 8e-18 * v6
        factor <- 0.5 * exp(-0.5 * q)
        c(factor * dq1, factor * dq2, factor * dq3, factor * dq4, factor * dq5, factor * dq6)
    }
    eq_fn <- function(x) {
        x[1] + 4e3 * x[2] - 1.76e4
    }
    eq_jac <- function(x) {
        matrix(c(1, 4e3, 0, 0, 0, 0), nrow = 1)
    }
    lower <- c(0, -10, 0, 0, 0, 0)
    upper <- c(2e4, 10, 1e7, 20, 1, 2e8)
    start <- c(6e3, 1.5, 4e6, 2, 3e-3, 5e7)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- -exp(-27 / 280)
    best_par <- c(9.16e4 / 7, 79 / 70, 2e6, 10, 1e-3, 1e8)
    list(
        name = "hs54",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs55_problem <- function()
{
    # solver killer problem
    fn <- function(x) {
        x14 <- x[1] * x[4]
        if (x14 > 10) x14 <- 10
        x[1] + 2 * x[2] + 4 * x[5] + exp(x14)
    }
    gr <- function(x) {
        x14 <- x[1] * x[4]
        if (x14 > 10) {
            v1 <- exp(10)
            dfdx1 <- 1 + x[4] * v1
            dfdx4 <- x[1] * v1
        } else {
            v1 <- exp(x14)
            dfdx1 <- 1 + x[4] * v1
            dfdx4 <- x[1] * v1
        }
        c(dfdx1, 2, 0, dfdx4, 4, 0)
    }
    eq_fn <- function(x) {
        c(x[1] + 2 * x[2] + 5 * x[5] - 6, x[1] + x[2] + x[3] - 3,
            x[4] + x[5] + x[6] - 2, x[1] + x[4] - 1,
            x[2] + x[5] - 2, x[3] + x[6] - 2)
    }
    eq_jac <- function(x) {
        matrix(c(1, 2, 0, 0, 5, 0,
                 1, 1, 1, 0, 0, 0,
                 0, 0, 0, 1, 1, 1,
                 1, 0, 0, 1, 0, 0,
                 0, 1, 0, 0, 1, 0,
                 0, 0, 1, 0, 0, 1), nrow = 6, byrow = TRUE)
    }
    lower <- c(0, 0, 0, 0, 0, 0)
    upper <- c(1, 1000, 1000, 1, 1000, 1000)
    start <- c(1, 2, 0, 0, 0, 2)
    eq_b <- rep(0, 6)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 19/3
    best_par <- c(0, 4/3, 5/3, 1, 2/3, 1/3)
    list(
        name = "hs55",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs56_problem <- function()
{
    fn <- function(x) {
        -x[1] * x[2] * x[3]
    }
    gr <- function(x) {
        c(-x[2] * x[3], -x[1] * x[3], -x[1] * x[2], 0, 0, 0, 0)
    }
    eq_fn <- function(x) {
        c(x[1] - 4.2 * sin(x[4])^2, x[2] - 4.2 * sin(x[5])^2,
          x[3] - 4.2 * sin(x[6])^2, x[1] + 2 * x[2] + 2 * x[3] - 7.2 * sin(x[7])^2)
    }
    eq_jac <- function(x) {
        matrix(c(1, 0, 0, -8.4 * sin(x[4]) * cos(x[4]), 0, 0, 0,
                 0, 1, 0, 0, -8.4 * sin(x[5]) * cos(x[5]), 0, 0,
                 0, 0, 1, 0, 0, -8.4 * sin(x[6]) * cos(x[6]), 0,
                 1, 2, 2, 0, 0, 0, -14.4 * sin(x[7]) * cos(x[7])), nrow = 4, byrow = TRUE)
    }
    lower <- rep(-1000, 7)
    upper <- rep(1000, 7)
    start <- c(
        1, 1, 1,
        asin(sqrt(1 / 4.2)),
        asin(sqrt(1 / 4.2)),
        asin(sqrt(1 / 4.2)),
        asin(sqrt(5 / 7.2))
    )
    eq_b <- rep(0, 4)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- -3.456
    best_par <- c(
        2.4, 1.2, 1.2,
        asin(sqrt(4 / 7)),
        asin(sqrt(2 / 7)),
        asin(sqrt(2 / 7)),
        2 * atan(1)
    )
    list(
        name = "hs56",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs57_problem <- function()
{
    A <- rep(0, 44); B <- rep(0, 44)

    # Step 2: Fortran style assignments (copy the block structure)
    A[1:2] <- 8; A[16+1:2] <- 18; A[30+1:2] <- 28; A[35+1:2] <- 32; A[38+1:2] <- 36; A[40+1:2] <- 38
    B[1:2] <- 0.49; B[6+1:2] <- 0.46; B[11+1:2] <- 0.43; B[14+1:2] <- 0.43; B[18+1:2] <- 0.42
    B[21+1:2] <- 0.41; B[25+1:2] <- 0.40; B[29+1:2] <- 0.41; B[36+1:2] <- 0.40; B[40+1:2] <- 0.40
    B[42+1:2] <- 0.39
    for (i in 1:3) {
        A[10+i] <- 14;  A[13+i] <- 16;  A[18+i] <- 20
        A[21+i] <- 22;  A[24+i] <- 24;  A[27+i] <- 26;  A[32+i] <- 30
        B[31+i] <- 0.40
    }
    for (i in 1:4) {
        A[2+i] <- 10; A[6+i] <- 12
    }
    A[38] <- 34; A[43] <- 40; A[44] <- 42
    B[3] <- 0.48; B[4] <- 0.47; B[5] <- 0.48; B[6] <- 0.47; B[9] <- 0.45
    B[10] <- 0.43; B[11] <- 0.45; B[14] <- 0.44; B[17] <- 0.46; B[18] <- 0.45
    B[21] <- 0.43; B[24] <- 0.40; B[25] <- 0.42; B[28] <- 0.41; B[29] <- 0.40
    B[35] <- 0.38; B[36] <- 0.41; B[39] <- 0.41; B[40] <- 0.38
    fn <- function(x) {
        F <- B - x[1] - (0.49 - x[1]) * exp(-x[2] * (A - 8))
        sum(F^2)
    }
    gr <- function(x) {
        F <- B - x[1] - (0.49 - x[1]) * exp(-x[2] * (A - 8))
        V1 <- exp(-x[2] * (A - 8))
        DF1 <- -1 + V1
        DF2 <- (A - 8) * (0.49 - x[1]) * V1
        g1 <- 2 * sum(F * DF1)
        g2 <- 2 * sum(F * DF2)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        -x[1] * x[2] + 0.49 * x[2] - 0.09
    }
    ineq_jac <- function(x) {
        matrix(c(-x[2], -x[1] + 0.49), nrow = 1)
    }
    lower <- c(0.4, -4)
    upper <- c(1000, 1000)
    start <- c(0.42, 5)
    ineq_lower <- 0
    ineq_upper <- 1e8
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    best_fn <- 0.0284596697213
    best_par <- c(0.419952674511, 1.28484562930)
    list(
        name = "hs57",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs58_problem <- function()
{
    fn <- function(x) {
        100 * (x[2] - x[1]^2)^2 + (1 - x[1])^2
    }
    gr <- function(x) {
        g2 <- 200 * (x[2] - x[1]^2)
        g1 <- -2 * (x[1] * (g2 - 1) + 1)
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        c(x[2]^2 - x[1], x[1]^2 - x[2], x[1]^2 + x[2]^2 - 1)
    }
    ineq_jac <- function(x) {
        matrix(c(-1, 2 * x[2],
                 2 * x[1], -1,
                 2 * x[1], 2 * x[2]), nrow = 3, byrow = TRUE)
    }
    lower <- c(-2, -1000)
    upper <- c(0.5, 1000)
    start <- c(-2, 1)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- rep(0, 3)
    ineq_upper <- rep(1e8, 3)
    best_fn <- 3.19033354957
    best_par <- c(-0.786150483331, 0.618034533851)
    list(
        name = "hs58",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs59_problem <- function()
{
    fn <- function(x) {
        x1 <- x[1]; x2 <- x[2]
        x11 <- x1; x12 <- x1^2; x13 <- x1^3; x14 <- x1^4
        x21 <- x2; x22 <- x2^2; x23 <- x2^3; x24 <- x2^4
        xx11 <- x1 * x2; xx12 <- x1 * x2^2; xx21 <- x1^2 * x2
        xx31 <- x1^3 * x2
        -75.196 + 3.8112 * x11 - 0.12694 * x12 + 2.0567e-3 * x13 - 1.0345e-5 * x14 +
            6.8306 * x21 - 3.0234e-2 * x11 * x21 + 1.28134e-3 * x12 * x21 -
            3.5256e-5 * x13 * x21 + 2.266e-7 * x14 * x21 -
            0.25645 * x22 + 3.4604e-3 * x23 - 1.3514e-5 * x24 +
            28.106 / (x21 + 1) + 5.2375e-6 * x12 * x22 + 6.3e-8 * x13 * x22 -
            7e-10 * x13 * x23 - 3.4054e-4 * x11 * x22 + 1.6638e-6 * x11 * x23 +
            2.8673 * exp(5e-4 * x11 * x21)
    }
    gr <- function(x) {
        x1 <- x[1]; x2 <- x[2]
        x11 <- x1; x12 <- x1^2; x13 <- x1^3; x14 <- x1^4
        x21 <- x2; x22 <- x2^2; x23 <- x2^3
        xx11 <- x1 * x2; xx12 <- x1 * x2^2; xx21 <- x1^2 * x2; xx31 <- x1^3 * x2

        expterm <- exp(5e-4 * xx11)
        g1 <- 3.8112 - 0.25388 * x1 + 6.1701e-3 * x12 - 4.138e-5 * x13 -
            3.0234e-2 * x2 + 2.56268e-3 * xx11 - 1.05768e-4 * xx21 + 9.064e-7 * xx31 +
            1.0475e-5 * xx12 + 1.89e-7 * x12 * x22 - 2.1e-9 * x12 * x23 -
            3.4054e-4 * x22 + 1.6638e-6 * x23 + 1.43365e-3 * x2 * expterm
        g2 <- 6.8306 - 3.0234e-2 * x1 + 1.28134e-3 * x12 - 3.5256e-5 * x13 +
            2.266e-7 * x14 - 0.5129 * x2 + 1.03812e-2 * x22 - 5.4056e-5 * x23 - 28.106 / (x2 + 1)^2 +
            1.0475e-5 * xx21 + 1.26e-7 * xx31 - 2.1e-9 * x13 * x22 - 6.8108e-4 * xx11 + 4.9914e-6 * xx12 +
            1.43365e-3 * x1 * expterm
        c(g1, g2)
    }
    ineq_fn <- function(x) {
        c(x[1] * x[2] - 700, x[2] - 8e-3 * x[1]^2, (x[2] - 50)^2 - 5 * (x[1] - 55))
    }
    ineq_jac <- function(x) {
        matrix(c(x[2], x[1],
                 -1.6e-2 * x[1], 1,
                 -5, 2 * (x[2] - 50)), nrow = 3, byrow = TRUE)
    }
    lower <- c(0, 0)
    upper <- c(75, 65)
    start <- c(90, 10)
    ineq_lower <- rep(0, 3)
    ineq_upper <- rep(1e8, 3)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    best_fn <- -7.80422632408
    best_par <- c(13.5501042366, 51.6601812877)
    list(
        name = "hs59",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs60_problem <- function()
{
    fn <- function(x) {
        (x[1] - 1)^2 + (x[1] - x[2])^2 + (x[2] - x[3])^4
    }
    gr <- function(x) {
        v1 <- 2 * (x[1] - x[2])
        g1 <- 2 * (x[1] - 1) + v1
        g3 <- -4 * (x[2] - x[3])^3
        g2 <- -g3 - v1
        c(g1, g2, g3)
    }
    eq_fn <- function(x) {
        x[1] * (1 + x[2]^2) + x[3]^4 - 4 - 3 * sqrt(2)
    }
    eq_jac <- function(x) {
        matrix(c(1 + x[2]^2, 2 * x[1] * x[2], 4 * x[3]^3), nrow = 1)
    }
    lower <- rep(-10, 3)
    upper <- rep(10, 3)
    start <- rep(2, 3)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 0.0325682002513
    best_par <- c(1.10485902423, 1.19667419413, 1.53526225739)
    list(
        name = "hs60",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs61_problem <- function()
{
    fn <- function(x) {
        4 * x[1]^2 + 2 * x[2]^2 + 2 * x[3]^2 - 33 * x[1] + 16 * x[2] - 24 * x[3]
    }
    gr <- function(x) {
        c(8 * x[1] - 33, 4 * x[2] + 16, 4 * x[3] - 24)
    }
    eq_fn <- function(x) {
        c(3 * x[1] - 2 * x[2]^2 - 7, 4 * x[1] - x[3]^2 - 11)
    }
    eq_jac <- function(x) {
        matrix(c(3, -4 * x[2], 0, 4, 0, -2 * x[3]), nrow = 2, byrow = TRUE)
    }
    lower <- rep(-1000, 3)
    upper <- rep(1000, 3)
    start <- c(0, 0, 0)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- -143.646142201
    best_par <- c(5.32677015744, -2.11899863998, 3.21046423906)
    list(
        name = "hs61",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs62_problem <- function()
{
    fn <- function(x) {
        # Variable substitutions
        B3 <- x[3] + 0.03
        C3 <- 0.13 * x[3] + 0.03
        B2 <- B3 + x[2]
        C2 <- B3 + 0.07 * x[2]
        B1 <- B2 + x[1]
        C1 <- B2 + 0.09 * x[1]
        V5 <- B1 / C1
        V6 <- B2 / C2
        V7 <- B3 / C3

        if (V5 <= 0 || V6 <= 0 || V7 <= 0) {
            # Penalty for infeasible region
            S <- sum((x - 5)^2)
            return(S + 1e3 - 2.67e4)
        }

        -32.174 * (255 * log(V5) + 280 * log(V6) + 290 * log(V7))
    }
    gr <- function(x) {
        # Variable substitutions
        B3 <- x[3] + 0.03
        C3 <- 0.13 * x[3] + 0.03
        B2 <- B3 + x[2]
        C2 <- B3 + 0.07 * x[2]
        B1 <- B2 + x[1]
        C1 <- B2 + 0.09 * x[1]
        # Reciprocals
        RB1 <- 1 / B1
        RB2 <- 1 / B2
        RB3 <- 1 / B3
        RC1 <- 1 / C1
        RC2 <- 1 / C2
        RC3 <- 1 / C3
        V1 <- -32.174 * 255
        V2 <- -32.174 * 280
        V3 <- -32.174 * 290
        V4 <- V1 * (RB1 - RC1)

        # Gradient (as per Fortran)
        g1 <- V1 * (RB1 - 0.09 * RC1)
        g2 <- V4 + V2 * (RB2 - 0.07 * RC2)
        g3 <- V4 + V2 * (RB2 - RC2) + V3 * (RB3 - 0.13 * RC3)
        c(g1, g2, g3)
    }
    eq_fn <- function(x) {
        sum(x) - 1
    }
    eq_jac <- function(x) {
        matrix(rep(1, 3), nrow = 1)
    }
    lower <- rep(0, 3)
    upper <- rep(1, 3)
    start <- c(0.7, 0.2, 0.1)
    eq_b <- 0
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- -26272.5144873
    best_par <- c(0.617813298210, 0.328202155786, 0.0539845460119)
    list(
        name = "hs62",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs63_problem <- function()
{
    fn <- function(x) {
        1e3 - x[1]^2 - 2 * x[2]^2 - x[3]^2 - x[1] * x[2] - x[1] * x[3]
    }
    gr <- function(x) {
        c(-2 * x[1] - x[2] - x[3], -4 * x[2] - x[1], -2 * x[3] - x[1])
    }
    eq_fn <- function(x) {
        c(8 * x[1] + 14 * x[2] + 7 * x[3] - 56,
          x[1]^2 + x[2]^2 + x[3]^2 - 25)
    }
    eq_jac <- function(x) {
        matrix(c(8, 14, 7, 2 * x[1], 2 * x[2], 2 * x[3]),nrow = 2, byrow = TRUE)
    }
    lower <- rep(0, 3)
    upper <- rep(1000, 3)
    start <- rep(2, 3)
    eq_b <- c(0, 0)
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    best_fn <- 961.715172127
    best_par <- c(3.51211841492, 0.216988174172, 3.55217403459)
    list(
        name = "hs63",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs64_problem <- function()
{
    fn <- function(x) {
        5 * x[1] + 5e4 / x[1] + 20 * x[2] + 7.2e4 / x[2] + 10 * x[3] + 1.44e5 / x[3]
    }
    gr <- function(x) {
        c(5 - 5e4 / x[1]^2, 20 - 7.2e4 / x[2]^2, 10 - 1.44e5 / x[3]^2)
    }
    ineq_fn <- function(x) {
        1 - 4 / x[1] - 32 / x[2] - 120 / x[3]
    }
    ineq_jac <- function(x) {
        matrix(c(4 / x[1]^2, 32 / x[2]^2, 120 / x[3]^2), nrow = 1)
    }
    lower <- rep(1e-5, 3)
    upper <- rep(1000, 3)
    start <- rep(1, 3)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- 6299.84242821
    best_par <- c(108.734717597, 85.1261394257, 204.324707858)
    list(
        name = "hs64",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

hs65_problem <- function()
{
    fn <- function(x) {
        (x[1] - x[2])^2 + ((x[1] + x[2] - 10)/3)^2 + (x[3] - 5)^2
    }
    gr <- function(x) {
        v1 <- 2 * (x[1] - x[2])
        v2 <- 2 * (x[1] + x[2] - 10) / 9
        c(v1 + v2, -v1 + v2, 2 * (x[3] - 5))
    }
    ineq_fn <- function(x) {
        48 - x[1]^2 - x[2]^2 - x[3]^2
    }
    ineq_jac <- function(x) {
        matrix(c(-2 * x[1], -2 * x[2], -2 * x[3]), nrow = 1)
    }
    lower <- c(-4.5, -4.5, -5)
    upper <- c(4.5, 4.5, 5)
    start <- c(-5, 5, 0)
    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- 0
    ineq_upper <- 1e8
    best_fn <- 0.953528856757
    best_par <- c(3.65046182158, 3.65046168940, 4.62041750754)
    list(
        name = "hs65",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = best_fn,
        best_par = best_par
    )
}

alkylation_problem <- function() {
    fn <- function(x) {
        -0.63 * x[4] * x[7] + 50.4 * x[1] + 3.5 * x[2] + x[3] + 33.6 * x[5]
    }

    gr <- NULL  # If you want to implement gradient later

    eq_fn <- function(x) {
        z1 <- 98 * x[3] - 0.1 * x[4] * x[6] * x[9] - x[3] * x[6]
        z2 <- 1000 * x[2] + 100 * x[5] - 100 * x[1] * x[8]
        z3 <- 122 * x[4] - 100 * x[1] - 100 * x[5]
        return(c(z1, z2, z3))
    }

    eq_jac <- NULL  # Optional, if analytic Jacobian is available

    ineq_fn <- function(x) {
        z1 <- (1.12 * x[1] + 0.13167 * x[1] * x[8] - 0.00667 * x[1] * x[8]^2) / x[4]
        z2 <- (1.098 * x[8] - 0.038 * x[8]^2 + 0.325 * x[6] + 57.25) / x[7]
        z3 <- (-0.222 * x[10] + 35.82) / x[9]
        z4 <- (3 * x[7] - 133) / x[10]
        return(c(z1, z2, z3, z4))
    }

    ineq_jac <- NULL  # Optional, if analytic Jacobian is available

    lower <- c(0, 0, 0, 10, 0, 85, 10, 3, 1, 145)
    upper <- c(20, 16, 120, 50, 20, 93, 95, 12, 4, 162)
    ineq_lower <- c(0.99, 0.99, 0.9, 0.99)
    ineq_upper <- c(100 / 99, 100 / 99, 10 / 9, 100 / 99)
    eq_b <- c(0, 0, 0)
    start <- c(17.45, 12, 110, 30, 19.74, 89.2, 92.8, 8, 3.6, 155)
    best_par <- c(16.996427, 16.000000, 57.685751, 30.324940, 20.000000, 90.565147, 95.000000, 10.590461, 1.561636, 153.535354)
    best_fn <- -172.642

    return(list(
        name = "alkylation",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}

wright4_problem <- function() {
    # Objective function
    fn <- function(x) {
        (x[1] - 1)^2 +
            (x[1] - x[2])^2 +
            (x[2] - x[3])^3 +
            (x[3] - x[4])^4 +
            (x[4] - x[5])^4
    }

    # Gradient of the objective
    gr <- function(x) {
        d1 <- 2 * (x[1] - 1) + 2 * (x[1] - x[2])
        d2 <- -2 * (x[1] - x[2]) + 3 * (x[2] - x[3])^2
        d3 <- -3 * (x[2] - x[3])^2 + 4 * (x[3] - x[4])^3
        d4 <- -4 * (x[3] - x[4])^3 + 4 * (x[4] - x[5])^3
        d5 <- -4 * (x[4] - x[5])^3
        return(c(d1, d2, d3, d4, d5))
    }

    # Equality constraints
    eq_fn <- function(x) {
        z1 <- x[1] + x[2]^2 + x[3]^3
        z2 <- x[2] - x[3]^2 + x[4]
        z3 <- x[1] * x[5]
        return(c(z1, z2, z3))
    }

    # Jacobian of equality constraints: 3 x 5 matrix
    eq_jac <- function(x) {
        jac <- matrix(0, nrow = 3, ncol = 5)

        # dz1/dx
        jac[1, 1] <- 1
        jac[1, 2] <- 2 * x[2]
        jac[1, 3] <- 3 * x[3]^2
        jac[1, 4] <- 0
        jac[1, 5] <- 0

        # dz2/dx
        jac[2, 1] <- 0
        jac[2, 2] <- 1
        jac[2, 3] <- -2 * x[3]
        jac[2, 4] <- 1
        jac[2, 5] <- 0

        # dz3/dx
        jac[3, 1] <- x[5]
        jac[3, 2] <- 0
        jac[3, 3] <- 0
        jac[3, 4] <- 0
        jac[3, 5] <- x[1]

        return(jac)
    }

    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL

    lower <- rep(-10, 5)
    upper <- rep(10, 5)
    eq_b <- c(2 + 3 * sqrt(2), -2 + 2 * sqrt(2), 2)
    start <- c(1, 1, 1, 1, 1)
    best_par <- c(1.116635, 1.220442, 1.537785, 1.972769, 1.791096)
    best_fn <- 0.02931083

    return(list(
        name = "wright4",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}

wright9_problem <- function() {
    fn <- function(x) {
        10 * x[1] * x[4] -
            6 * x[3] * x[2]^2 +
            x[2] * x[1]^3 +
            9 * sin(x[5] - x[3]) +
            x[5]^4 * x[4]^2 * x[2]^3
    }

    gr <- function(x) {
        df_dx1 <- 10 * x[4] + 3 * x[2] * x[1]^2
        df_dx2 <- -12 * x[3] * x[2] + x[1]^3 + 3 * x[5]^4 * x[4]^2 * x[2]^2
        df_dx3 <- -6 * x[2]^2 - 9 * cos(x[5] - x[3])
        df_dx4 <- 10 * x[1] + 2 * x[5]^4 * x[4] * x[2]^3
        df_dx5 <- 9 * cos(x[5] - x[3]) + 4 * x[5]^3 * x[4]^2 * x[2]^3
        return(c(df_dx1, df_dx2, df_dx3, df_dx4, df_dx5))
    }

    eq_fn <- NULL
    eq_jac <- NULL

    ineq_fn <- function(x) {
        z1 <- x[1]^2 + x[2]^2 + x[3]^2 + x[4]^2 + x[5]^2
        z2 <- x[1]^2 * x[3] - x[4] * x[5]
        z3 <- x[2]^2 * x[4] + 10 * x[1] * x[5]
        return(c(z1, z2, z3))
    }

    ineq_jac <- function(x) {
        # Rows: g1, g2, g3; Columns: partials w.r.t. x1 to x5
        jac <- matrix(0, nrow = 3, ncol = 5)

        # g1
        jac[1, ] <- 2 * x

        # g2
        jac[2, 1] <- 2 * x[1] * x[3]
        jac[2, 2] <- 0
        jac[2, 3] <- x[1]^2
        jac[2, 4] <- -x[5]
        jac[2, 5] <- -x[4]

        # g3
        jac[3, 1] <- 10 * x[5]
        jac[3, 2] <- 2 * x[2] * x[4]
        jac[3, 3] <- 0
        jac[3, 4] <- x[2]^2
        jac[3, 5] <- 10 * x[1]

        return(jac)
    }

    lower <- rep(-5, 5)
    upper <- rep(5, 5)
    ineq_lower <- c(-100, -2, 5)
    ineq_upper <- c(20, 100, 100)
    eq_b <- NULL
    start <- c(1, 1, 1, 1, 1)
    best_par <- c(-0.08145219, 3.69237756, 2.48741102, 0.37713392, 0.17398257)
    best_fn <- -210.4078

    return(list(
        name = "wright9",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}

entropy_problem <- function() {
    # Objective function
    fn <- function(x) {
        m <- length(x)
        f <- -sum(log(x))
        f - log(sqrt(sum((x - 1)^2)) + 0.1)
    }

    # Gradient of the objective function
    gr <- function(x) {
        # Compute vector norm of (x - 1)
        diff <- x - 1
        norm_diff <- sqrt(sum(diff^2))

        # First term: -1 / x
        g1 <- -1 / x

        # Second term: -(x_i - 1) / [norm(x - 1) * (norm(x - 1) + 0.1)]
        g2 <- -diff / (norm_diff * (norm_diff + 0.1))

        return(g1 + g2)
    }

    # Equality constraint: sum(x) == 10
    eq_fn <- function(x) sum(x)

    # Jacobian of equality constraint
    eq_jac <- function(x) matrix(1, nrow = 1, ncol = length(x))

    # Inequality: none
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL

    # Bounds and other values
    n <- 10
    lower <- rep(0, n)
    upper <- rep(1000, n)
    eq_b <- 10
    start <- runif(n, 0, 1000)
    best_fn <- 0.1854782
    best_par <- c(
        2.2801555, 0.8577605, 0.8577605, 0.8577605, 0.8577605,
        0.8577605, 0.8577605, 0.8577605, 0.8577605, 0.8577605
    )

    return(list(
        name = "entropy",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}

box_problem <- function() {
    # Objective function
    fn <- function(x) {
        -x[1] * x[2] * x[3]
    }

    # Gradient of the objective function
    gr <- function(x) {
        c(
            -x[2] * x[3],
            -x[1] * x[3],
            -x[1] * x[2]
        )
    }

    # Equality constraint function: 4*x1*x2 + 2*x2*x3 + 2*x3*x1 == 100
    eq_fn <- function(x) {
        4 * x[1] * x[2] + 2 * x[2] * x[3] + 2 * x[3] * x[1]
    }

    # Jacobian of equality constraint: 1 x 3 matrix
    eq_jac <- function(x) {
        matrix(c(
            4 * x[2] + 2 * x[3],
            4 * x[1] + 2 * x[3],
            2 * x[2] + 2 * x[1]
        ), nrow = 1)
    }

    # No inequality constraints
    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL

    lower <- rep(1, 3)
    upper <- rep(10, 3)
    eq_b <- 100
    start <- c(1.1, 1.1, 9)
    best_par <- c(2.886751, 2.886751, 5.773503)
    best_fn <- -48.11252

    return(list(
        name = "box",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}

rosen_suzuki_problem <- function() {
    # Objective function
    fn <- function(x) {
        x[1]^2 + x[2]^2 + 2 * x[3]^2 + x[4]^2 - 5 * x[1] - 5 * x[2] - 21 * x[3] + 7 * x[4]
    }

    # Gradient of the objective function
    gr <- function(x) {
        c(
            2 * x[1] - 5,
            2 * x[2] - 5,
            4 * x[3] - 21,
            2 * x[4] + 7
        )
    }

    # Inequality constraint function
    ineq_fn <- function(x) {
        z1 <- 8 - x[1]^2 - x[2]^2 - x[3]^2 - x[4]^2 - x[1] + x[2] - x[3] + x[4]
        z2 <- 10 - x[1]^2 - 2 * x[2]^2 - x[3]^2 - 2 * x[4]^2 + x[1] + x[4]
        z3 <- 5 - 2 * x[1]^2 - x[2]^2 - x[3]^2 - 2 * x[1] + x[2] + x[4]
        return(c(z1, z2, z3))
    }

    # Jacobian of the inequality constraints: 3 x 4 matrix
    ineq_jac <- function(x) {
        jac <- matrix(0, nrow = 3, ncol = 4)

        # dz1/dx
        jac[1, 1] <- -2 * x[1] - 1
        jac[1, 2] <- -2 * x[2] + 1
        jac[1, 3] <- -2 * x[3] - 1
        jac[1, 4] <- -2 * x[4] + 1

        # dz2/dx
        jac[2, 1] <- -2 * x[1] + 1
        jac[2, 2] <- -4 * x[2]
        jac[2, 3] <- -2 * x[3]
        jac[2, 4] <- -4 * x[4] + 1

        # dz3/dx
        jac[3, 1] <- -4 * x[1] - 2
        jac[3, 2] <- -2 * x[2] + 1
        jac[3, 3] <- -2 * x[3]
        jac[3, 4] <- 1

        return(jac)
    }

    eq_fn <- NULL
    eq_jac <- NULL
    eq_b <- NULL

    lower <- rep(-10, 4)
    upper <- rep(10, 4)
    ineq_lower <- rep(0, 3)
    ineq_upper <- rep(1000, 3)
    start <- c(1, 1, 1, 1)
    best_par <- c(2.502771e-07, 9.999997e-01, 2.000000e+00, -1.000000e+00)
    best_fn <- -44

    return(list(
        name = "rosen_suzuki",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}


powell_problem <- function() {
    # Objective function
    fn <- function(x) {
        exp(x[1] * x[2] * x[3] * x[4] * x[5])
    }

    # Gradient of the objective function
    gr <- function(x) {
        prod_term <- x[1] * x[2] * x[3] * x[4] * x[5]
        fval <- exp(prod_term)

        c(
            fval * x[2] * x[3] * x[4] * x[5],
            fval * x[1] * x[3] * x[4] * x[5],
            fval * x[1] * x[2] * x[4] * x[5],
            fval * x[1] * x[2] * x[3] * x[5],
            fval * x[1] * x[2] * x[3] * x[4]
        )
    }

    # Equality constraint function (3 constraints)
    eq_fn <- function(x) {
        z1 <- x[1]^2 + x[2]^2 + x[3]^2 + x[4]^2 + x[5]^2
        z2 <- x[2] * x[3] - 5 * x[4] * x[5]
        z3 <- x[1]^3 + x[2]^3
        return(c(z1, z2, z3))
    }

    # Jacobian of equality constraints: 3 x 5 matrix
    eq_jac <- function(x) {
        jac <- matrix(0, nrow = 3, ncol = 5)

        # dz1/dx
        jac[1, ] <- 2 * x

        # dz2/dx
        jac[2, 1] <- 0
        jac[2, 2] <- x[3]
        jac[2, 3] <- x[2]
        jac[2, 4] <- -5 * x[5]
        jac[2, 5] <- -5 * x[4]

        # dz3/dx
        jac[3, 1] <- 3 * x[1]^2
        jac[3, 2] <- 3 * x[2]^2
        jac[3, 3] <- 0
        jac[3, 4] <- 0
        jac[3, 5] <- 0

        return(jac)
    }

    ineq_fn <- NULL
    ineq_jac <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL

    lower <- rep(-10, 5)
    upper <- rep(10, 5)
    eq_b <- c(10, 0, -1)
    start <- c(-2, 2, 2, -1, -1)
    best_par <- c(-1.717144, 1.595710, 1.827245, 0.763643, 0.763643)
    best_fn <- 0.05394985

    return(list(
        name = "powell",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}

himmelblau5_problem <- function() {
    # Objective function
    fn <- function(x) {
        (x[1] - 5)^2 + (x[2] - 5)^2
    }

    # Gradient of objective
    gr <- function(x) {
        c(2 * (x[1] - 5), 2 * (x[2] - 5))
    }

    # Equality constraint: x1 - 3 * x2 = 0
    eq_fn <- function(x) {
        x[1] - 3 * x[2]
    }

    eq_jac <- function(x) {
        matrix(c(1, -3), nrow = 1)
    }

    # Inequality constraint: x1^2 + x2^2 <= 25
    ineq_fn <- function(x) {
        x[1]^2 + x[2]^2
    }

    ineq_jac <- function(x) {
        matrix(c(2 * x[1], 2 * x[2]), nrow = 1)
    }

    lower <- rep(-5, 2)
    upper <- rep(5, 2)
    eq_b <- 0
    ineq_lower <- -1000
    ineq_upper <- 25
    start <- c(1, 1)
    best_par <- c(4.74342, 1.58114)
    best_fn <- 11.7544

    return(list(
        name = "himmelblau5",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}


hs118_problem <- function() {
    # Objective function
    fn <- function(x) {
        obj <- 0
        for (i in 0:4) {
            xi <- x[(3 * i + 1):(3 * i + 3)]
            obj <- obj + (2.3 * xi[1] + 0.0001 * xi[1]^2 +
                              1.7 * xi[2] + 0.0001 * xi[2]^2 +
                              2.2 * xi[3] + 0.00015 * xi[3]^2)
        }
        return(obj)
    }

    gr <- function(x) {
        g <- numeric(15)
        for (k in 0:4) {
            i <- 3 * k + 1
            g[i]     <- 2.3 + 0.0002 * x[i]
            g[i + 1] <- 1.7 + 0.0002 * x[i + 1]
            g[i + 2] <- 2.2 + 0.0003 * x[i + 2]
        }
        return(g)
    }

    # Inequality constraints
    ineq_fn <- function(x) {
        c(
            # Temporal order constraints (lower and upper bounds handled separately)
            x[4]  - x[1]  + 7,
            x[7]  - x[4]  + 7,
            x[10] - x[7]  + 7,
            x[13] - x[10] + 7,

            x[5]  - x[2]  + 7,
            x[8]  - x[5]  + 7,
            x[11] - x[8]  + 7,
            x[14] - x[11] + 7,

            x[6]  - x[3]  + 7,
            x[9]  - x[6]  + 7,
            x[12] - x[9]  + 7,
            x[15] - x[12] + 7,

            # Demand constraints (LHS ≥ threshold)
            60 - (x[1] + x[2] + x[3]),
            50 - (x[4] + x[5] + x[6]),
            70 - (x[7] + x[8] + x[9]),
            85 - (x[10] + x[11] + x[12]),
            100 - (x[13] + x[14] + x[15])
        )
    }

    ineq_jac <- function(x) {
        J <- matrix(0, nrow = 17, ncol = 15)

        # Time progression constraints (12 rows)
        idx_pairs <- list(
            c(1, 4), c(4, 7), c(7, 10), c(10, 13),
            c(2, 5), c(5, 8), c(8, 11), c(11, 14),
            c(3, 6), c(6, 9), c(9, 12), c(12, 15)
        )

        for (i in seq_along(idx_pairs)) {
            idx1 <- idx_pairs[[i]][1]
            idx2 <- idx_pairs[[i]][2]
            J[i, idx2] <- 1
            J[i, idx1] <- -1
        }

        # Demand constraints (5 rows)
        for (k in 0:4) {
            row <- 12 + k + 1
            cols <- 3 * k + 1:3
            J[row, cols] <- -1
        }

        return(J)
    }

    # Bounds
    lower <- c(8, 43, 3,
               rep(0, 12))
    upper <- c(21, 57, 16,
               rep(c(90, 120, 60), 4))

    # Initial values
    start <- c(20, 55, 15,
               20, 60, 20,
               20, 60, 20,
               20, 60, 20,
               20, 60, 20)

    # Known best solution
    best_fn <- 664.82045
    best_par <- NULL  # Not given exactly — optional

    return(list(
        name = "hs118",
        fn = fn,
        gr = gr,
        eq_fn = NULL,
        eq_jac = NULL,
        eq_b = NULL,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = c(rep(0, 12), rep(-1000, 5)),  # >= 0 for time constraints, >= threshold for demand
        ineq_upper = c(rep(13, 12), rep(0, 5)),    # <= 13 for time constraints, demand constraint expressed as ≤ 0
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}


hs119_problem <- function() {
    # Constants
    c_vals <- c(2.5, 1.1, -3.1, -3.5, 1.3, 2.1, 2.3, -1.5)

    # Objective: sum of 46 nonlinear terms
    fn <- function(x) {
        square_expr <- function(z) (z^2 + z + 1)
        t <- numeric(46)
        t[1] <- square_expr(x[1]) * square_expr(x[1])
        t[2] <- square_expr(x[1]) * square_expr(x[4])
        t[3] <- square_expr(x[1]) * square_expr(x[7])
        t[4] <- square_expr(x[1]) * square_expr(x[8])
        t[5] <- square_expr(x[1]) * square_expr(x[16])
        t[6] <- square_expr(x[2]) * square_expr(x[2])
        t[7] <- square_expr(x[2]) * square_expr(x[3])
        t[8] <- square_expr(x[2]) * square_expr(x[7])
        t[9] <- square_expr(x[2]) * square_expr(x[10])
        t[10] <- square_expr(x[3]) * square_expr(x[3])
        t[11] <- square_expr(x[3]) * square_expr(x[7])
        t[12] <- square_expr(x[3]) * square_expr(x[9])
        t[13] <- square_expr(x[3]) * square_expr(x[10])
        t[14] <- square_expr(x[3]) * square_expr(x[14])
        t[15] <- square_expr(x[4]) * square_expr(x[4])
        t[16] <- square_expr(x[4]) * square_expr(x[7])
        t[17] <- square_expr(x[4]) * square_expr(x[11])
        t[18] <- square_expr(x[4]) * square_expr(x[15])
        t[19] <- square_expr(x[5]) * square_expr(x[5])
        t[20] <- square_expr(x[5]) * square_expr(x[6])
        t[21] <- square_expr(x[5]) * square_expr(x[10])
        t[22] <- square_expr(x[5]) * square_expr(x[12])
        t[23] <- square_expr(x[5]) * square_expr(x[16])
        t[24] <- square_expr(x[6]) * square_expr(x[6])
        t[25] <- square_expr(x[6]) * square_expr(x[8])
        t[26] <- square_expr(x[6]) * square_expr(x[15])
        t[27] <- square_expr(x[7]) * square_expr(x[7])
        t[28] <- square_expr(x[7]) * square_expr(x[11])
        t[29] <- square_expr(x[7]) * square_expr(x[13])
        t[30] <- square_expr(x[8]) * square_expr(x[8])
        t[31] <- square_expr(x[8]) * square_expr(x[10])
        t[32] <- square_expr(x[8]) * square_expr(x[15])
        t[33] <- square_expr(x[9]) * square_expr(x[9])
        t[34] <- square_expr(x[9]) * square_expr(x[12])
        t[35] <- square_expr(x[9]) * square_expr(x[16])
        t[36] <- square_expr(x[10]) * square_expr(x[10])
        t[37] <- square_expr(x[10]) * square_expr(x[14])
        t[38] <- square_expr(x[11]) * square_expr(x[11])
        t[39] <- square_expr(x[11]) * square_expr(x[13])
        t[40] <- square_expr(x[11]) * square_expr(x[12])
        t[41] <- square_expr(x[12]) * square_expr(x[14])
        t[42] <- square_expr(x[13]) * square_expr(x[13])
        t[43] <- square_expr(x[13]) * square_expr(x[14])
        t[44] <- square_expr(x[14]) * square_expr(x[14])
        t[45] <- square_expr(x[15]) * square_expr(x[15])
        t[46] <- square_expr(x[16]) * square_expr(x[16])
        return(sum(t))
    }

    gr <- function(x) {
        fi  <- x^2 + x + 1          # f_i(x)
        dfi <- 2*x + 1              # f'_i(x)

        ## list the 46 (a,b) index pairs exactly as used in `fn`
        idx <- matrix(c(
            1,1,  1,4, 1,7, 1,8, 1,16,
            2,2,  2,3, 2,7, 2,10,
            3,3,  3,7, 3,9, 3,10, 3,14,
            4,4,  4,7, 4,11, 4,15,
            5,5,  5,6, 5,10, 5,12, 5,16,
            6,6,  6,8, 6,15,
            7,7,  7,11, 7,13,
            8,8,  8,10, 8,15,
            9,9,  9,12, 9,16,
            10,10, 10,14,
            11,11, 11,13, 11,12,
            12,14,
            13,13, 13,14,
            14,14,
            15,15,
            16,16
        ), byrow = TRUE, ncol = 2)

        g <- numeric(16)

        for (k in seq_len(nrow(idx))) {
            i <- idx[k, 1]
            j <- idx[k, 2]

            if (i == j) {
                ## diagonal term f_i^2
                g[i] <- g[i] + 2 * fi[i] * dfi[i]
            } else {
                ## off-diagonal product f_i * f_j
                g[i] <- g[i] + fi[j] * dfi[i]
                g[j] <- g[j] + fi[i] * dfi[j]
            }
        }
        g
    }

    # Equality constraints: s[1:8] = c[1:8]
    eq_fn <- function(x) {
        c(
            0.22*x[1] + 0.2*x[2] + 0.19*x[3] + 0.25*x[4] + 0.15*x[5] + 0.11*x[6] + 0.12*x[7] + 0.13*x[8] + x[9],
            -1.46*x[1] -1.3*x[3] + 1.82*x[4] -1.15*x[5] + 0.8*x[7] + x[10],
            1.29*x[1] -0.89*x[2] -1.16*x[5] -0.96*x[6] -0.49*x[8] + x[11],
            -1.1*x[1] -1.06*x[2] + 0.95*x[3] -0.54*x[4] -1.78*x[6] -0.41*x[7] + x[12],
            -1.43*x[4] + 1.51*x[5] + 0.59*x[6] -0.33*x[7] -0.43*x[8] + x[13],
            -1.72*x[2] -0.33*x[3] + 1.62*x[5] + 1.24*x[6] + 0.21*x[7] -0.26*x[8] + x[14],
            1.12*x[1] + 0.31*x[4] + 1.12*x[7] -0.36*x[9] + x[15],
            0.45*x[2] + 0.26*x[3] -1.1*x[4] + 0.58*x[5] -1.03*x[7] + 0.1*x[8] + x[16]
        )
    }

    # Jacobian of the equality constraints (8x16)
    eq_jac <- function(x) {
        J <- matrix(0, nrow = 8, ncol = 16)
        # Fill based on coefficients from model
        J[1, 1:9] <- c(0.22, 0.2, 0.19, 0.25, 0.15, 0.11, 0.12, 0.13, 1)
        J[2, c(1, 3, 4, 5, 7, 10)] <- c(-1.46, -1.3, 1.82, -1.15, 0.8, 1)
        J[3, c(1, 2, 5, 6, 8, 11)] <- c(1.29, -0.89, -1.16, -0.96, -0.49, 1)
        J[4, c(1, 2, 3, 4, 6, 7, 12)] <- c(-1.1, -1.06, 0.95, -0.54, -1.78, -0.41, 1)
        J[5, c(4, 5, 6, 7, 8, 13)] <- c(-1.43, 1.51, 0.59, -0.33, -0.43, 1)
        J[6, c(2, 3, 5, 6, 7, 8, 14)] <- c(-1.72, -0.33, 1.62, 1.24, 0.21, -0.26, 1)
        J[7, c(1, 4, 7, 9, 15)] <- c(1.12, 0.31, 1.12, -0.36, 1)
        J[8, c(2, 3, 4, 5, 7, 8, 16)] <- c(0.45, 0.26, -1.1, 0.58, -1.03, 0.1, 1)
        return(J)
    }

    lower <- rep(0, 16)
    upper <- rep(5, 16)
    start <- rep(2, 16)
    eq_b <- c(2.5, 1.1, -3.1, -3.5, 1.3, 2.1, 2.3, -1.5)

    best_par <- c(
        0.03984735, 0.7919832, 0.2028703, 0.8443579,
        1.126991, 0.9347387, 1.681962, 0.1553009,
        1.567870, 0, 0, 0,
        0.6602041, 0, 0.6742559, 0
    )
    best_fn <- 244.899698

    return(list(
        name = "hs119",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_jac = eq_jac,
        eq_b = eq_b,
        ineq_fn = NULL,
        ineq_jac = NULL,
        ineq_lower = NULL,
        ineq_upper = NULL,
        lower = lower,
        upper = upper,
        start = start,
        best_par = best_par,
        best_fn = best_fn
    ))
}



hs110_problem <- function()
{
    # Objective function: sum of obj[1:11]
    fn <- function(x) {
        # x is length 10
        # Intermediates
        prod_vec <- numeric(11)
        prod_vec[1] <- 1
        for (i in 2:11) {
            prod_vec[i] <- x[i-1] * prod_vec[i-1]
        }
        obj1_10 <- (log(x - 2))^2 + (log(10 - x))^2
        obj11 <- -prod_vec[11]^0.2
        sum(obj1_10) + obj11
    }
    # Analytic gradient
    gr <- function(x) {
        grad <- numeric(10)
        # Compute intermediates for obj11
        prod_vec <- numeric(11)
        prod_vec[1] <- 1
        for (i in 2:11) {
            prod_vec[i] <- x[i-1] * prod_vec[i-1]
        }
        log_term1 <- log(x - 2)
        log_term2 <- log(10 - x)
        # Derivatives of obj[1:10]
        grad_obj <- 2 * log_term1 / (x - 2) - 2 * log_term2 / (10 - x)
        # Derivative of obj11
        for (j in 1:10) {
            if (prod_vec[11] > 0) {
                dprodj <- prod_vec[11] / x[j]
                grad[j] <- grad_obj[j] + (-0.2) * prod_vec[11]^(-0.8) * dprodj
            } else {
                # Outside feasible region, set NaN
                grad[j] <- NaN
            }
        }
        grad
    }
    lower <- rep(2.001, 10)
    upper <- rep(9.999, 10)
    start <- rep(9, 10)
    eq_fn <- NULL
    eq_jac <- NULL
    ineq_fn <- NULL
    ineq_jac <- NULL
    eq_b <- NULL
    ineq_lower <- NULL
    ineq_upper <- NULL
    list(
        name = "hs110",
        fn = fn,
        gr = gr,
        eq_fn = eq_fn,
        eq_b = eq_b,
        eq_jac = eq_jac,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = -45.77846971,
        best_par = NULL
    )
}

garch_problem <- function()
{
    n <- 1500
    set.seed(100)
    z <- rnorm(n)
    x <- numeric(n)
    sigma2 <- numeric(n)
    x[1] <- 0.1315172
    sigma2[1] <- 0.2211
    mu <- -0.006184353
    omega <- 0.010760430
    alpha <- 0.153408326
    beta <- 0.805877422
    for (t in 2:n) {
        sigma2[t] <- omega + alpha * (x[t - 1] - mu)^2 + beta * sigma2[t - 1]
        x[t] <- mu + z[t] * sqrt(sigma2[t])
    }

    returns <- x
    garch11_negloglik <- function(par) {
        # Parameterization: par = c(mu, omega, alpha, beta)
        mu    <- par[1]
        omega <- par[2]
        alpha <- par[3]
        beta  <- par[4]

        T <- length(returns)
        eps <- returns - mu
        sig2 <- numeric(T)

        # Initialize sigma^2 with unconditional variance
        sig2[1] <- mean(eps^2)

        for (t in 2:T) {
            sig2[t] <- omega + alpha * eps[t - 1]^2 + beta * sig2[t - 1]
        }

        nll <- 0.5 * sum(log(2*pi) + log(sig2) + eps^2 / sig2)
        return(nll)
    }

    garch11_grad <- function(par) {
        mu    <- par[1]
        omega <- par[2]
        alpha <- par[3]
        beta  <- par[4]
        T <- length(returns)
        eps <- returns - mu
        sig2 <- numeric(T)
        dsig2_domega <- numeric(T)
        dsig2_dalpha <- numeric(T)
        dsig2_dbeta  <- numeric(T)
        dsig2_dmu    <- numeric(T)

        # Initialize with sample variance of residuals
        sig2[1] <- mean(eps^2)
        dsig2_domega[1] <- 0
        dsig2_dalpha[1] <- 0
        dsig2_dbeta[1]  <- 0
        dsig2_dmu[1]    <- -2 * mean(eps)

        # Recursion
        for (t in 2:T) {
            dsig2_domega[t] <- 1 + beta * dsig2_domega[t - 1]
            dsig2_dalpha[t] <- eps[t - 1]^2 + beta * dsig2_dalpha[t - 1]
            dsig2_dbeta[t]  <- sig2[t - 1] + beta * dsig2_dbeta[t - 1]
            dsig2_dmu[t]    <- -2 * alpha * eps[t - 1] + beta * dsig2_dmu[t - 1]
            sig2[t] <- omega + alpha * eps[t - 1]^2 + beta * sig2[t - 1]
        }

        # Gradients
        dmu <- -sum(eps / sig2) + sum(0.5 * (1 / sig2 - (eps^2) / (sig2^2)) * dsig2_dmu)
        domega <- sum((1 / (2 * sig2) - eps^2 / (2 * sig2^2)) * dsig2_domega)
        dalpha <- sum((1 / (2 * sig2) - eps^2 / (2 * sig2^2)) * dsig2_dalpha)
        dbeta  <- sum((1 / (2 * sig2) - eps^2 / (2 * sig2^2)) * dsig2_dbeta)

        return(c(dmu, domega, dalpha, dbeta))
    }
    garch11_ineq_fn <- function(par) {
        # par = (mu, omega, alpha, beta)
        alpha <- par[3]
        beta  <- par[4]
        return(alpha + beta - 1)
    }

    garch11_ineq_jac <- function(par) {
        # Jacobian: returns a 1x4 row vector
        jac <- numeric(4)
        jac[3] <- 1  # derivative wrt alpha
        jac[4] <- 1  # derivative wrt beta
        return(matrix(jac, nrow = 1))
    }

    lower <- c(-1, 1e-12, 1e-12, 1e-12)
    upper <- c(1, 2, 1, 1)
    start <- c(mu = 0, omega = 0.1, alpha = 0.05, beta = 0.9)
    eq_fn <- NULL
    eq_jac <- NULL
    ineq_fn <- garch11_ineq_fn
    ineq_jac <- garch11_ineq_jac
    eq_b <- NULL
    ineq_lower <- -1
    ineq_upper <- 0
    list(
        name = "garch",
        fn = garch11_negloglik,
        gr = garch11_grad,
        eq_fn = NULL,
        eq_b = NULL,
        eq_jac = NULL,
        ineq_fn = ineq_fn,
        ineq_jac = ineq_jac,
        ineq_lower = ineq_lower,
        ineq_upper = ineq_upper,
        lower = lower,
        upper = upper,
        start = start,
        best_fn = 1074.36,
        best_par = c(-0.006184353, 0.010760430, 0.153408326, 0.805877422)
    )
}

---
title: "SOLNP Test Suite"
output: 
    rmarkdown::html_vignette:
        css: custom.css
        code_folding: hide
        citation_package: natbib
        toc: yes
bibliography: references.bib
bibliography-style: apalike
natbiboptions: round
vignette: >
  %\VignetteIndexEntry{SOLNP Test Suite}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo=FALSE}
library(Rsolnp)
```

## Introduction


The purpose of the new SOLNP Test Suite is to provide a comprehensive, 
standardized environment for benchmarking and validating nonlinear optimization 
algorithms, with a particular focus on the SOLNP family of solvers. The suite is 
anchored around the well-known Hock-Schittkowski (HS) test problems, a diverse 
collection of nonlinear programming challenges that have become a benchmark 
standard in the optimization literature.

This suite currently implements 64 of the HS problems, spanning a range of 
complexities, constraint types, and objective function characteristics. Future 
updates will expand coverage to include the entire set of 306 Hock-Schittkowski 
problems, ensuring a thorough and robust testing ground for both algorithmic 
development and comparative analysis.

The problems are currently stored in a form compatible with the SOLNP solver, 
enabling immediate use for benchmarking and analysis. However, recognizing that 
nonlinear programming (NLP) problems can be specified in various formats—sometimes 
with differing conventions for constraints, bounds, and objective function 
representation—the suite includes a utility function called `solnp_standardize_problem`
which rewrites the test in the NLP standard form (e.g. as used by [nloptr](https://CRAN.R-project.org/package=nloptr)).

A distinctive feature of several HS problems is that their optimal solutions 
often occur on or very near the boundary of the feasible region, where one or more 
inequality constraints are active. In other words, the optimal solution frequently 
coincides with points where the constraints are "tight", and further movement 
would violate feasibility.

This aspect is of particular importance for algorithm development, as it tests 
a solver's ability to:

  * Correctly identify the active set of constraints at the optimum,
  * Handle boundary solutions robustly, and
  * Maintain numerical stability and feasibility in the presence of active inequality constraints.

It should be noted that SOLNP does not employ an active set strategy. Instead, 
its augmented Lagrangian approach treats constraints via penalty terms, and the 
optimizer does not explicitly identify or maintain sets of active constraints 
during the iterations. As a result, SOLNP solutions typically remain in the 
interior of the feasible region whenever possible, only approaching the boundary 
when required for optimality.



## Usage

The function `solnp_problems_table` lists the problem currently implemented:

```{r}
head(solnp_problems_table())
```


whilst the function `solnp_problem_suite` can either return a specific problem
or all problems within a Suite:

```{r}
prob <- solnp_problem_suite(suite = "Hock-Schittkowski", number = 60)
sol <- csolnp(prob$start, fn = prob$fn, gr = prob$gr, eq_fn = prob$eq_fn, 
              eq_b = prob$eq_b, eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, 
              ineq_lower = prob$ineq_lower, ineq_upper = prob$ineq_upper, 
              ineq_jac = prob$ineq_jac, lower = prob$lower, upper = prob$upper, 
              control = list(trace = 0, min_iter = 1000, max_iter = 300, tol = 1e-8, rho = 1))
print(c("solnp" = sol$objective, "benchmark" = prob$best_fn))
print(rbind("solnp" = sol$pars, "benchmark" = prob$best_par))
```

---
title: "Introduction to SOLNP ver 2.0.0"
output: 
    rmarkdown::html_vignette:
        css: custom.css
        code_folding: hide
        citation_package: natbib
        toc: yes
bibliography: references.bib
bibliography-style: apalike
natbiboptions: round
vignette: >
  %\VignetteIndexEntry{Introduction to SOLNP ver 2.0.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo=FALSE}
library(Rsolnp)
```

## Introduction

The SOLNP algorithm, first proposed by @Ye1989, is a Sequential Quadratic 
Programming (SQP) approach designed to solve general nonlinear optimization 
problems with both equality and inequality constraints. In 2010, it was 
implemented in R as the Rsolnp package, originally to fill the need for a robust 
and flexible constrained optimizer required by the author’s own GARCH modeling 
package, [rugarch](https://cran.r-project.org/package=rugarch). Since then, 
Rsolnp has remained one of the few readily available and user-friendly options 
for nonlinear constrained optimization in R.

Despite the growth of R’s ecosystem in statistics and data science, the development 
and availability of modern nonlinear constrained optimization solvers in R has 
lagged behind other languages, with notable exceptions being [nloptr](https://cran.r-project.org/package=nloptr) (which wraps NLopt solvers) and [ipoptr](https://github.com/jyypma/ipoptr) (an 
R interface to the IPOPT solver)^[the latter is no longer available on CRAN due 
to licensing issues]. Moreover, in the nonlinear programming (NLP) space, most 
state-of-the-art solvers remain closed source and commercial, in contrast to other 
algorithmic advances that have transitioned from academia to open-source availability.

Rsolnp does not claim to be competitive with commercial grade nonlinear solvers,
but it has generally served its purpose reasonably well for small to medium 
sized problems.

## Problem Statement

The general nonlinear optimization problem addressed by SOLNP can be stated as follows:

$$
\begin{align}
& \min_{x \in \mathbb{R}^n} \quad f(x) \\\\
& \text{subject to:} \\\\
& \quad l_g \leq g(x) \leq u_g \qquad \text{(general inequality constraints)} \\\\
& \quad h(x) = B \qquad \text{(equality constraints with bound vector \(B\))} \\\\
& \quad x^L \leq x \leq x^U \qquad \text{(variable bounds)}
\end{align}
$$

where:

 - $f(x)$: Nonlinear objective function to minimize
 - $g(x)$: Vector of general nonlinear inequality constraint functions, with lower and upper bounds $l_g$, $u_g$
 - $h(x)$: Vector of nonlinear equality constraint functions, required to be equal to vector B
 - $x^L$, $x^U$: Lower and upper bounds for the variables

Note:

 - Standard equality constraints of the form $h(x) = 0$ can be recovered by setting B = 0.
 - General inequality constraints allow for two-sided bounds (e.g., $l_g < g(x) < u_g$), 
 not just upper or lower bounds individually.


### Inequalities

Each inequality constraint $l_i \leq g_i(x) \leq u_i$ is transformed into two 
equality constraints via slack variables. Introduce slack variable $s_i \geq 0$, 
then define:

$$
h_i(x, s_i) = g_i(x) - s_i = l_i \quad \text{if } g_i(x) \geq l_i \\
h_i(x, s_i) = g_i(x) + s_i = u_i \quad \text{if } g_i(x) \leq u_i
$$


This can be unified into:

$$
g_i(x) - s_i = l_i \quad \text{and} \quad g_i(x) + s_i = u_i
$$


So every bounded inequality becomes an equality constraint with a slack, and the 
slack becomes part of the optimization variable vector.

In the actual implementation, if a constraint is double-bounded, a 
"penalty equality constraint" is introduced $h_i(x, s_i) := g_i(x) - s_i = m_i$
and add a penalty term to objective if $s_i < 0$.

### Slack Variable Embedding

The augmented optimization vector becomes:

$\tilde{x} = \begin{bmatrix} x \\ s \end{bmatrix}$

And the original constraints are reformulated into a system of equalities only:

$$
\tilde{g}(x, s) =
\begin{bmatrix}
g_{\text{eq}}(x) \\
g_{\text{ineq}}(x) - s - l \\
g_{\text{ineq}}(x) + s - u
\end{bmatrix} = 0
$$

### Optimization via Augmented Lagrangian

Unlike traditional SQP, SOLNP does not explicitly solve for the Karush–Kuhn–Tucker 
(KKT) conditions. Instead, it uses a partial augmented Lagrangian:

$$
L(x, \lambda, \rho) = f(x) + \lambda^\top g(x) + \frac{\rho}{2} \|g(x)\|^2
$$

where:

  - $g(x)$ contains only equality constraints (as inequalities are transformed)
  - $\lambda$ are Lagrange multipliers (estimated iteratively)
  - $\rho$ is the penalty parameter

### No Stationarity-Based KKT Step

Unlike SQP or interior-point methods, SOLNP does not solve the full KKT system 
at each iteration, instead using:

  - a quadratic local model
  - a trust-region style step
  - penalty-based adjustment (via $\rho$)

The stationarity check is only implicit in terms of convergence of $\nabla f + J^\top \lambda \to 0$, 
not enforced directly.

### Why Higher Tolerances Are Not Preferred

Setting the tolerance in SOLNP (or augmented Lagrangian solvers) much tighter 
than 1e-8—such as 1e-12—can actually degrade solution quality because the algorithm 
becomes sensitive to numerical round-off and finite precision errors, especially 
in gradient and Hessian computations or matrix solves. As the tolerance approaches 
the limits of double-precision arithmetic, the solver may "chase noise" rather than 
real improvements, leading to erratic or even worse solutions, with unreliable 
convergence and possible violation of constraints. For most problems, practical 
and reliable results are achieved with tolerances in the 1e-6 to 1e-8 range.


## C++ Version

The original solnp function did not make use of analytic gradients or Jacobians,
and was written entirely in R, a direct translation from the Matlab code. Since 
version 2.0.0, a new function `csolnp`, written in C++^[Using [Rcpp](https://cran.r-project.org/package=Rcpp) 
and [RcppArmadillo](https://cran.r-project.org/package=RcppArmadillo)] allows the user 
to provide analytic gradient and/or Jacobians for the inequality and equality constraints. 
In the absence of analytic functions, finite differences are used using the functions from the [numDeriv](https://cran.r-project.org/package=numDeriv) package.

The function signature for the csolnp function is now slightly different from that
of solnp in both content and code styling:

```{r}
args(csolnp)
```

vs 

```{r}
args(solnp)
```


Speedup of anywhere up to 10x can be expected for some problems. Additionally, 
certain enhancements have been made in the C++ code in regards to criteria for
early termination to avoid stalling in the line search phase.

## Multi-Start Solver

A new function `csolnp_ms` implements a multi-start strategy, similar to the 
existing `gosolnp` function. However, it differs in the way starting candidate 
values are calculated, with the original function using rejection sampling to 
find feasible starting points, whilst the new function uses a combination of 
buffered box sampling and a fast interior-point feasibility projection. This 
approach efficiently generates initial values that strictly satisfy parameter 
bounds and nonlinear constraints by first sampling points slightly away from the 
boundaries and then rapidly projecting them into the feasible region using a 
penalized minimization. This results in more reliable and diverse initial 
candidates for multi-start optimization, especially in problems with complex or 
tightly constrained feasible regions.

## Test Suite

As of version 2.0.0, a new test suite based on @Hock1980 and @Schittkowski2012
has been included, translated from the Fortran codes 
[here](https://klaus-schittkowski.de/tpnp.htm). Currently, about 60 problems
of the 306 have been translated, and it is the intention of the author to 
eventually translate all the tests.

Each test, returns a list with the following information:

 - fn : the objective function
 - gr : the analytic gradient function
 - eq_fn : the equality function
 - eq_b : the equality bounds
 - eq_jac : the Jacobian of the equality function
 - ineq_fn : the inequality function
 - ineq_lower : the lower bounds for the inequalities
 - ineq_upper : the upper bounds for the inequalities
 - ineq_jac : the Jacobian of the inequality function
 - lower : lower parameter bounds
 - upper : upper parameter bounds
 - start : initialization parameters
 - best_fn : best known optimal objective
 - best_par : best known optimal parameters
 - name : test name
 
The suite can be called at once or by reference to a specific test, as illustrated
below.

```{r}
prob <- solnp_problem_suite(number = 10)
sol <- csolnp(pars = prob$start, fn = prob$fn, gr = prob$gr, eq = prob$eq_fn, eq_b = prob$eq_b, 
              eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, ineq_lower = prob$ineq_lower, 
              ineq_upper = prob$ineq_upper, ineq_jac = prob$ineq_jac, lower = prob$lower,
              upper = prob$upper)
print(prob$name)
print(c("convergence" = sol$convergence))
print(c("csolnp objective" = sol$objective, "best objective" = prob$best_fn))
print(sol$elapsed_time)
```

We also test a more specific problem based on the GARCH(1,1) model with analytic
derivatives:

```{r}
prob <- solnp_problem_suite(suite = "Other", number = 4)
sol <- csolnp(pars = prob$start, fn = prob$fn, gr = prob$gr, eq = prob$eq_fn, eq_b = prob$eq_b, 
              eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, ineq_lower = prob$ineq_lower, 
              ineq_upper = prob$ineq_upper, ineq_jac = prob$ineq_jac, lower = prob$lower,
              upper = prob$upper)
print(prob$name)
print(c("convergence" = sol$convergence))
print(c("csolnp objective" = sol$objective, "best objective" = prob$best_fn))
print(sol$elapsed_time)
```

Finally, we take a look at a `solver-killer` problem, HS55 (from the Hock-Schittkowski suite),
which poses the following challenges for NLP solvers:

 * Extremely Flat Objective Near Optimum: The objective function in HS55 is nearly constant (very flat) near the solution, so gradients become very small and the solver can struggle to find a meaningful descent direction. This leads to slow progress or stalling.
 * Tight and Nearly Active Nonlinear Constraints: The feasible region is defined by nonlinear constraints that are nearly active at the optimum, meaning any small move risks violating feasibility. This makes line search and step acceptance delicate.
  * Ill-conditioning and Sensitivity: The combination of flatness and tight constraints results in an ill-conditioned optimization landscape. Small numerical errors or poorly scaled steps can cause the solver to jump out of the feasible region, or make the algorithm oscillate or stagnate.
  * Difficult Constraint Jacobians: The constraint gradients in HS55 can be nearly linearly dependent or poorly scaled near the optimum, leading to numerical instability when solving the KKT system or updating multipliers.
  * Slow Convergence/Failure to Converge: Many solvers take thousands of iterations and may still not converge to high accuracy, or may falsely report convergence with a suboptimal or infeasible solution.

Instead of using `csolnp` which fails in this case, we try out the 
`csolnp_ms` approach.


```{r,warning=FALSE}
prob <- solnp_problem_suite(number = 55)
sol <- csolnp_ms(prob$fn, gr = prob$gr, eq_fn = prob$eq_fn, eq_b = prob$eq_b, 
              eq_jac = prob$eq_jac, ineq_fn = prob$ineq_fn, ineq_lower = prob$ineq_lower, 
              ineq_upper = prob$ineq_upper, ineq_jac = prob$ineq_jac, lower = prob$lower,
              upper = prob$upper, n_candidates = 200, penalty = 1e5, 
              control = list(min_iter = 1000, max_iter = 100, tol = 1e-8),
              seed = 300)

print(paste0("Solution : ", round(sol$objective,3), " | Best Objective :", round(prob$best_fn,3)))
print(paste0("Equaility Violation : ", round(sol$kkt_diagnostics$eq_violation,3)))
```

The result is is not too bad, though the equality violation is still > 1e-6 which
is what we were aiming for.


## Differences with full Active-Set/SQP solvers

Below is a short summary of the key differences between the SOLNP algorithm
and full Active-set/SQP methods:

| Aspect                | SOLNP                                    | Active-Set SQP / QP                  |
|-----------------------|------------------------------------------|--------------------------------------|
| Inequality handling   | Slacks, penalty, all treated equally     | Active set, complementarity enforced |
| Subproblem            | Least squares (QR, no QP)                | Quadratic programming                |
| KKT stationarity      | Not always enforced                      | Enforced at every step               |
| Complementarity       | Not enforced                             | Enforced at every step               |
| Scaling sensitivity   | Can be significant                       | Generally better                     |
| Globalization         | Penalty parameter tuning crucial         | Line search/trust region, more robust|
| Numerical behavior    | Simple, robust for easy problems         | More complex, better for hard ones   |



## Conclusion

The C++ rewrite of solnp, inclusion of analytic derivative options and the new
test suite should hopefully elevate the quality of the Rsolnp package. Future
enhancements will include expansion of the test suite and a deeper investigation
into some of the underlying approaches used and ways to enhance the solver.


## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmark_suite.R
\name{solnp_standardize_problem}
\alias{solnp_standardize_problem}
\title{Standardize an Optimization Problem to NLP Standard Form}
\usage{
solnp_standardize_problem(prob)
}
\arguments{
\item{prob}{A list specifying the problem in SOLNP-compatible format, with components
\code{fn}, \code{eq_fn}, \code{eq_jac}, \code{eq_b}, \code{ineq_fn}, \code{ineq_jac},
\code{ineq_lower}, \code{ineq_upper}, and others.}
}
\value{
A list with the same structure as the input, but with \code{eq_fn} and \code{ineq_fn}
standardized to the forms \eqn{e(x) = 0} and \eqn{g(x) \leq 0}, and with
\code{eq_b}, \code{ineq_lower}, and \code{ineq_upper} removed.
}
\description{
Converts a problem specified with two-sided inequalities and nonzero equality right-hand sides
to the standard nonlinear programming (NLP) form.
}
\details{
The standard form given by the following set of equations:

\deqn{ \min_x\ f(x) }
\deqn{ \textrm{subject to}\quad e(x) = 0 }
\deqn{ \qquad\qquad\qquad g(x) \leq 0 }

Specifically:
\itemize{
\item All equality constraints are standardized to \eqn{e(x) = e(x) - b = 0}
\item Each two-sided inequality \eqn{l \leq g(x) \leq u} is converted to one or two
one-sided constraints: \eqn{l - g(x) \leq 0}, \eqn{g(x) - u \leq 0}
}

The returned problem object has all equalities as \eqn{e(x) = 0}, all inequalities as \eqn{g(x) \leq 0},
and any right-hand side or bounds are absorbed into the standardized constraint functions.
}
\examples{
# Alkylation problem
p <- solnp_problem_suite(suite = "Other", number = 1)
ps <- solnp_standardize_problem(p)
ps$eq_fn(ps$start)    # standardized equalities: e(x) = 0
ps$ineq_fn(ps$start)  # standardized inequalities: g(x) <= 0

}
\seealso{
\code{\link{solnp_problem_suite}}
}

\name{benchmarkids}
\alias{benchmarkids}
\title{
The Rsolnp Benchmark Problems Suite problem id's.
}
\description{
Returns the id's of available benchmark in the Rsolnp Benchmark Problems Suite.
}
\usage{
benchmarkids()
}
\value{
A character vector of problem id's.
}
\references{
W.Hock and K.Schittkowski, \emph{Test Examples for Nonlinear Programming Codes}, 
Lecture Notes in Economics and Mathematical Systems. Springer Verlag, 1981.\cr 
Y.Ye, \emph{Interior algorithms for linear, quadratic, and linearly constrained 
non linear programming}, PhD Thesis, Department of EES Stanford University, 
Stanford CA.\cr
}
\author{
Alexios Ghalanos and Stefan Theussl\cr
Y.Ye (original matlab version of solnp)
}
\examples{
benchmarkids()
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/csolnp.R
\name{csolnp}
\alias{csolnp}
\title{Nonlinear optimization using augmented Lagrange method (C++ version)}
\usage{
csolnp(
  pars,
  fn,
  gr = NULL,
  eq_fn = NULL,
  eq_b = NULL,
  eq_jac = NULL,
  ineq_fn = NULL,
  ineq_lower = NULL,
  ineq_upper = NULL,
  ineq_jac = NULL,
  lower = NULL,
  upper = NULL,
  control = list(),
  use_r_version = FALSE,
  ...
)
}
\arguments{
\item{pars}{an numeric vector of decision variables (length n).}

\item{fn}{the objective function (must return a scalar).}

\item{gr}{an optional function for computing the analytic gradient of the function (must return
a vector of length n).}

\item{eq_fn}{an optional function for calculating equality constraints.}

\item{eq_b}{a vector of the equality bounds (if eq_fn provided).}

\item{eq_jac}{an optional function for computing the analytic jacobian of the equality.
function (a matrix with number of columns n and number of rows the same length as the number of equalities).}

\item{ineq_fn}{an optional function for calculating inequality constraints.}

\item{ineq_lower}{the lower bounds for the inequality (must be finite)}

\item{ineq_upper}{the upper bounds for the inequalitiy (must be finite)}

\item{ineq_jac}{an optional function for computing the analytic jacobian of the inequality (a matrix
with number of columns n and number of rows the same length as the number of inequalities).}

\item{lower}{lower bounds for the parameters. This is strictly required.}

\item{upper}{upper bounds for the parameters. This is strictly required.}

\item{control}{a list of solver control parameters (see details).}

\item{use_r_version}{(logical) used for debugging and validation. Uses the R version of the solver
rather than the C++ version. Will be deprecated in future releases.}

\item{...}{additional arguments passed to the supplied functions (common to all functions supplied).}
}
\value{
A list with the following slot:
\describe{
\item{pars}{ The parameters at the optimal solution found.}
\item{objective}{ The value of the objective at the optimal solution found.}
\item{objective_history}{ A vector of objective values obtained at each outer iteration.}
\item{out_iterations}{The number of outer iterations used to arrive at the solution.}
\item{convergence}{The convergence code (0 = converged).}
\item{message}{The convergence message.}
\item{kkt_diagnostics}{A list of optimal solution diagnostics.}
\item{lagrange}{ The vector of Lagrange multipliers at the optimal solution found.}
\item{n_eval}{ The number of function evaluations.}
\item{elapsed}{ The time taken to find a solution.}
\item{hessian}{ The Hessian at the optimal solution.}
}
}
\description{
Nonlinear optimization using augmented Lagrange method (C++ version)
}
\details{
The optimization problem solved by \code{csolnp} is formulated as:
\deqn{
\begin{aligned}
\min_{x \in \mathbb{R}^n} \quad & f(x) \\
\text{s.t.} \quad & g(x) = b \\
& h_l \le h(x) \le h_u\\
& x_l \le x \le x_u\\
\end{aligned}
}

where \eqn{f(x)} is the objective function, \eqn{g(x)} is the vector of equality constraints
with target value \eqn{b}, \eqn{h(x)} is the vector of inequality constraints bounded
by \eqn{h_l} and \eqn{h_u}, with parameter bounds \eqn{x_l} and \eqn{x_u}. Internally,
inequality constraints are converted into equality constraints using slack variables
and solved using an augmented Lagrangian approach.
This function is based on the original R code, but converted to C++, making use of
\code{Rcpp} and \code{RcppArmadillo}.
Additionally, it allows the user to pass in analytic gradient and Jacobians, else
finite differences using functions from the \code{numDeriv} package are used.

The control list consists of the following options:
\describe{
\item{rho}{Numeric. Initial penalty parameter for the augmented Lagrangian. Controls the weight given to constraint violation in the objective. Default is \code{1}.}
\item{max_iter}{Integer. Maximum number of major (outer) iterations allowed. Default is \code{400}.}
\item{min_iter}{Integer. Maximum number of minor (inner) iterations (per major iteration) for the quadratic subproblem solver. Default is \code{800}.}
\item{tol}{Numeric. Convergence tolerance for both feasibility (constraint violation) and optimality (change in objective). The algorithm terminates when changes fall below this threshold. Default is \code{1e-8}.}
\item{trace}{Integer If \code{1}, prints progress, \code{2} includes diagnostic information during optimization. Default is \code{0}.}
}

Tracing information provides the following:
\describe{
\item{Iter}{The current major iteration number.}
\item{Obj}{The value of the objective function \eqn{f(x)} at the current iterate.}
\item{||Constr||}{The norm of the current constraint violation, summarizing how well all constraints (equality and inequality) are satisfied. Typically the Euclidean or infinity norm.}
\item{RelObj}{The relative change in the objective function value compared to the previous iteration, i.e., \eqn{|f_k - f_{k-1}| / max(1, |f_{k-1}|)}.}
\item{Step}{The norm of the parameter update taken in this iteration, i.e., \eqn{||x_k - x_{k-1}||}.}
\item{Penalty}{The current value of the penalty parameter (\eqn{\rho}) in the augmented Lagrangian. This parameter is adaptively updated to balance objective minimization and constraint satisfaction.}
}
}
\author{
Alexios Galanos
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gosolnp.R
\name{startpars}
\alias{startpars}
\title{Generates and returns a set of starting parameters by sampling the parameter
space based on the evaluation of the function and constraints.}
\usage{
startpars(
  pars = NULL,
  fixed = NULL,
  fun,
  eqfun = NULL,
  eqB = NULL,
  ineqfun = NULL,
  ineqLB = NULL,
  ineqUB = NULL,
  LB = NULL,
  UB = NULL,
  distr = rep(1, length(LB)),
  distr.opt = list(),
  n.sim = 20000,
  cluster = NULL,
  rseed = NULL,
  bestN = 15,
  eval.type = 1,
  trace = FALSE,
  ...
)
}
\arguments{
\item{pars}{The starting parameter vector. This is not required unless the fixed option is
also used.}

\item{fixed}{The numeric index which indicates those parameters which should stay fixed
instead of being randomly generated.}

\item{fun}{The main function which takes as first argument the parameter vector and returns
a single value.}

\item{eqfun}{(Optional) The equality constraint function returning the vector of evaluated
equality constraints.}

\item{eqB}{(Optional) The equality constraints.}

\item{ineqfun}{(Optional) The inequality constraint function returning the vector of evaluated
inequality constraints.}

\item{ineqLB}{(Optional) The lower bound of the inequality constraints.}

\item{ineqUB}{(Optional) The upper bound of the inequality constraints.}

\item{LB}{The lower bound on the parameters. This is not optional in this function.}

\item{UB}{The upper bound on the parameters. This is not optional in this function.}

\item{distr}{A numeric vector of length equal to the number of parameters, indicating the
choice of distribution to use for the random parameter generation. Choices are
uniform (1), truncated normal (2), and normal (3).}

\item{distr.opt}{If any choice in \code{distr} was anything other than uniform (1), this is a
list equal to the length of the parameters with sub-components for the mean and
sd, which are required in the truncated normal and normal distributions.}

\item{n.sim}{The number of random parameter sets to generate.}

\item{cluster}{If you want to make use of parallel functionality, initialize and pass a cluster
object from the parallel package (see details), and remember to terminate it!}

\item{rseed}{(Optional) A seed to initiate the random number generator, else system time will
be used.}

\item{bestN}{The best N (less than or equal to n.sim) set of parameters to return.}

\item{eval.type}{Either 1 (default) for the direction evaluation of the function (excluding
inequality constraint violations) or 2 for the penalty barrier method.}

\item{trace}{(logical) Whether to display the progress of the function evaluation.}

\item{...}{(Optional) Additional parameters passed to the main, equality or inequality
functions}
}
\value{
A matrix of dimension bestN x (no.parameters + 1). The last column is the
evaluated function value.
}
\description{
A simple penalty barrier function is formed which is then evaluated at randomly
sampled points based on the upper and lower parameter bounds
(when \code{eval.type} = 2), else the objective function directly for values not
violating any inequality constraints (when \code{eval.type} = 1). The sampled
points can be generated from the uniform, normal or truncated normal
distributions.
}
\details{
Given a set of lower and upper bounds, the function generates, for those
parameters not set as fixed, random values from one of the 3 chosen
distributions. For simple functions with only inequality constraints, the direct
method (\code{eval.type} = 1) might work better. For more complex setups with
both equality and inequality constraints the penalty barrier method
(\code{eval.type} = 2)might be a better choice.
}
\note{
The choice of which distribution to use for randomly sampling the parameter
space should be driven by the user's knowledge of the problem and confidence or
lack thereof of the parameter distribution. The uniform distribution indicates a
lack of confidence in the location or dispersion of the parameter, while the
truncated normal indicates a more confident choice in both the location and
dispersion. On the other hand, the normal indicates perhaps a lack
of knowledge in the upper or lower bounds, but some confidence in the location
and dispersion of the parameter. In using choices (2) and (3) for \code{distr},
the \code{distr.opt} list must be supplied with \code{mean} and \code{sd} as
subcomponents for those parameters not using the uniform.
}
\examples{
\dontrun{
library(Rsolnp)
library(parallel)
# Windows
cl = makePSOCKcluster(2)
# Linux:
# makeForkCluster(nnodes = getOption("mc.cores", 2L), ...)

gofn = function(dat, n)
{

	x = dat[1:n]
	y = dat[(n+1):(2*n)]
	z = dat[(2*n+1):(3*n)]
	ii = matrix(1:n, ncol = n, nrow = n, byrow = TRUE)
	jj = matrix(1:n, ncol = n, nrow = n)
	ij = which(ii<jj, arr.ind = TRUE)
	i = ij[,1]
	j = ij[,2]
	#  Coulomb potential
	potential = sum(1.0/sqrt((x[i]-x[j])^2 + (y[i]-y[j])^2 + (z[i]-z[j])^2))
	potential
}

goeqfn = function(dat, n)
{
	x = dat[1:n]
	y = dat[(n+1):(2*n)]
	z = dat[(2*n+1):(3*n)]
	apply(cbind(x^2, y^2, z^2), 1, "sum")
}
n = 25
LB  = rep(-1, 3*n)
UB  = rep( 1, 3*n)
eqB = rep( 1,   n)

sp = startpars(pars = NULL, fixed = NULL, fun = gofn , eqfun = goeqfn,
eqB = eqB, ineqfun = NULL, ineqLB = NULL, ineqUB = NULL, LB = LB, UB = UB,
distr = rep(1, length(LB)), distr.opt = list(), n.sim = 2000,
cluster = cl, rseed = 100, bestN = 15, eval.type = 2, n = 25)
#stop cluster
stopCluster(cl)
# the last column is the value of the evaluated function (here it is the barrier
# function since eval.type = 2)
print(round(apply(sp, 2, "mean"), 3))
# remember to remove the last column
ans = solnp(pars=sp[1,-76],fun = gofn , eqfun = goeqfn , eqB = eqB, ineqfun = NULL,
ineqLB = NULL, ineqUB = NULL, LB = LB, UB = UB, n = 25)
# should get a value of around 243.8162
}
}
\author{
Alexios Galanos and Stefan Theussl\cr
}

\name{benchmark}
\Rdversion{1.1}
\alias{benchmark}
\title{
The Rsolnp Benchmark Problems Suite.
}
\description{
The function implements a set of benchmark problems against the MINOS solver of 
Murtagh and Saunders.
}
\usage{
benchmark(id = "Powell")
}
\arguments{
  \item{id}{
The name of the benchmark problem. A call to the function \code{\link{benchmarkids}}
will return the available benchmark problems.
}
}
\details{
The benchmarks were run on dual xeon server with 24GB of memory and windows 7 
operating system. The MINOS solver was used via the tomlab interface.
}
\value{
A data.frame containing the benchmark data. The description of the benchmark 
problem can be accessed throught the \code{description} attribute of 
the data.frame.
}
\references{
W.Hock and K.Schittkowski, \emph{Test Examples for Nonlinear Programming Codes}, 
Lecture Notes in Economics and Mathematical Systems. Springer Verlag, 1981.\cr 
Y.Ye, \emph{Interior algorithms for linear, quadratic, and linearly constrained 
non linear programming}, PhD Thesis, Department of EES Stanford University, 
Stanford CA.\cr
B.A.Murtagh and M.A.Saunders, \emph{MINOS 5.5 User's Guide, Report SOL 83-20R}, 
Systems Optimization Laboratory, Stanford University (revised July 1998).\cr
P. E. Gill, W. Murray, and M. A. Saunders, \emph{SNOPT An SQP algorithm for 
large-scale constrained optimization}, SIAM J. Optim., 12 (2002), pp.979-1006.
}
\author{
Alexios Ghalanos and Stefan Theussl\cr
Y.Ye (original matlab version of solnp)
}
\examples{
\dontrun{
benchmarkids()
benchmark(id = "Powell")
benchmark(id = "Alkylation")
benchmark(id = "Box")
benchmark(id = "RosenSuzuki")
benchmark(id = "Wright4")
benchmark(id = "Wright9")
benchmark(id = "Electron")
benchmark(id = "Permutation")
# accessing the description
test = benchmark(id = "Entropy")
attr(test, "description")
}
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solnp.R
\name{solnp}
\alias{solnp}
\title{Nonlinear optimization using augmented Lagrange method (original version)}
\usage{
solnp(
  pars,
  fun,
  eqfun = NULL,
  eqB = NULL,
  ineqfun = NULL,
  ineqLB = NULL,
  ineqUB = NULL,
  LB = NULL,
  UB = NULL,
  control = list(),
  ...
)
}
\arguments{
\item{pars}{an numeric vector of decision variables (length n).}

\item{fun}{the objective function (must return a scalar).}

\item{eqfun}{an optional function for calculating equality constraints.}

\item{eqB}{a vector of the equality bounds (if eq_fn provided).}

\item{ineqfun}{an optional function for calculating inequality constraints.}

\item{ineqLB}{the lower bounds for the inequality (must be finite)}

\item{ineqUB}{the upper bounds for the inequalitiy (must be finite)}

\item{LB}{lower bounds on decision variables}

\item{UB}{upper bounds on decision variables}

\item{control}{a list of solver control parameters (see details).}

\item{...}{additional arguments passed to the supplied functions (common to all functions supplied).}
}
\value{
An list with the following slot:
\describe{
\item{pars}{Optimal Parameters.}
\item{convergence }{Indicates whether the solver has converged (0) or not (1 or 2).}
\item{values}{Vector of function values during optimization with last one the
value at the optimal.}
\item{lagrange}{The vector of Lagrange multipliers.}
\item{hessian}{The Hessian of the augmented problem at the optimal solution.}
\item{ineqx0}{The estimated optimal inequality vector of slack variables used
for transforming the inequality into an equality constraint.}
\item{nfuneval}{The number of function evaluations.}
\item{elapsed}{Time taken to compute solution.}
}
}
\description{
Nonlinear optimization using augmented Lagrange method (original version)
}
\details{
The optimization problem solved by \code{csolnp} is formulated as:
\deqn{
\begin{aligned}
\min_{x \in \mathbb{R}^n} \quad & f(x) \\
\text{s.t.} \quad & g(x) = b \\
& h_l \le h(x) \le h_u\\
& x_l \le x \le x_u\\
\end{aligned}
}

where \eqn{f(x)} is the objective function, \eqn{g(x)} is the vector of equality constraints
with target value \eqn{b}, \eqn{h(x)} is the vector of inequality constraints bounded
by \eqn{h_l} and \eqn{h_u}, with parameter bounds \eqn{x_l} and \eqn{x_u}. Internally,
inequality constraints are converted into equality constraints using slack variables
and solved using an augmented Lagrangian approach.
The control is a list with the following options:
\describe{
\item{rho}{This is used as a penalty weighting scaler for infeasibility in the
augmented objective function. The higher its value the more the weighting to
bring the solution into the feasible region (default 1). However, very high
values might lead to numerical ill conditioning or significantly slow down
convergence.}
\item{outer.iter}{Maximum number of major (outer) iterations (default 400).}
\item{inner.iter}{Maximum number of minor (inner) iterations (default 800).}
\item{delta}{Relative step size in forward difference evaluation (default 1.0e-7).}
\item{tol}{ Relative tolerance on feasibility and optimality (default 1e-8).}
\item{trace}{The value of the objective function and the parameters is printed
at every major iteration (default 1).}
}
}
\examples{
{
# From the original paper by Y.Ye
# see the unit tests for more....
# POWELL Problem
fn1 = function(x)
{
    exp(x[1] * x[2] * x[3] * x[4] * x[5])
}
eqn1 = function(x){
    z1 = x[1] * x[1] + x[2] * x[2] + x[3] * x[3] + x[4] * x[4] + x[5] * x[5]
    z2 = x[2] * x[3] - 5 * x[4] * x[5]
    z3 = x[1] * x[1] * x[1] + x[2] * x[2] * x[2]
    return(c(z1, z2, z3))
}
x0 = c(-2, 2, 2, -1, -1)
}
powell = solnp(x0, fun = fn1, eqfun = eqn1, eqB = c(10, 0, -1))
}
\author{
Alexios Galanos
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsolnp-package.R
\docType{package}
\name{Rsolnp-package}
\alias{Rsolnp}
\alias{Rsolnp-package}
\title{Rsolnp: General Non-Linear Optimization}
\description{
General Non-linear Optimization Using Augmented Lagrange Multiplier Method.
}
\author{
\strong{Maintainer}: Alexios Galanos \email{alexios@4dscape.com} (\href{https://orcid.org/0009-0000-9308-0457}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Yinyu Ye (Original author of the solnp Matlab code, Stanford University)
}

Other contributors:
\itemize{
  \item Stefan Theussl \email{Stefan.Theussl@R-project.org} (\href{https://orcid.org/0000-0002-6523-4620}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gosolnp.R
\name{gosolnp}
\alias{gosolnp}
\title{Random Initialization and Multiple Restarts of the solnp solver.}
\usage{
gosolnp(
  pars = NULL,
  fixed = NULL,
  fun,
  eqfun = NULL,
  eqB = NULL,
  ineqfun = NULL,
  ineqLB = NULL,
  ineqUB = NULL,
  LB = NULL,
  UB = NULL,
  control = list(),
  distr = rep(1, length(LB)),
  distr.opt = list(),
  n.restarts = 1,
  n.sim = 20000,
  cluster = NULL,
  rseed = NULL,
  ...
)
}
\arguments{
\item{pars}{The starting parameter vector. This is not required unless the fixed option is
also used.}

\item{fixed}{The numeric index which indicates those parameters which should stay fixed
instead of being randomly generated.}

\item{fun}{The main function which takes as first argument the parameter vector and returns
a single value.}

\item{eqfun}{(Optional) The equality constraint function returning the vector of evaluated
equality constraints.}

\item{eqB}{(Optional) The equality constraints.}

\item{ineqfun}{(Optional) The inequality constraint function returning the vector of evaluated
inequality constraints.}

\item{ineqLB}{(Optional) The lower bound of the inequality constraints.}

\item{ineqUB}{(Optional) The upper bound of the inequality constraints.}

\item{LB}{The lower bound on the parameters. This is not optional in this function.}

\item{UB}{The upper bound on the parameters. This is not optional in this function.}

\item{control}{(Optional) The control list of optimization parameters. The \code{eval.type}
option in this control list denotes whether to evaluate the function as is and
exclude inequality violations in the final ranking (default, value = 1),
else whether to evaluate a penalty barrier function comprised of the objective
and all constraints (value = 2). See \code{solnp} function documentation for
details of the remaining control options.}

\item{distr}{A numeric vector of length equal to the number of parameters, indicating the
choice of distribution to use for the random parameter  generation. Choices are
uniform (1), truncated normal (2), and normal (3).}

\item{distr.opt}{If any choice in \code{distr} was anything other than uniform (1), this is a
list equal to the length of the parameters with sub-components for the mean and
sd, which are required in the truncated normal and normal distributions.}

\item{n.restarts}{The number of solver restarts required.}

\item{n.sim}{The number of random parameters to generate for every restart of the solver.
Note that there will always be significant rejections if inequality bounds are
present. Also, this choice should also be motivated by the width of the upper
and lower bounds.}

\item{cluster}{If you want to make use of parallel functionality, initialize and pass a cluster
object from the parallel package (see details), and remember to terminate it!}

\item{rseed}{(Optional) A seed to initiate the random number generator, else system time will
be used.}

\item{...}{(Optional) Additional parameters passed to the main, equality or inequality
functions}
}
\value{
A list containing the following values:
\item{pars}{Optimal Parameters.}
\item{convergence }{Indicates whether the solver has converged (0) or not (1).}
\item{values}{Vector of function values during optimization with last one the
value at the optimal.}
\item{lagrange}{The vector of Lagrange multipliers.}
\item{hessian}{The Hessian at the optimal solution.}
\item{ineqx0}{The estimated optimal inequality vector of slack variables used
for transforming the inequality into an equality constraint.}
\item{nfuneval}{The number of function evaluations.}
\item{elapsed}{Time taken to compute solution.}
\item{start.pars}{The parameter vector used to start the solver}
}
\description{
When the objective function is non-smooth or has many local minima, it is hard
to judge the optimality of the solution, and this usually depends critically on
the starting parameters. This function enables the generation of a set of
randomly chosen parameters from which to initialize multiple restarts of the
solver (see note for details).
}
\details{
Given a set of lower and upper bounds, the function generates, for those
parameters not set as fixed, random values from one of the 3 chosen
distributions. Depending on the \code{eval.type} option of the \code{control}
argument, the function is either directly evaluated for those points not
violating any inequality constraints, or indirectly via a penalty barrier
function jointly comprising the objective and constraints. The resulting values
are then sorted, and the best N (N = random.restart) parameter vectors
(corresponding to the best N objective function values) chosen in order to
initialize the solver. Since version 1.14, it is up to the user to prepare and
pass a cluster object from the parallel package for use with gosolnp, after
which the parLapply function is used. If your function makes use of additional
packages, or functions, then make sure to export them via the \code{clusterExport}
function of the parallel package. Additional arguments passed to the solver via the
\dots option are evaluated and exported by gosolnp to the cluster.
}
\note{
The choice of which distribution to use for randomly sampling the parameter
space should be driven by the user's knowledge of the problem and confidence or
lack thereof of the parameter distribution. The uniform distribution indicates
a lack of confidence in the location or dispersion of the parameter, while the
truncated normal indicates a more confident choice in both the location and
dispersion. On the other hand, the normal indicates perhaps a lack of knowledge
in the upper or lower bounds, but some confidence in the location and dispersion
of the parameter. In using choices (2) and (3) for \code{distr},
the \code{distr.opt} list must be supplied with \code{mean} and \code{sd} as
subcomponents for those parameters not using the uniform (the examples section
hopefully clarifies the usage).
}
\examples{
\dontrun{
# [Example 1]
# Distributions of Electrons on a Sphere Problem:
# Given n electrons, find the equilibrium state distribution (of minimal Coulomb
# potential) of the electrons positioned on a conducting sphere. This model is
# from the COPS benchmarking suite. See http://www-unix.mcs.anl.gov/~more/cops/.
gofn = function(dat, n)
{

	x = dat[1:n]
	y = dat[(n+1):(2*n)]
	z = dat[(2*n+1):(3*n)]
	ii = matrix(1:n, ncol = n, nrow = n, byrow = TRUE)
	jj = matrix(1:n, ncol = n, nrow = n)
	ij = which(ii<jj, arr.ind = TRUE)
	i = ij[,1]
	j = ij[,2]
	#  Coulomb potential
	potential = sum(1.0/sqrt((x[i]-x[j])^2 + (y[i]-y[j])^2 + (z[i]-z[j])^2))
	potential
}

goeqfn = function(dat, n)
{
	x = dat[1:n]
	y = dat[(n+1):(2*n)]
	z = dat[(2*n+1):(3*n)]
	apply(cbind(x^2, y^2, z^2), 1, "sum")
}

n = 25
LB = rep(-1, 3*n)
UB = rep(1,  3*n)
eqB = rep(1, n)
ans = gosolnp(pars  = NULL, fixed = NULL, fun = gofn, eqfun = goeqfn, eqB = eqB,
LB = LB, UB = UB, control = list(outer.iter = 100, trace = 1),
distr = rep(1, length(LB)), distr.opt = list(), n.restarts = 2, n.sim = 20000,
rseed = 443, n = 25)
# should get a function value around 243.813

# [Example 2]
# Parallel functionality for solving the Upper to Lower CVaR problem (not properly
# formulated...for illustration purposes only).

mu =c(1.607464e-04, 1.686867e-04, 3.057877e-04, 1.149289e-04, 7.956294e-05)
sigma = c(0.02307198,0.02307127,0.01953382,0.02414608,0.02736053)
R = matrix(c(1, 0.408, 0.356, 0.347, 0.378,  0.408, 1, 0.385, 0.565, 0.578, 0.356,
0.385, 1, 0.315, 0.332, 0.347, 0.565, 0.315, 1, 0.662, 0.378, 0.578,
0.332, 0.662, 1), 5,5, byrow=TRUE)
# Generate Random deviates from the multivariate Student distribution
set.seed(1101)
v = sqrt(rchisq(10000, 5)/5)
S = chol(R)
S = matrix(rnorm(10000 * 5), 10000) \%*\% S
ret = S/v
RT = as.matrix(t(apply(ret, 1, FUN = function(x) x*sigma+mu)))
# setup the functions
.VaR = function(x, alpha = 0.05)
{
	VaR = quantile(x, probs = alpha, type = 1)
	VaR
}

.CVaR = function(x, alpha = 0.05)
{
	VaR = .VaR(x, alpha)
	X = as.vector(x[, 1])
	CVaR = VaR - 0.5 * mean(((VaR-X) + abs(VaR-X))) / alpha
	CVaR
}
.fn1 = function(x,ret)
{
	port=ret\%*\%x
	obj=-.CVaR(-port)/.CVaR(port)
	return(obj)
}

# abs(sum) of weights ==1
.eqn1  = function(x,ret)
{
	sum(abs(x))
}

LB=rep(0,5)
UB=rep(1,5)
pars=rep(1/5,5)
ctrl = list(delta = 1e-10, tol = 1e-8, trace = 0)
cl = makePSOCKcluster(2)
# export the auxilliary functions which are used and cannot be seen by gosolnp
clusterExport(cl, c(".CVaR", ".VaR"))
ans = gosolnp(pars, fun = .fn1, eqfun = .eqn1, eqB = 1, LB = LB, UB = UB,
n.restarts = 2, n.sim=500, cluster = cl, ret = RT)
ans
# don't forget to stop the cluster!
stopCluster(cl)
}
}
\references{
Y.Ye, \emph{Interior algorithms for linear, quadratic, and linearly constrained
non linear programming}, PhD Thesis, Department of EES Stanford University,
Stanford CA.\cr
Hu, X. and Shonkwiler, R. and Spruill, M.C. \emph{Random Restarts in Global
Optimization}, 1994, Georgia Institute of technology, Atlanta.
}
\author{
Alexios Galanos and Stefan Theussl\cr
Y.Ye (original matlab version of solnp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/csolnp.R
\name{kkt_diagnose}
\alias{kkt_diagnose}
\title{Summarize KKT Condition Diagnostics}
\usage{
kkt_diagnose(kkt, tol = 1e-08)
}
\arguments{
\item{kkt}{A named list containing numeric entries for KKT diagnostics.
Required names are \code{"kkt_stationarity"}, \code{"eq_violation"},
\code{"ineq_violation"}, \code{"dual_feas_violation"}, \code{"compl_slackness"}.}

\item{tol}{Numeric tolerance for considering a condition as "satisfied". Default is \code{1e-8}.}
}
\value{
An object of class "solnp_kkt_summary" (a data.frame with columns: condition, value, status, tol).
}
\description{
Given a list of KKT diagnostic statistics (stationarity, feasibility, complementarity, etc.),
this function prints a clear summary indicating which KKT conditions are satisfied at a specified tolerance.
}
\examples{
kkt <- list(
  kkt_stationarity = 5.828909e-06,
  eq_violation = 0,
  ineq_violation = 0,
  dual_feas_violation = 0.4380053,
  compl_slackness = 0
)
kkt_diagnose(kkt, tol = 1e-8)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmark_suite.R
\name{solnp_problem_suite}
\alias{solnp_problem_suite}
\title{Retrieve Implemented Test Problems for the SOLNP Suite}
\usage{
solnp_problem_suite(
  suite = "Hock-Schittkowski",
  number = 1,
  return_all = FALSE
)
}
\arguments{
\item{suite}{Character. The test suite to draw from. Must be one of \dQuote{Hock-Schittkowski} or \dQuote{Other}.
Default is \dQuote{Hock-Schittkowski}.}

\item{number}{Integer or vector of integers. One or more problem numbers to retrieve. Ignored if return_all = TRUE.}

\item{return_all}{Logical. If TRUE, returns all implemented problems in the specified suite. Default is FALSE.}
}
\value{
If one problem is requested and implemented, the evaluated problem object is returned directly.
Otherwise, an unnamed list of evaluated problem objects is returned.
}
\description{
Returns a list (or a single object) of implemented test problems corresponding to a selected suite.
Problem functions must follow the naming convention \sQuote{problem_name_problem} and return a list
describing the optimization problem (e.g., objective, constraints, bounds).
}
\details{
\itemize{
\item Problems are matched by number within the selected suite, using the table from \code{\link[=solnp_problems_table]{solnp_problems_table()}}.
\item If a requested problem is valid but not yet implemented (i.e., the corresponding function does not exist),
a message will inform the user.
\item If a problem number exceeds the allowable range (e.g., > 306 for Hock-Schittkowski), an error is raised.
}
}
\examples{
\dontrun{
# Retrieve a single HS problem
prob <- solnp_problem_suite(number = 1)

# Retrieve multiple HS problems
probs <- solnp_problem_suite(number = c(1, 2, 3))

# Retrieve problem in "Other" suite
other_prob <- solnp_problem_suite(suite = "Other", number = 1)
}

}
\seealso{
\code{\link[=solnp_problems_table]{solnp_problems_table()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmark_suite.R
\name{solnp_problems_table}
\alias{solnp_problems_table}
\title{List of Valid Test Problems for the SOLNP Suite}
\usage{
solnp_problems_table()
}
\value{
A data.frame with the following columns:
\describe{
\item{Suite}{A character string indicating the suite the problem belongs to.
One of \dQuote{Hock-Schittkowski} or \dQuote{Other}.}
\item{Problem}{The base name of the problem function (without the \sQuote{_problem} suffix).}
\item{Number}{An integer identifier used to index or request problems programmatically.}
}
}
\description{
Returns a data.frame of known and registered test problems used with the SOLNP solver.
The list includes problems from the Hock-Schittkowski suite as well as a selection
of other classic optimization problems.
}
\details{
\itemize{
\item All problem functions are expected to follow the naming convention
\sQuote{Problem_problem} (e.g., \sQuote{hs01_problem}).
\item For Hock-Schittkowski problems, numbers range from 1 to 64, with a few
selected extras (e.g., 110, 118, 119).
\item The \dQuote{Other} suite includes named problems like \sQuote{box},
\sQuote{alkylation}, \sQuote{entropy}, \sQuote{garch}, etc.,
and are numbered sequentially.
}
}
\examples{
# View all known problems
tail(solnp_problems_table())

# Filter only HS problems
head(subset(solnp_problems_table(), Suite == "Hock-Schittkowski"))

}
\seealso{
\code{\link[=solnp_problem_suite]{solnp_problem_suite()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_start.R
\name{csolnp_ms}
\alias{csolnp_ms}
\title{Multi-start version of csolnp}
\usage{
csolnp_ms(
  fn,
  gr = NULL,
  eq_fn = NULL,
  eq_b = NULL,
  eq_jac = NULL,
  ineq_fn = NULL,
  ineq_lower = NULL,
  ineq_upper = NULL,
  ineq_jac = NULL,
  lower = NULL,
  upper = NULL,
  control = list(),
  n_candidates = 20,
  penalty = 10000,
  eq_tol = 1e-06,
  ineq_tol = 1e-06,
  seed = NULL,
  return_all = FALSE,
  ...
)
}
\arguments{
\item{fn}{the objective function (must return a scalar).}

\item{gr}{an optional function for computing the analytic gradient of the
function (must return a vector of length n).}

\item{eq_fn}{an optional function for calculating equality constraints.}

\item{eq_b}{a vector of the equality bounds (if eq_fn provided).}

\item{eq_jac}{an optional function for computing the analytic Jacobian of the
equality function (a matrix with number of columns n and number of rows
equal to the number of equalities).}

\item{ineq_fn}{an optional function for calculating inequality constraints.}

\item{ineq_lower}{the lower bounds for the inequality constraints (must be finite).}

\item{ineq_upper}{the upper bounds for the inequality constraints (must be finite).}

\item{ineq_jac}{an optional function for computing the analytic Jacobian of the
inequality function (a matrix with number of columns n and number of rows equal
to the number of inequalities).}

\item{lower}{lower bounds for the parameters. This is strictly required.}

\item{upper}{upper bounds for the parameters. This is strictly required.}

\item{control}{a list of solver control parameters (see details).}

\item{n_candidates}{integer. The number of initial feasible candidate points
to generate for multi-start optimization. Default is 20.}

\item{penalty}{numeric. The penalty parameter used when projecting to feasibility
for candidate generation. Default is 1e4.}

\item{eq_tol}{Numeric. Tolerance for equality constraint violation
(default is 1e-6). Candidate solutions with \code{kkt_diagnostics\$eq_violation}
less than or equal to this value are considered feasible with respect
to equality constraints.}

\item{ineq_tol}{Numeric. Tolerance for inequality constraint violation
(default is 1e-6). Candidate solutions with \code{kkt_diagnostics\$ineq_violation}
less than or equal to this value are considered feasible with respect to
inequality constraints.}

\item{seed}{an optional random seed used to initialize the random number
generator for the random samples.}

\item{return_all}{logical. Whether to return all solutions as a list. This
may be useful for debugging.}

\item{...}{additional arguments passed to the supplied functions
(common to all functions supplied).}
}
\value{
A list containing the best solution found by multi-start, with elements analogous
to those returned by \code{csolnp}. If \code{return_all} is TRUE, then a list of
all solutions is returned instead.
}
\description{
Runs the csolnp solver from multiple diverse feasible starting values and
returns the best solution found.
}
\details{
This function automates the process of generating multiple feasible starting
points (using lower, upper, and constraint information), runs \code{csolnp} from
each, and returns the solution with the lowest objective value. It is useful
for problems where local minima are a concern or the objective surface is
challenging.

\strong{Candidate Generation:}

The \code{generate_feasible_starts} approach creates a diverse set of initial
parameter vectors (candidates) that are feasible with respect to box and
(optionally) nonlinear constraints. The process is as follows:

\enumerate{
\item For each candidate, a random point is sampled inside the parameter
box constraints (\code{lower} and \code{upper}) but a small distance away
from the boundaries, to avoid numerical issues. This is achieved by a helper
function that applies a user-specified buffer (\code{eps}).
\item If nonlinear inequality constraints (\code{ineq_fn}) are provided,
each sampled point is projected towards the feasible region using a
fast penalized minimization. This step does not solve the feasibility problem
exactly, but quickly produces a point that satisfies the constraints to
within a specified tolerance, making it suitable as a starting point for
optimization.
\item If only box constraints are present, the sampled point is used
directly as a feasible candidate.
\item The set of feasible candidates is ranked by the objective with lower
values considered better. This allows prioritization of candidates that start
closer to optimality.
}

This method efficiently creates a diverse set of robust initial values, improving
the chances that multi-start optimization will identify the global or a
high-quality local solution, especially in the presence of non-convexities or
challenging constraint boundaries.
\strong{Solution Selection:}
For each candidate starting point, \code{csolnp_ms} runs the \code{csolnp} solver
and collects the resulting solutions and their associated KKT diagnostics. If
equality or inequality constraints are present, candidate solutions are first
filtered to retain only those for which the maximum violation of
equality (\code{kkt_diagnostics\$eq_violation}) and/or
inequality (\code{kkt_diagnostics\$ineq_violation}) constraints are less than
or equal to user-specified tolerances (\code{eq_tol} and \code{ineq_tol}).
Among the feasible solutions (those satisfying all constraints within tolerance),
the solution with the lowest objective value is selected and returned as the
best result. If no candidate fully satisfies the constraints, the solution with
the smallest total constraint violation is returned, with a warning issued to
indicate that strict feasibility was not achieved. This two-stage selection
process ensures that the final result is both feasible (when possible) and
optimally minimizes the objective function among all feasible candidates.
}
\seealso{
\code{\link{csolnp}}
}
\author{
Alexios Galanos
}
